/*****************************************************************************
 *
 * oacam.h -- camera API header
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#ifndef OPENASTRO_CAMERA_H
#define OPENASTRO_CAMERA_H

#include <openastro/oacam/controls.h>
#include <openastro/oacam/formats.h>
#include <openastro/oacam/features.h>
#include <openastro/errno.h>
#include <openastro/debug.h>

enum oaInterfaceType {
  OA_IF_V4L2				= 1,
  OA_IF_PWC				= 2,
  OA_IF_ZWASI				= 3,
  OA_IF_QHY				= 4,
  OA_IF_IIDC				= 5,
  OA_IF_UVC				= 6
};

// This may need looking at...

#define	OA_MAX_RESOLUTIONS		31
#define	OA_MAX_FRAMERATES		31

typedef struct FRAMESIZE {
  int			x;
  int			y;
} FRAMESIZE;

typedef struct FRAMERATE {
  int			numerator;
  int			denominator;
} FRAMERATE;

typedef struct START_PARMS {
  FRAMESIZE	size;
  FRAMERATE	rate;
} START_PARMS;

// ...to here

struct oaCamera;
struct oaDevice;

typedef struct oaCameraFuncs {
  struct oaCamera* ( *initCamera )( struct oaDevice* );
  int              ( *closeCamera )( struct oaCamera* );
  int              ( *resetCamera )( struct oaCamera* );
  int              ( *startCapture )( struct oaCamera* );
  int              ( *startCaptureExtended )( struct oaCamera*,
                       START_PARMS* );
  int              ( *stopCapture )( struct oaCamera* );

  int              ( *hasLoadableFirmware )( struct oaCamera* );
  int              ( *isFirmwareLoaded )( struct oaCamera* );
  int              ( *loadFirmware )( struct oaCamera* );

  const char*      ( *getName )( struct oaCamera* );
  const char*      ( *getMenuString )( struct oaCamera*, int, int );

  int              ( *getControlRange )( struct oaCamera*, int, int*, int*,
                       int*, int* );
  int              ( *readControl )( struct oaCamera*, int );
  int              ( *setControl )( struct oaCamera*, int, int );
  int              ( *testControl )( struct oaCamera*, int, int );
  int              ( *setControlMulti )( struct oaCamera* );
  int              ( *getAutoWBManualSetting )( struct oaCamera* );
  int              ( *hasAuto )( struct oaCamera*, int );
  int              ( *isAuto )( struct oaCamera*, int );

  int              ( *startReadFrame )( struct oaCamera*, unsigned int );
  int              ( *readFrame )( struct oaCamera*, void** );	
  int              ( *finishReadFrame )( struct oaCamera* );

  const FRAMESIZE* ( *enumerateFrameSizes )( struct oaCamera* );
  const FRAMERATE* ( *enumerateFrameRates )( struct oaCamera*, int, int );
  int	           ( *hasFixedFrameRates )( struct oaCamera*, int, int );

  int              ( *getFramePixelFormat )( struct oaCamera*, int );
  int              ( *testROISize )( struct oaCamera*, int, int, int*, int* );
} oaCameraFuncs;

#define OA_MAX_NAME_LEN			40
#define OA_MAX_DEVICES			32

#if HAVE_LIBDC1394
#include <openastro/oacam/IIDCstate.h>
#endif
#include <openastro/oacam/PWCstate.h>
#include <openastro/oacam/QHYstate.h>
#if HAVE_LIBV4L2
#include <openastro/oacam/V4L2state.h>
#endif
#include <openastro/oacam/ZWASIstate.h>
#if HAVE_LIBUVC
#include <openastro/oacam/UVCstate.h>
#endif

typedef struct oaCamera {
  enum oaInterfaceType  interface;
  char			deviceName[OA_MAX_NAME_LEN+1];
  oaCameraFuncs		funcs;
  uint8_t		controls[OA_CAM_CTRL_LAST_P1];
  uint32_t              min[ OA_CAM_CTRL_LAST_P1 ];
  uint32_t              max[ OA_CAM_CTRL_LAST_P1 ];
  uint32_t              step[ OA_CAM_CTRL_LAST_P1 ];
  uint32_t              def[ OA_CAM_CTRL_LAST_P1 ];
  oaCameraFeatures	features;
  union {
#if HAVE_LIBDC1394
    IIDC_STATE           _iidc;
#endif
    PWC_STATE           _pwc;
    QHY_STATE           _qhy;
#if HAVE_LIBV4L2
    V4L2_STATE          _v4l2;
#endif
    ZWASI_STATE         _zwasi;
#if HAVE_LIBUVC
    UVC_STATE           _uvc;
#endif
  };
} oaCamera;

typedef struct oaDevice {
  enum oaInterfaceType	interface;
  char			deviceName[OA_MAX_NAME_LEN+1];
  oaCamera*		( *initCamera )( struct oaDevice* );
  int           	( *hasLoadableFirmware )( struct oaDevice* );
  int           	( *isFirmwareLoaded )( struct oaDevice* );
  int           	( *loadFirmware )( struct oaDevice* );
  unsigned long		_devIndex;
  unsigned long		_devType;
  unsigned short	_vendorId;
  unsigned short	_productId;
  unsigned long		_misc;
  uint64_t		_guid;
  uint8_t		_unit;
  unsigned int		_colour;
  char			sysPath[ PATH_MAX+1 ];
  unsigned int		hasFirmware;
  unsigned int		firmwareLoaded;
} oaDevice;

extern int		oaGetCameras ( oaDevice** );
extern unsigned		oaGetCameraAPIVersion ( void );
extern const char*	oaGetCameraAPIVersionStr ( void );
extern void		oaSetCameraDebugLevel ( int );
extern int		oaGetAutoForControl ( int );
extern int		oaGetControlForAuto ( int );
extern int		oaIsAuto ( int );
extern void		oaSetRootPath ( const char* );

// These may no longer be required

#define OA_FRAMESIZES_DISCRETE		1
#define OA_FRAMESIZES_CONTINUOUS	2
#define OA_FRAMESIZES_STEPWISE		3

#define OA_FRAMERATES_DISCRETE		1
#define OA_FRAMERATES_CONTINUOUS	2
#define OA_FRAMERATES_STEPWISE		3

#define	OA_BIN_MODE_NONE		1
#define	OA_BIN_MODE_2x2			2
#define	OA_BIN_MODE_3x3			3
#define	OA_BIN_MODE_4x4			4

#endif	/* OPENASTRO_CAMERA_H */
