/*****************************************************************************
 *
 * IIDCcontrol.c -- control functions for IIDC cameras
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#if HAVE_LIBDC1394

#include <openastro/oacam.h>

#include "IIDCoacam.h"

#define	cameraState		camera->_iidc


int
oaIIDCCameraSetControl ( oaCamera* camera, int control, int val )
{
  int i, found, iidcControl;

  for ( i = 0, found = -1; found < 0 && i < numIIDCControls; i++ ) {
    if ( dc1394Controls[i].oaControl == control ||
        dc1394Controls[i].oaAutoControl == control ) {
      iidcControl = dc1394Controls[i].iidcControl;
      found = i;
    }
  }

  if ( OA_CAM_CTRL_EXPOSURE == control || OA_CAM_CTRL_EXPOSURE_ABSOLUTE ==
      control ) {
    iidcControl = DC1394_FEATURE_SHUTTER;
  }

  if ( oaIsAuto ( control )) {
    if ( dc1394_feature_set_mode ( cameraState.iidcHandle, iidcControl,
        val ? DC1394_FEATURE_MODE_AUTO : DC1394_FEATURE_MODE_MANUAL ) !=
        DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_feature_set_mode failed for control %d\n",
          __FUNCTION__, iidcControl );
      return -1;
    }
    return 0;
  }

  if ( found >= 0 || OA_CAM_CTRL_EXPOSURE == control ) {
    if ( dc1394_feature_set_value ( cameraState.iidcHandle, iidcControl,
        val ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_feature_set_value failed for "
          "control %d\n", __FUNCTION__, control );
    }
    return 0;
  }

  if ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE == control ) {
    // need to handle absolute exposure time separately
    float decval;
    decval = val / 1000000.0;
    if ( dc1394_feature_set_absolute_value ( cameraState.iidcHandle,
        iidcControl, decval ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_feature_set_absolute_value %d failed\n",
          __FUNCTION__, control );
      return -1;
    }
    return 0;
  }

  fprintf ( stderr, "Unrecognised control %d in %s\n", control, __FUNCTION__ );
  return -1;

}


int
oaIIDCCameraReadControl ( oaCamera* camera, int control )
{
  int i, found, iidcControl;
  unsigned int val;

  for ( i = 0, found = 0; !found && i < numIIDCControls; i++ ) {
    if ( dc1394Controls[i].oaControl == control ||
        dc1394Controls[i].oaAutoControl == control ) {
      iidcControl = dc1394Controls[i].iidcControl;
      found = 1;
    }
  }

  if ( found ) {
    if ( iidcControl != DC1394_FEATURE_SHUTTER || OA_CAM_CTRL_EXPOSURE ==
        control || OA_CAM_CTRL_AUTO_EXPOSURE == control ) {
      if ( !oaIsAuto ( control )) {
        if ( dc1394_feature_get_value ( cameraState.iidcHandle, iidcControl,
            &val ) != DC1394_SUCCESS ) {
          fprintf ( stderr, "%s: dc1394_feature_get_value failed for "
              "control %d\n", __FUNCTION__, control );
        val = 0;
        }
      } else {
        dc1394feature_mode_t mode;
        if ( dc1394_feature_get_mode ( cameraState.iidcHandle, iidcControl,
            &mode ) != DC1394_SUCCESS ) {
          fprintf ( stderr, "%s: dc1394_feature_get failed for feature %d\n",
              __FUNCTION__, iidcControl );
          val = 0;
        } else {
          val = ( mode == DC1394_FEATURE_MODE_AUTO ) ? 1 : 0;
        }
      }
    }
    return val;
  }

  if ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE == control ) {
    // need to handle absolute exposure time separately
    float decval;
    if ( dc1394_feature_get_absolute_value ( cameraState.iidcHandle,
        DC1394_FEATURE_SHUTTER, &decval ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_feature_get_absolute_value %d failed\n",
          __FUNCTION__, iidcControl );
      val = 0;
    } else {
      val = decval * 1000000.0;
    }
    return val;
  }

  fprintf ( stderr, "Unrecognised control %d in %s\n", control, __FUNCTION__ );
  return 0;
}


int
oaIIDCCameraTestControl ( oaCamera* camera, int control, int arg1 )
{
  // FIX ME -- this is horribly wrong.  Should be done by control type
  // rather than by the actual control itself

  switch ( control ) {

    case OA_CAM_CTRL_BRIGHTNESS:
    case OA_CAM_CTRL_CONTRAST:
    case OA_CAM_CTRL_SATURATION:
    case OA_CAM_CTRL_HUE:
    case OA_CAM_CTRL_SHARPNESS:
    case OA_CAM_CTRL_GAMMA:
    case OA_CAM_CTRL_WHITE_BALANCE_TEMP:
    case OA_CAM_CTRL_GAIN:
    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      if ( arg1 >= camera->min[ control ] &&
          arg1 <= camera->max[ control ] &&
          ( 0 == ( arg1 - camera->min[ control ] ) %
          camera->step[ control ] )) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CAM_CTRL_AUTO_WHITE_BALANCE:
    case OA_CAM_CTRL_AUTO_WHITE_BALANCE_TEMP:
    case OA_CAM_CTRL_HUE_AUTO:
      return OA_ERR_NONE;
      break;

    case OA_CAM_CTRL_AUTO_EXPOSURE:
      // We're not currently going to worry about shutter priority or
      // aperture priority here
      // FIX ME -- probably no longer true
      return OA_ERR_NONE;
      break;     

    case OA_CAM_CTRL_BIT_DEPTH:
    case OA_CAM_CTRL_BINNING:
      return -OA_ERR_INVALID_CONTROL;
      break;

    default:
      // If we reach here it's because we don't recognise the control
      // FIX ME -- make this debug only
      fprintf ( stderr, "Unrecognised control %d in %s\n", control,
          __FUNCTION__ );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }

  // And if we reach here it's because the value wasn't valid
  return -OA_ERR_OUT_OF_RANGE;
}


int
oaIIDCCameraSetROI ( oaCamera* camera, int x, int y )
{
  fprintf ( stderr, "attempt to set ROI of %dx%d in %s\n", x, y,
      __FUNCTION__ );
  return 0;
}


int
oaIIDCCameraStart ( oaCamera* camera, START_PARMS* parms )
{
  // configure the camera for the requested resolution and frame rate
  // with the current colour mode and start capturing frames

  int i, j, thisOne, requiredFrameRate, ret;
  dc1394video_modes_t   videoModes;
  dc1394color_coding_t  codec;
  dc1394framerates_t    framerates;

  requiredFrameRate = 0;
  for ( i = 0; i < numIIDCFrameRates; i++ ) {
    if ( dc1394FrameRates[i].numerator == parms->rate.numerator &&
        dc1394FrameRates[i].denominator == parms->rate.denominator ) {
      requiredFrameRate = dc1394FrameRates[i].iidcFrameRate;
    }
  }

  if ( !requiredFrameRate ) {
    fprintf ( stderr, "%s: no frame rate matching %d/%d found\n",
        __FUNCTION__, parms->rate.numerator, parms->rate.denominator );
    return -1;
  }

  if ( dc1394_video_get_supported_modes ( cameraState.iidcHandle,
      &videoModes ) != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_get_supported_modes failed",
        __FUNCTION__ );
    return -1;
  }

  for ( i = 0, thisOne = -1; thisOne < 0 && i < videoModes.num; i++ ) {
    uint32_t w, h;

    if ( videoModes.modes[i] >= DC1394_VIDEO_MODE_FORMAT7_MIN &&
        videoModes.modes[i] <= DC1394_VIDEO_MODE_FORMAT7_MAX ) {
      continue;
    }

    if ( dc1394_get_image_size_from_video_mode ( cameraState.iidcHandle,
        videoModes.modes[i], &w, &h ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_get_image_size_from_video_mode failed",
          __FUNCTION__ );
    } else {
      if ( w == parms->size.x && h == parms->size.y ) {
        if ( dc1394_get_color_coding_from_video_mode ( cameraState.iidcHandle,
            videoModes.modes[i], &codec ) != DC1394_SUCCESS ) {
          fprintf ( stderr, "%s: dc1394_get_color_coding_from_video_mode "
              "failed", __FUNCTION__ );
        } else {
          if ( codec == cameraState.currentCodec ) {
            if ( dc1394_video_get_supported_framerates (
                cameraState.iidcHandle, videoModes.modes[i], &framerates )
                != DC1394_SUCCESS ) {
              fprintf ( stderr, "%s: dc1394_video_get_supported_framerates "
                  "failed\n", __FUNCTION__ );
            } else {
              for ( j = 0; thisOne < 0 && j < framerates.num; j++ ) {
                if ( framerates.framerates[j] == requiredFrameRate ) {
                  thisOne = i;
                }
              }
            }
          }
        }
      }
    }
  }

  if ( thisOne < 0 ) {
    fprintf ( stderr, "%s: can't find compatible video mode", __FUNCTION__ );
    return -1;
  }

  if ( dc1394_video_set_mode ( cameraState.iidcHandle,
      videoModes.modes[ thisOne ]) != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_set_mode failed\n", __FUNCTION__ );
    return -1;
  }
  if ( dc1394_video_set_framerate ( cameraState.iidcHandle,
      requiredFrameRate ) != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_set_framerate failed\n", __FUNCTION__ );
    return -1;
  }
  if ( !cameraState.started ) {
    if (( ret = dc1394_capture_setup ( cameraState.iidcHandle,
        NUM_IIDC_BUFFERS, DC1394_CAPTURE_FLAGS_DEFAULT )) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_capture_setup failed: %d\n", __FUNCTION__,
          ret );
      return -1;
    }
  }
  cameraState.started = 1;
  if ( dc1394_video_set_transmission ( cameraState.iidcHandle, DC1394_ON ) !=
      DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_set_transmission failed\n",
        __FUNCTION__ );
    return -1;
  }

  return 0;
}


int
oaIIDCCameraStop ( oaCamera* camera )
{
  if ( dc1394_video_set_transmission ( cameraState.iidcHandle, DC1394_OFF )
      != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_set_transmission failed\n",
      __FUNCTION__ );
    return -1;
  }
  if ( cameraState.started ) {
    if ( dc1394_capture_stop ( cameraState.iidcHandle ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_capture_stop failed\n", __FUNCTION__ );
      return -1;
    }
    cameraState.started = 0;
  }
  return 0;
}


int
oaIIDCCameraReset ( oaCamera* camera )
{
  if ( dc1394_video_set_transmission ( cameraState.iidcHandle, DC1394_OFF )
      != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_set_transmission failed\n",
      __FUNCTION__ );
    return -1;
  }
  if ( cameraState.started ) {
    if ( dc1394_capture_stop ( cameraState.iidcHandle ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_capture_stop failed\n", __FUNCTION__ );
      return -1;
    }
    cameraState.started = 0;
  }
  if ( dc1394_capture_setup ( cameraState.iidcHandle, NUM_IIDC_BUFFERS,
      DC1394_CAPTURE_FLAGS_DEFAULT ) != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_capture_setup failed\n", __FUNCTION__ );
    return -1;
  }
  cameraState.started = 1;
  return 0;
}


const char*
oaIIDCCameraGetMenuString ( oaCamera* camera, int control, int index )
{
  static char buff[40];

  if ( control != OA_CAM_CTRL_AUTO_EXPOSURE ) {
    fprintf ( stderr, "%s: control not implemented\n", __FUNCTION__ );
    return "";
  }

  switch ( index ) {
    case 1:
      return "Manual";
      break;
    case 2:
      return "Auto";
      break;
    case 4:
      return "Shutter Priority";
      break;
    case 8:
      return "Aperture Priority";
      break;
  }

  return "Unknown";
}

#endif /* HAVE_LIBDC1394 */
