/*****************************************************************************
 *
 * IIDCinit.c -- Initialise IEEE1394/IIDC cameras
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#if HAVE_LIBDC1394

#include <openastro/oacam.h>
#include <dc1394/dc1394.h>

#include "IIDCoacam.h"
#include "oacamprivate.h"

#define	cameraState		camera->_iidc

static void _IIDCInitFunctionPointers ( oaCamera* );

struct iidcCtrl dc1394Controls[] = {
  { DC1394_FEATURE_BRIGHTNESS, OA_CAM_CTRL_BRIGHTNESS,
      OA_CAM_CTRL_AUTO_BRIGHTNESS },
  // The IIDC spec says this one is like contrast.  Who am I to argue?
  { DC1394_FEATURE_EXPOSURE, OA_CAM_CTRL_CONTRAST, OA_CAM_CTRL_AUTO_CONTRAST },
  { DC1394_FEATURE_SHARPNESS, OA_CAM_CTRL_SHARPNESS, 0 },
  { DC1394_FEATURE_WHITE_BALANCE, OA_CAM_CTRL_WHITE_BALANCE,
      OA_CAM_CTRL_AUTO_WHITE_BALANCE },
  { DC1394_FEATURE_HUE, OA_CAM_CTRL_HUE, OA_CAM_CTRL_HUE_AUTO },
  { DC1394_FEATURE_SATURATION, OA_CAM_CTRL_SATURATION, 0 },
  { DC1394_FEATURE_GAMMA, OA_CAM_CTRL_GAMMA, OA_CAM_CTRL_AUTO_GAMMA },
  { DC1394_FEATURE_SHUTTER, -1, -1 },
  { DC1394_FEATURE_GAIN, OA_CAM_CTRL_GAIN, OA_CAM_CTRL_AUTO_GAIN },
  { DC1394_FEATURE_IRIS, 0, 0 },
  { DC1394_FEATURE_FOCUS, 0, 0 },
  { DC1394_FEATURE_TEMPERATURE, 0, 0 }, // FIX ME -- actually we have this one
  { DC1394_FEATURE_TRIGGER, 0, 0 },
  { DC1394_FEATURE_TRIGGER_DELAY, 0, 0 },
  { DC1394_FEATURE_WHITE_SHADING, 0, 0 },
  { DC1394_FEATURE_FRAME_RATE, -1, -1 },
  { DC1394_FEATURE_ZOOM, 0, 0 },
  { DC1394_FEATURE_PAN, 0, 0 },
  { DC1394_FEATURE_TILT, 0, 0 },
  { DC1394_FEATURE_OPTICAL_FILTER, 0, 0 },
  { DC1394_FEATURE_CAPTURE_SIZE, 0, 0 },
  { DC1394_FEATURE_CAPTURE_QUALITY, 0, 0 }
};

unsigned int numIIDCControls = sizeof ( dc1394Controls ) /
    sizeof ( struct iidcCtrl );

struct iidcFrameRate dc1394FrameRates[] = {
  { DC1394_FRAMERATE_1_875, 8, 15 },
  { DC1394_FRAMERATE_3_75, 4, 14 },
  { DC1394_FRAMERATE_7_5, 2, 15 },
  { DC1394_FRAMERATE_15, 1, 15 },
  { DC1394_FRAMERATE_30, 1, 30 },
  { DC1394_FRAMERATE_60, 1, 60 },
  { DC1394_FRAMERATE_120, 1, 120 },
  { DC1394_FRAMERATE_240, 1, 240 }
};

unsigned int numIIDCFrameRates = sizeof ( dc1394FrameRates ) /
    sizeof ( struct iidcFrameRate );

/**
 * Initialise a given camera device
 */

oaCamera*
oaIIDCInitCamera ( oaDevice* device )
{
  dc1394_t*	        iidcContext;
  dc1394error_t	        ret;
  dc1394camera_t*       iidcCam;
  dc1394video_modes_t   videoModes;
  dc1394featureset_t    features;
  dc1394color_coding_t  codec;
  oaCamera*             camera;
  int                   i, j, oaControl, oaAutoControl, use1394B;
  int                   found, numResolutions, preferredFound;

  dc1394framerates_t framerates;
  dc1394framerate_t framerate;
  dc1394video_mode_t video_mode = 0;
  dc1394color_coding_t coding;
  dc1394video_frame_t* frame;

  iidcContext = dc1394_new();
  if ( !iidcContext ) {
    fprintf ( stderr, "%s: Can't get IIDC context\n", __FUNCTION__ );
    return 0;
  }

  if (!( iidcCam = dc1394_camera_new_unit ( iidcContext, device->_guid,
      device->_unit ))) {
    fprintf ( stderr, "%s: dc1394_camera_new_unit failed\n", __FUNCTION__ );
    return 0;
  }

  if ( dc1394_feature_get_all ( iidcCam, &features ) != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_feature_get_all failed\n", __FUNCTION__ );
    return 0;
  }

  if (!( camera = ( oaCamera* ) malloc ( sizeof ( oaCamera )))) {
    fprintf ( stderr, "%s: malloc oaCamera failed\n", __FUNCTION__ );
    // FIX ME -- set errno?
    return 0;
  }
  OA_CLEAR ( cameraState );
  OA_CLEAR ( camera->controls );
  OA_CLEAR ( camera->features );

  camera->interface = device->interface;
  ( void ) strcpy ( camera->deviceName, device->deviceName );
  cameraState.initialised = 0;

  // Reset the camera here.  That way hopefully we can be sure that the
  // values we read from the current settings are actually the default
  // values for those controls

  if ( dc1394_camera_reset ( iidcCam ) != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_camera_reset failed\n", __FUNCTION__ );
  }

  _oaInitCameraFunctionPointers ( camera );
  _IIDCInitFunctionPointers ( camera );

  // There's a lot of work still to be done here.  For most stuff I'm
  // ignoring absolute_capable because it uses completely different
  // ranges and units and everything becomes a nightmare :(
  // I'm also (mostly) ignoring on/off, readout and polarity

  for ( i = 0; i < DC1394_FEATURE_NUM; i++ ) {
    if ( features.feature[ i ].available &&
        (( features.feature[ i ].on_off_capable &&
        features.feature[ i ].is_on ) ||
        !features.feature[ i ].on_off_capable )) {

      oaControl = dc1394Controls[ i ].oaControl;
      oaAutoControl = dc1394Controls[ i ].oaAutoControl;

      switch ( i + DC1394_FEATURE_MIN ) {

        case DC1394_FEATURE_BRIGHTNESS:
        case DC1394_FEATURE_SHARPNESS:
        case DC1394_FEATURE_WHITE_BALANCE:
        case DC1394_FEATURE_HUE:
        case DC1394_FEATURE_SATURATION:
        case DC1394_FEATURE_GAMMA:
        case DC1394_FEATURE_GAIN:
        case DC1394_FEATURE_EXPOSURE:
        {
          int j;

          for ( j = 0; j < features.feature[i].modes.num; j++ ) {
            switch ( features.feature[i].modes.modes[j] ) {

              case DC1394_FEATURE_MODE_MANUAL:
                camera->controls[ oaControl ] = OA_CAM_CTRL_TYPE_INT32;
                camera->min[ oaControl ] = features.feature[i].min;
                camera->max[ oaControl ] = features.feature[i].max;
                camera->step[ oaControl ] = 1; // arbitrary
                camera->def[ oaControl ] = features.feature[i].value;
                break;

              case DC1394_FEATURE_MODE_AUTO:
                if ( oaAutoControl ) {
                  camera->controls[ oaAutoControl ] = OA_CAM_CTRL_TYPE_BOOLEAN;
                  camera->min[ oaAutoControl ] = 0;
                  camera->max[ oaAutoControl ] = 1;
                  camera->step[ oaAutoControl ] = 1;
                  camera->def[ oaAutoControl ] = ( DC1394_FEATURE_MODE_AUTO ==
                      features.feature[i].current_mode ) ? 1 : 0;
                } else {
                  fprintf ( stderr, "%s: have auto for control %d, but "
                      "liboacam does not", __FUNCTION__, oaControl );
                }
                break;

              default:
                break;
            }
          }
          break;
        }
        case DC1394_FEATURE_SHUTTER:
        {
          unsigned int min, max, step, def;
          // shutter is actually exposure time.  exposure is something
          // else
          oaAutoControl = OA_CAM_CTRL_AUTO_EXPOSURE;
          if ( features.feature[ i ].absolute_capable ) {
            // have to set the control to use the absolute settings...
            if ( dc1394_feature_set_absolute_control ( iidcCam,
                DC1394_FEATURE_SHUTTER, DC1394_ON ) != DC1394_SUCCESS ) {
              fprintf ( stderr, "%s: dc1394_feature_set_absolute_control "
                  "failed\n", __FUNCTION__ );
              free ( camera );
              return 0;
            }
            // and disable auto mode, as we don't support that at the moment
            // FIX ME -- need to deal with this when we do -- auto mode in
            // IIDC cameras may be more than just on/off
            if ( dc1394_feature_set_mode ( iidcCam, DC1394_FEATURE_SHUTTER,
                DC1394_FEATURE_MODE_MANUAL ) != DC1394_SUCCESS ) {
              fprintf ( stderr, "%s: dc1394_feature_set_mode failed for "
                  "shutter speed\n", __FUNCTION__ );
              free ( camera );
              return 0;
            }

            oaControl = OA_CAM_CTRL_EXPOSURE_ABSOLUTE;
            min = features.feature[ i ].abs_min * 1000000.0;
            max = features.feature[ i ].abs_max * 1000000.0;
            step = 1000; // arbitrary
            def = features.feature[ i ].abs_value * 1000000.0;
          } else {
            oaControl = OA_CAM_CTRL_EXPOSURE;
            min = features.feature[i].min;
            max = features.feature[i].max;
            step = 1; // arbitrary
            def = features.feature[ i ].value;
          }
          camera->controls[ oaControl ] = OA_CAM_CTRL_TYPE_INT32;
          camera->min[ oaControl ] = min;
          camera->max[ oaControl ] = max;
          camera->step[ oaControl ] = step;
          camera->def[ oaControl ] = def;
          break;
        }
        case DC1394_FEATURE_FRAME_RATE:
          // this is used, but handled elsewhere
          break;

        case DC1394_FEATURE_IRIS:
        case DC1394_FEATURE_FOCUS:
        case DC1394_FEATURE_TEMPERATURE:
        case DC1394_FEATURE_TRIGGER:
        case DC1394_FEATURE_TRIGGER_DELAY:
        case DC1394_FEATURE_WHITE_SHADING:
        case DC1394_FEATURE_ZOOM:
        case DC1394_FEATURE_PAN:
        case DC1394_FEATURE_TILT:
        case DC1394_FEATURE_OPTICAL_FILTER:
        case DC1394_FEATURE_CAPTURE_SIZE:
        case DC1394_FEATURE_CAPTURE_QUALITY:
          fprintf ( stderr, "%s: unsupported IIDC control %d\n", __FUNCTION__,
              i );
          break;

        default:
          fprintf ( stderr, "%s: unknown IIDC control %d\n", __FUNCTION__, i );
          break;
      }
    }
  }

  // Looking at the specs it appears that a number of compressed video
  // modes are supported (YUV422 etc.) as well as RGB and 8-bit or 16-bit
  // mono should
  // be supported.  For the moment we'll ignore YUY2 and assume the camera
  // is colour if RGB shows up, monochrome otherwise.

  if ( dc1394_video_get_supported_modes ( iidcCam, &videoModes ) !=
      DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_get_supported_modes failed",
        __FUNCTION__ );
    free (( void* ) camera );
    return 0;
  }

  // There are problems here if not all colour modes are supported in
  // all resolutions

  cameraState.videoRGB24 = 0;
  cameraState.videoGrey16 = 0;
  cameraState.videoGrey = 0;
  cameraState.mosaicFormat = 0;
  cameraState.videoCurrent = 0;
  cameraState.currentCodec = 0;

  camera->features.rawMode = camera->features.demosaicMode = 0;

  numResolutions = preferredFound = 0;
  for ( i = 0; i < videoModes.num && i < OA_MAX_RESOLUTIONS; i++ ) {
    uint32_t w, h;

    if ( videoModes.modes[i] >= DC1394_VIDEO_MODE_FORMAT7_MIN &&
        videoModes.modes[i] <= DC1394_VIDEO_MODE_FORMAT7_MAX ) {
      continue;
    }

    if ( dc1394_get_image_size_from_video_mode ( iidcCam, videoModes.modes[i],
        &w, &h ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_get_image_size_from_video_mode failed",
          __FUNCTION__ );
      free (( void* ) camera );
      return 0;
    }
    found = 0;
    if ( numResolutions ) {
      for ( j = 0; j < numResolutions && !found; j++ ) {
        if ( cameraState.resolutions[j].x == w &&
            cameraState.resolutions[j].y == h ) {
          found = 1;
        }
      }
    }
    if ( !found ) {
      cameraState.resolutions[ numResolutions ].x = w;
      cameraState.resolutions[ numResolutions ].y = h;
      numResolutions++;
    }

    if ( dc1394_get_color_coding_from_video_mode ( iidcCam,
        videoModes.modes[i], &codec ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_get_color_coding_from_video_mode failed",
          __FUNCTION__ );
      free (( void* ) camera );
      return 0;
    }

    switch ( codec ) {
      case DC1394_COLOR_CODING_MONO8:
        if ( !cameraState.videoRGB24 ) {
          preferredFound = 1;
          cameraState.videoCurrent = videoModes.modes[i];
          cameraState.currentCodec = codec;
        }
        cameraState.videoGrey = 1;
        break;

      case DC1394_COLOR_CODING_RGB8:
        preferredFound = 1;
        cameraState.videoCurrent = videoModes.modes[i];
        cameraState.currentCodec = codec;
        cameraState.videoRGB24 = 1;
        camera->features.demosaicMode = 1;
        break;

      case DC1394_COLOR_CODING_MONO16:
        if ( !preferredFound ) {
          preferredFound = 1;
          cameraState.videoCurrent = videoModes.modes[i];
          cameraState.currentCodec = codec;
        }
        cameraState.videoGrey16 = 1;
        break;

      case DC1394_COLOR_CODING_RAW8:
        if ( !preferredFound ) {
          preferredFound = 1;
          cameraState.videoCurrent = videoModes.modes[i];
          cameraState.currentCodec = codec;
        }
        camera->features.rawMode = 1;
        break;

      default:
        fprintf ( stderr, "%s: unhandled video codec %d\n", __FUNCTION__,
            codec );
        break;
    }
  }

  if ( !numResolutions ) {
    fprintf ( stderr, "%s: no suitable resolutions found\n", __FUNCTION__ );
    free (( void* ) camera );
    return 0;
  }

  cameraState.resolutions[ numResolutions ].x = 0;
  cameraState.resolutions[ numResolutions ].y = 0;

  if ( !cameraState.videoCurrent ) {
    fprintf ( stderr, "%s: No suitable video format found", __FUNCTION__ );
    free (( void* ) camera );
    return 0;
  }

  if ( dc1394_video_get_supported_framerates ( iidcCam,
      cameraState.videoCurrent, &framerates ) != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_get_supported_framerates failed\n",
       __FUNCTION__ );
  }

  if ( framerates.num > 1 ) {
    camera->features.frameRates = 1;
  }

  use1394B = 1;
  if ( dc1394_video_set_operation_mode ( iidcCam, DC1394_OPERATION_MODE_1394B )
      != DC1394_SUCCESS ) {
    use1394B = 0;
    fprintf ( stderr, "Setting 1394B mode failed, falling back to 1394\n" );
    if ( dc1394_video_set_operation_mode ( iidcCam,
        DC1394_OPERATION_MODE_LEGACY ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_video_set_operation_mode failed\n",
          __FUNCTION__ );
      return 0;
    }
  }

  if ( dc1394_video_set_iso_speed ( iidcCam, use1394B ? DC1394_ISO_SPEED_400 :
      DC1394_ISO_SPEED_800 ) != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_set_iso_speed failed\n",
       __FUNCTION__ );
  }

/*
  if ( dc1394_video_set_mode ( iidcCam, cameraState.videoCurrent ) !=
      DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_set_mode failed\n", __FUNCTION__ );
    free (( void* ) camera );
    return 0;
  }

  if ( dc1394_capture_setup ( iidcCam, NUM_IIDC_BUFFERS,
      DC1394_CAPTURE_FLAGS_DEFAULT ) != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_capture_setup failed\n", __FUNCTION__ );
    free (( void* ) camera );
    return 0;
  }
*/

  cameraState.iidcHandle = iidcCam;
  cameraState.initialised = 1;

  return camera;
}


static void
_IIDCInitFunctionPointers ( oaCamera* camera )
{
  camera->funcs.initCamera = oaIIDCInitCamera;
  camera->funcs.closeCamera = oaIIDCCloseCamera;
  camera->funcs.resetCamera = oaIIDCCameraReset;
  // camera->funcs.startCapture = oaIIDCCameraStart;
  camera->funcs.startCaptureExtended = oaIIDCCameraStart;
  camera->funcs.stopCapture = oaIIDCCameraStop;
  camera->funcs.hasLoadableFirmware = oaIIDCCameraHasLoadableFirmware;
  // camera->funcs.isFirmwareLoaded = _cameraIsFirmwareLoaded;
  // camera->funcs.loadFirmware = _cameraLoadFirmware;
  camera->funcs.getName = oaIIDCCameraGetName;
  camera->funcs.getControlRange = oaIIDCCameraGetControlRange;
  camera->funcs.readControl = oaIIDCCameraReadControl;
  camera->funcs.setControl = oaIIDCCameraSetControl;
  camera->funcs.testControl = oaIIDCCameraTestControl;
  // camera->funcs.hasAuto = oaHasAuto;
  // camera->funcs.isAuto = _isAuto;
  // camera->funcs.setControlMulti = _setControlMulti;
  camera->funcs.startReadFrame = oaIIDCCameraStartReadFrame;
  camera->funcs.readFrame = oaIIDCCameraReadFrame;
  camera->funcs.finishReadFrame = oaIIDCCameraFinishReadFrame;
  camera->funcs.enumerateFrameSizes = oaIIDCCameraGetFrameSizes;
  camera->funcs.enumerateFrameRates = oaIIDCCameraGetFrameRates;
  camera->funcs.getFramePixelFormat = oaIIDCCameraGetFramePixelFormat;
  // camera->funcs.testROISize = _testROISize;
  // camera->funcs.getMenuString = oaIIDCCameraGetMenuString;
  // camera->funcs.getAutoWBManualSetting = oaIIDCCameraGetAutoWBManualSetting;
  camera->funcs.hasFixedFrameRates = oaIIDCCameraHasFixedFrameRates;
}

#endif	/* HAVE_LIBDC1394 */
