/*****************************************************************************
 *
 * QHY5.c -- QHY5 camera interface
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>
#include <openastro/oacam.h>
#include <openastro/errno.h>

#include <libusb-1.0/libusb.h>

#include "QHYoacam.h"
#include "QHY5.h"
#include "oacamprivate.h"
#include "QHYusb.h"

#define QHY5_SENSOR_WIDTH	1558
#define QHY5_SENSOR_HEIGHT	1048
#define QHY5_IMAGE_WIDTH	1280
#define QHY5_IMAGE_HEIGHT	1024
#define	QHY5_BUFFERS		5
#define VBLANK			26
#define DARK_WIDTH_X		36
#define PIXEL_RATE		24000

#define BUFFER_SIZE		19
#define USB_TIMEOUT		5000
#define DEFAULT_EXPOSURE	100


static int      _setParameters ( oaCamera*, int, int, int );
static void     _startExposure ( oaCamera*, unsigned int );
static int      _readExposure ( oaCamera*, void*, unsigned int );
static void*    _capture ( void* );
static void     _init ( oaCamera* );
static void	_QHY5InitFunctionPointers ( oaCamera* );
static void	_storeWordBE ( unsigned char*, unsigned int );


static int      setControl ( oaCamera*, int, int );
static int      getFramePixelFormat ( oaCamera*, int );
static const FRAMESIZE* getFrameSizes ( oaCamera* );
static const char* getName ( oaCamera* );
static int      reset ( oaCamera* );
static int      stop ( oaCamera* );
static int      start ( oaCamera*, START_PARMS* );
static int      getControlRange ( oaCamera*, int, int*, int*, int*, int* );
static int      startReadFrame ( oaCamera*, unsigned int );
static int      finishReadFrame ( oaCamera* );
static int      readFrame ( oaCamera*, void** );
static int      closeCamera ( oaCamera* );

/**
 * Initialise a given camera device
 */

int
_QHY5InitCamera ( oaCamera* camera )
{
  int i, j;

  OA_CLEAR ( camera->controls );
  OA_CLEAR ( camera->features );
  _QHY5InitFunctionPointers ( camera );

  camera->controls[ OA_CAM_CTRL_GAIN ] = OA_CAM_CTRL_TYPE_INT32;
  camera->min[ OA_CAM_CTRL_GAIN ] = 1;
  camera->max[ OA_CAM_CTRL_GAIN ] = 74;
  camera->step[ OA_CAM_CTRL_GAIN ] = 1;
  camera->def[ OA_CAM_CTRL_GAIN ] = 10; // completely arbitrary

  camera->controls[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = OA_CAM_CTRL_TYPE_INT32;
  camera->min[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 0;
  camera->max[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 10000000; // I made this up
  camera->step[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 1;
  camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = DEFAULT_EXPOSURE * 1000;

  camera->controls[ OA_CAM_CTRL_DROPPED ] = OA_CAM_CTRL_TYPE_READONLY;
  camera->controls[ OA_CAM_CTRL_DROPPED_RESET ] = OA_CAM_CTRL_TYPE_BUTTON;

  cameraState.maxResolutionX = QHY5_IMAGE_WIDTH;
  cameraState.maxResolutionY = QHY5_IMAGE_HEIGHT;

  cameraState.videoRGB24 = cameraState.videoGrey16 = 0;
  cameraState.videoGrey = 1;
  cameraState.videoCurrent = OA_PIX_FMT_GREY8;

  cameraState.resolutions[0].x = cameraState.maxResolutionX;
  cameraState.resolutions[0].y = cameraState.maxResolutionY;
  cameraState.resolutions[1].x = 0;
  cameraState.resolutions[1].y = 0;

  cameraState.xSize = cameraState.maxResolutionX;
  cameraState.ySize = cameraState.maxResolutionY;
  cameraState.xOffset = DARK_WIDTH_X;

  cameraState.buffers = 0;
  cameraState.configuredBuffers = 0;

  camera->features.ROI = 1;

  cameraState.captureLength = QHY5_SENSOR_WIDTH *
      ( QHY5_IMAGE_HEIGHT + VBLANK );
  if (!( cameraState.xferBuffer = malloc ( cameraState.captureLength ))) {
    fprintf ( stderr, "malloc of transfer buffer failed in %s\n",
        __FUNCTION__ );
    return 0;
  }

  cameraState.currentGain = camera->def[ OA_CAM_CTRL_GAIN ];
  cameraState.currentExposure = DEFAULT_EXPOSURE;

  // _init ( camera );

  cameraState.imageBufferLength = cameraState.maxResolutionX *
      cameraState.maxResolutionY;
  cameraState.buffers = calloc ( QHY5_BUFFERS, sizeof ( struct QHYbuffer ));
  for ( i = 0; i < QHY5_BUFFERS; i++ ) {
    void* m = malloc ( cameraState.imageBufferLength );
    if ( m ) {
      cameraState.buffers[i].start = m;
      cameraState.configuredBuffers++;
    } else {
      fprintf ( stderr, "%s malloc failed\n", __FUNCTION__ );
      if ( i ) {
        for ( j = 0; j < i; j++ ) {
          free (( void* ) cameraState.buffers[j].start );
        }
      }
      free ( cameraState.xferBuffer );
      return 0;
    }
  }

  return -1;
}


static int
_setParameters ( oaCamera* camera, int height, int gain, int firstCall )
{
  // Weirdness starts here.
  //
  // First, gain settings.  Gain may be set for individual colour planes
  // (not applicable for the mono QHY5, obviously) and globally, but the
  // settings aren't linear.  Everyone seems to go for a lookup table for
  // this, so I guess I'll follow suit.

  static const int gainLookup[] = {
      0x000,0x004,0x005,0x006,0x007,0x008,0x009,0x00A,0x00B,0x00C,0x00D,0x00E,
      0x00F,0x010,0x011,0x012,0x013,0x014,0x015,0x016,0x017,0x018,0x019,0x01A,
      0x01B,0x01C,0x01D,0x01E,0x01F,0x051,0x052,0x053,0x054,0x055,0x056,0x057,
      0x058,0x059,0x05A,0x05B,0x05C,0x05D,0x05E,0x05F,0x6CE,0x6CF,0x6D0,0x6D1,
      0x6D2,0x6D3,0x6D4,0x6D5,0x6D6,0x6D7,0x6D8,0x6D9,0x6DA,0x6DB,0x6DC,0x6DD,
      0x6DE,0x6DF,0x6E0,0x6E1,0x6E2,0x6E3,0x6E4,0x6E5,0x6E6,0x6E7,0x6FC,0x6FD,
      0x6FE,0x6FF
  };
  static const int gainTableSize = ( sizeof ( gainLookup ) / sizeof ( int ));

  unsigned int offset, frameSize, frameSizeMSW, frameSizeLSW;
  unsigned int greenGain, blueGain, redGain, globalGain;
  unsigned char regs[ BUFFER_SIZE ];

  if ( gain > gainTableSize ) {
    gain = gainTableSize;
  }
  greenGain = blueGain = redGain = globalGain = gainLookup[ gain - 1 ];

  // height must be a multiple of 4
  height -= height % 4;
  // how far from the top edge of the sensor we start
  offset = ( QHY5_SENSOR_HEIGHT - height ) / 2;
  frameSize = ( QHY5_SENSOR_WIDTH * ( height + VBLANK ));
  frameSizeMSW = frameSize >> 16;
  frameSizeLSW = frameSize & 0xffff;

  // The registers we appear to have access to are:
  // 0x01 - first row to read
  // 0x02 - first column to read
  // 0x03 - how many rows to read (less one)
  // 0x04 - how many columns to read (less one)
  // 0x09 - how many rows to integrate (including vblank?)
  // 0x2b - green gain
  // 0x2c - blue gain
  // 0x2d - red gain
  // 0x2e - global gain

  _storeWordBE ( regs, greenGain );        // MT9M001 register 0x2b
  _storeWordBE ( regs + 2, blueGain );     // MT9M001 register 0x2c
  _storeWordBE ( regs + 4, redGain );      // MT9M001 register 0x2d
  _storeWordBE ( regs + 6, globalGain );   // MT9M001 register 0x2e
  _storeWordBE ( regs + 8, offset );       // MT9M001 register 0x01
  _storeWordBE ( regs + 10, 0 );           // MT9M001 register 0x02
  _storeWordBE ( regs + 12, height - 1 );  // MT9M001 register 0x03
  _storeWordBE ( regs + 14, 1313 );        // MT9M001 register 0x04 (why 1314?)
  _storeWordBE ( regs + 16, height + VBLANK - 1 ); // MT9M001 register 0x09
  regs[18] = 0xcc; // Magic number!

  // Lots of magic numbers :(
  // FIX ME -- check for errors here?

  _usbControlMsg ( camera, 0x42, 0x13, frameSizeLSW, frameSizeMSW, regs,
      BUFFER_SIZE, USB_TIMEOUT );
  _usbControlMsg ( camera, 0x42, 0x14, 0x31a5, 0, 0, 0, USB_TIMEOUT );
  _usbControlMsg ( camera, 0x42, 0x16, firstCall ? 1 : 0, 0, 0, 0,
      USB_TIMEOUT );

  return ( height + VBLANK );
}


static void
_startExposure ( oaCamera* camera, unsigned int exposureTime )
{
  unsigned char buff[8] = "DEADBEEF";
  // FIX ME -- more magic numbers
  _usbControlMsg ( camera, 0xc2, 0x12, exposureTime & 0xffff,
      exposureTime >> 16, buff, 2, USB_TIMEOUT );
}


static int
_readExposure ( oaCamera* camera, void* buffer, unsigned int readLength )
{
  int ret;
  int readSize;

  // FIX ME -- references to droppedFrames here should probably be mutexed

  ret = _usbBulkTransfer ( camera, 0x82, buffer, readLength, &readSize,
      USB_TIMEOUT );
  if ( ret ) {
    cameraState.droppedFrames++;
    return ret;
  }
  if ( readSize != readLength ) {
    fprintf ( stderr, "readExposure: USB bulk transfer was short. %d != %d\n",
        readSize, readLength );
    cameraState.droppedFrames++;
    return -1;
  }
  return 0;
}


static int
setControl ( oaCamera* camera, int control, int value )
{
  switch ( control ) {

    case OA_CAM_CTRL_GAIN:
      cameraState.currentGain = value;
      if ( cameraState.captureThreadStarted ) {
        pthread_mutex_lock ( &cameraState.captureMutex );
        cameraState.doSetParameters = 1;
        pthread_mutex_unlock ( &cameraState.captureMutex );
      } else {
        _setParameters ( camera, cameraState.ySize,
            cameraState.currentGain, 0 );
      }
      break;

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      value /= 1000;
      if ( value < 1 ) { value = 1; }
      pthread_mutex_lock ( &cameraState.captureMutex );
      cameraState.transferTime = QHY5_SENSOR_WIDTH * ( cameraState.ySize +
          VBLANK ) / PIXEL_RATE;
      if ( cameraState.transferTime > value ) {
        value = cameraState.transferTime;
      }
      cameraState.currentExposure = value;
      cameraState.requestedExposure = value - cameraState.transferTime;
      pthread_mutex_unlock ( &cameraState.captureMutex );
      break;

    case OA_CAM_CTRL_DROPPED_RESET:
      // FIX ME -- droppedFrames should probably be mutexed
      cameraState.droppedFrames = 0;
      break;

    default:
      fprintf ( stderr, "QHY5: %s not yet implemented for control %d\n",
          __FUNCTION__, control );
      return -1;
      break;
  }

  return 0;
}


static int
readControl ( oaCamera* camera, int control )
{
  switch ( control ) {

    case OA_CAM_CTRL_GAIN:
      return cameraState.currentGain;
      break;

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      return cameraState.currentExposure * 1000;
      break;

    case OA_CAM_CTRL_DROPPED:
      return cameraState.droppedFrames;
      break;

    default:
      fprintf ( stderr,
          "Unrecognised control %d in QHY5:%s\n", control, __FUNCTION__ );
      return 0;
      break;
  }
  return -1;
}


static int
testControl ( oaCamera* camera, int control, int value )
{
  switch ( control ) {

    case OA_CAM_CTRL_GAIN:
    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      if ( value >= camera->min[ control ] &&
          value <= camera->max[ control ] &&
          ( 0 == ( value - camera->min[ control ]) %
          camera->step[ control ])) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CAM_CTRL_BINNING:
    case OA_CAM_CTRL_BIT_DEPTH:
    case OA_CAM_CTRL_COLOUR_MODE:
      return -OA_ERR_INVALID_CONTROL;
      break;

    default:
      fprintf ( stderr, "QHY5: %s not yet implemented for control %d\n",
          __FUNCTION__, control );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }

  // And if we reach here it's because the value wasn't valid
  return -OA_ERR_OUT_OF_RANGE;
}


static int
getFramePixelFormat ( oaCamera* camera, int depth )
{
  if ( depth > 0 && depth != 8 ) {
    return -OA_ERR_INVALID_BIT_DEPTH;
  }
  return OA_PIX_FMT_GREY8;
}


const static FRAMESIZE*
getFrameSizes ( oaCamera* camera )
{
  return cameraState.resolutions;
}


static const char*
getName ( oaCamera* camera )
{
  return "QHY5";
}


static int
reset ( oaCamera* camera )
{
  unsigned char data = 0;
  int xferred;

  stop ( camera );
  _usbBulkTransfer ( camera, 1, &data, 1, &xferred, USB_TIMEOUT );
  // _init ( camera );
  return 0;
}


static int
stop ( oaCamera* camera )
{
  if ( !cameraState.captureThreadStarted ) {
    return 0;
  }

  pthread_mutex_lock ( &cameraState.captureMutex );
  cameraState.captureThreadExit = 1;
  pthread_mutex_unlock ( &cameraState.captureMutex );
  pthread_join ( cameraState.captureThread, 0 );
  cameraState.captureThreadStarted = 0;
  return 0;
}


static int
start ( oaCamera* camera, START_PARMS* params )
{
  cameraState.captureThreadStarted = 0;
  cameraState.captureThreadExit = 0;
  cameraState.xSize = params->size.x;
  cameraState.ySize = params->size.y;
  cameraState.xOffset = ( QHY5_IMAGE_WIDTH - DARK_WIDTH_X -
      cameraState.xSize ) / 2 + DARK_WIDTH_X;
  cameraState.captureHeight = _setParameters ( camera, cameraState.ySize,
      cameraState.currentGain, 0 );
  cameraState.imageBufferLength = cameraState.xSize * cameraState.ySize;
  _init ( camera );
  pthread_create ( &cameraState.captureThread, 0, &_capture, camera );
  cameraState.captureThreadStarted = 1;
  return 0;
}


static void*
_capture ( void* param )
{
  oaCamera*     camera = param;
  int           quitThread, nextBuffer, x, y, doSetParameters;
  unsigned int  requestedTime;
  uint8_t*      s;
  uint8_t*      t;
  uint8_t*      startOfFirstLine;
  uint8_t*      startOfLine;

  pthread_mutex_lock ( &cameraState.captureMutex );
  cameraState.haveDataToRead = 0;
  cameraState.nextBufferToRead = -1;
  cameraState.doSetParameters = 0;
  pthread_mutex_unlock ( &cameraState.captureMutex );
  nextBuffer = 0;

  startOfFirstLine = cameraState.xferBuffer + cameraState.xOffset;
  while ( 1 ) {
    pthread_mutex_lock ( &cameraState.captureMutex );
    quitThread = cameraState.captureThreadExit;
    if ( nextBuffer == cameraState.nextBufferToRead ) {
      nextBuffer = ( nextBuffer + 1 ) % cameraState.configuredBuffers;
    }
    requestedTime = cameraState.requestedExposure;
    pthread_mutex_unlock ( &cameraState.captureMutex );
    if ( quitThread ) {
      break;
    }

    // usleep ( 20000 );
    _startExposure ( camera, requestedTime );
    // usleep ( requestedTime < 20000 ? 20000 : requestedTime );
    usleep ( requestedTime );
    _readExposure ( camera, cameraState.xferBuffer,
        cameraState.captureLength );

    t = cameraState.buffers[ nextBuffer ].start;
    startOfLine = startOfFirstLine;
    for ( y = 0; y < cameraState.ySize; y++ ) {
      s = startOfLine;
      for ( x = 0; x < cameraState.xSize; x++ ) {
        *t++ = *s++;
      }
      startOfLine += QHY5_SENSOR_WIDTH;
    }

    pthread_mutex_lock ( &cameraState.captureMutex );
    cameraState.nextBufferToRead = nextBuffer;
    cameraState.haveDataToRead = 1;
    doSetParameters = cameraState.doSetParameters;
    cameraState.doSetParameters = 0;
    pthread_cond_signal ( &cameraState.frameAvailable );
    pthread_mutex_unlock ( &cameraState.captureMutex );
    nextBuffer = ( nextBuffer + 1 ) % cameraState.configuredBuffers;

    if ( doSetParameters ) {
      _setParameters ( camera, cameraState.ySize, cameraState.currentGain, 0 );
    }
  }
  return 0;
}


static void
_init ( oaCamera* camera )
{
  static int firstCall = 1;

  cameraState.captureHeight = _setParameters ( camera, cameraState.ySize,
      8, firstCall );
  firstCall = 0;

  cameraState.captureLength = cameraState.captureHeight * QHY5_SENSOR_WIDTH;

  // Now we need to capture a single frame for some reason.

  // FIX ME -- check for errors here?
  _startExposure ( camera, cameraState.currentExposure );
  usleep ( cameraState.currentExposure );
  _readExposure ( camera, cameraState.xferBuffer,
      cameraState.captureLength );
  cameraState.captureHeight = _setParameters ( camera, cameraState.ySize,
      cameraState.currentGain, 0 );
}


static int
getControlRange ( oaCamera* camera, int control, int* min, int* max,
    int* step, int* def )
{
  if ( !camera->controls[ control ] ) {
    return -OA_ERR_INVALID_CONTROL;
  }

  *min = camera->min[ control ];
  *max = camera->max[ control ];
  *step = camera->step[ control ];
  *def = camera->def[ control ];
  return 0;
}


static int
startReadFrame ( oaCamera* camera, unsigned int frameTime )
{
  cameraState.frameTime = frameTime;
  return 0;
}


static int
readFrame ( oaCamera* camera, void** buffer )
{
  int haveData = 0;
  int bufferToRead;
  struct timespec waitUntil;
  struct timeval now;
  unsigned long frameSec = 0;
  unsigned long frameNanoSec = 0;

  frameNanoSec = cameraState.frameTime * 1000000;
  frameSec = cameraState.frameTime / 1000 + 5;

  pthread_mutex_lock ( &cameraState.captureMutex );
  while ( !cameraState.haveDataToRead ) {

    gettimeofday ( &now, 0 );
    waitUntil.tv_sec = now.tv_sec + frameSec;
    if (( now.tv_usec * 1000 + frameNanoSec ) > 1000000000 ) {
      waitUntil.tv_sec++;
    }
    waitUntil.tv_nsec = ( now.tv_usec * 1000 + frameNanoSec ) % 1000000000;

    pthread_cond_timedwait ( &cameraState.frameAvailable,
        &cameraState.captureMutex, &waitUntil );
  }
  haveData = cameraState.haveDataToRead;
  bufferToRead = cameraState.nextBufferToRead;
  cameraState.haveDataToRead = 0;
  pthread_mutex_unlock ( &cameraState.captureMutex );

  *buffer = haveData ? cameraState.buffers[ bufferToRead ].start : 0;
  return ( haveData ? cameraState.imageBufferLength : 0 );
}


static int
finishReadFrame ( oaCamera* camera )
{
  return 0;
}


static int
closeCamera ( oaCamera* camera )
{
  libusb_release_interface ( cameraState.usbHandle, 0 );
  libusb_close ( cameraState.usbHandle );
  libusb_exit ( cameraState.usbContext );
  return 0;
}


int
testROISize ( oaCamera* camera, int tryX, int tryY, int* suggX, int* suggY )
{
  if ( 0 == tryX % 2 && 0 == tryY % 4 ) {
    if ( tryX <= cameraState.resolutions[0].x &&
        tryY <= cameraState.resolutions[0].y ) {
      return OA_ERR_NONE;
    }
  }

  tryX -= tryX % 2;
  tryY -= tryY % 4;
  if ( tryX == 0 ) {
    tryX = 2;
  }
  if ( tryY == 0 ) {
    tryY = 4;
  }

  *suggX = tryX;
  *suggY = tryY;

  return -OA_ERR_INVALID_SIZE;
}


static void
_storeWordBE ( unsigned char* p, unsigned int val )
{
  *p++ = ( val >> 8 ) & 0xff;
  *p = val & 0xff;
}


static void
_QHY5InitFunctionPointers ( oaCamera* camera )
{
  // Set by QHYinit()
  // camera->funcs.initCamera = oaQHYInitCamera;
  camera->funcs.closeCamera = closeCamera;
  camera->funcs.resetCamera = reset;
  // camera->funcs.startCapture = start;
  camera->funcs.startCaptureExtended = start;
  camera->funcs.stopCapture = stop;
  // stick with the device versions of these
  // camera->funcs.hasLoadableFirmware = _cameraHasLoadableFirmware;
  // camera->funcs.isFirmwareLoaded = _cameraIsFirmwareLoaded;
  // camera->funcs.loadFirmware = _cameraLoadFirmware;
  camera->funcs.getName = getName;
  camera->funcs.getControlRange = getControlRange;
  camera->funcs.readControl = readControl;
  camera->funcs.setControl = setControl;
  camera->funcs.testControl = testControl;
  // camera->funcs.hasAuto = oaHasAuto;
  // camera->funcs.isAuto = _isAuto;
  // camera->funcs.setControlMulti = _setControlMulti;
  camera->funcs.startReadFrame = startReadFrame;
  camera->funcs.readFrame = readFrame;
  camera->funcs.finishReadFrame = finishReadFrame;
  camera->funcs.enumerateFrameSizes = getFrameSizes;
  camera->funcs.getFramePixelFormat = getFramePixelFormat;
  camera->funcs.testROISize = testROISize;
  // camera->funcs.getMenuString = _getMenuString;
  // camera->funcs.getAutoWBManualSetting = _getAutoWBManualSetting;
  // camera->funcs.hasFixedFrameRates = _hasFixedFrameRates;
}

