/*****************************************************************************
 *
 * QHY5LII.c -- QHY5LII camera interface
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>
#include <openastro/oacam.h>
#include <openastro/errno.h>

#if HAVE_MATH_H
#include <math.h>
#endif

#include <libusb-1.0/libusb.h>

#include "QHYoacam.h"
#include "QHY5LII.h"
#include "oacamprivate.h"
#include "QHYusb.h"

#define QHY5LII_IMAGE_WIDTH	1280
#define QHY5LII_IMAGE_HEIGHT	960
#define DEFAULT_EXPOSURE	100
#define DEFAULT_SPEED		1
#define DEFAULT_USBTRAFFIC	30
#define USB_TIMEOUT		5000
#define	QHY5LII_BUFFERS		5
#define QHY5LII_IMAGE_OFFSET	5
#define PIXEL_RATE              24000

#define MONO_GAIN_MIN		0
#define MONO_GAIN_MAX		796
#define COLOUR_GAIN_MIN         0
#define COLOUR_GAIN_MAX         398

#define	MT9M034_Y_ADDR_START			0x3002
#define	MT9M034_X_ADDR_START			0x3004
#define	MT9M034_Y_ADDR_END			0x3006
#define	MT9M034_X_ADDR_END			0x3008
#define	MT9M034_FRAME_LENGTH_LINES		0x300A
#define	MT9M034_LINE_LENGTH_PCK			0x300C
#define MT9M034_COARSE_INTEGRATION_TIME		0x3012
#define MT9M034_FINE_INT_TIME			0x3014
#define	MT9M034_COARSE_INT_TIME_CB		0x3016
#define MT9M034_FINE_INT_TIME_CB		0x3018
#define MT9M034_RESET_REGISTER			0x301A
#define MT9M034_DATA_PEDESTAL			0x301E
#define MT9M034_VT_SYS_CLK_DIV			0x302A
#define MT9M034_VT_PIX_CLK_DIV			0x302C
#define MT9M034_PRE_PLL_CLK_DIV			0x302E
#define MT9M034_PLL_MULTIPLIER			0x3030
#define MT9M034_DIGITAL_BINNING			0x3032
#define	MT9M034_READ_MODE			0x3040
#define MT9M034_DARK_CONTROL			0x3044
#define MT9M034_GREEN1_GAIN			0x3056
#define MT9M034_BLUE_GAIN			0x3058
#define MT9M034_RED_GAIN			0x305A
#define MT9M034_GREEN2_GAIN			0x305C
#define MT9M034_GLOBAL_GAIN			0x305E
#define MT9M034_EMBEDDED_DATA_CTRL		0x3064
#define MT9M034_TEST_PATTERN			0x3070
#define MT9M034_TEST_RAW_MODE			0x307A
#define MT9M034_MODE_CTRL			0x3082
#define MT9M034_SEQ_DATA_PORT			0x3086
#define MT9M034_SEQ_CTRL_PORT			0x3088
#define MT9M034_ERS_PROG_START_ADDR		0x309E
#define	MT9M034_X_ODD_INC			0x30A2
#define	MT9M034_Y_ODD_INC			0x30A6
#define MT9M034_DIGITAL_TEST			0x30B0
#define MT9M034_TEMPERATURE			0x30B2
#define MT9M034_TEMPERATURE_CONTROL		0x30B4
#define MT9M034_DIGITAL_CTRL			0x30BA
#define MT9M034_GREEN1_GAIN_CB			0x30BC
#define MT9M034_BLUE_GAIN_CB			0x30BE
#define MT9M034_RED_GAIN_CB			0x30C0
#define MT9M034_GREEN2_GAIN_CB			0x30C2
#define MT9M034_GLOBAL_GAIN_CB			0x30C4
#define MT9M034_TEMPERATURE_CALIB_1		0x30C6
#define MT9M034_TEMPERATURE_CALIB_2		0x30C8
#define MT9M034_COLUMN_CORRECTION		0x30D4
#define MT9M034_ADC_BITS_2_3			0x30E0
#define MT9M034_ADC_BITS_4_5			0x30E2
#define MT9M034_ADC_BITS_6_7			0x30E4
#define MT9M034_ADC_CONFIG1			0x30E6
#define MT9M034_ADC_CONFIG2			0x30E8
#define MT9M034_AE_CTRL_REG			0x3100
#define MT9M034_AE_LUMA_TARGET_REG		0x3102
#define MT9M034_AE_HIST_TARGET_REG		0x3104
#define MT9M034_AE_DCG_EXPOSURE_HIGH_REG	0x3112
#define MT9M034_AE_DCG_EXPOSURE_LOW_REG		0x3114
#define MT9M034_AE_DCG_GAIN_FACTOR_REG		0x3116
#define MT9M034_AE_DCG_GAIN_FACTOR_INV_REG	0x3118
#define MT9M034_AE_MAX_EXPOSURE_REG		0x311C
#define MT9M034_AE_MIN_EXPOSURE_REG		0x311E
#define MT9M034_AE_ALPHA_V1_REG			0x3126
#define MT9M034_HDR_COMP			0x31D0
#define MT9M034_ANALOG_REG			0x3ED6
#define MT9M034_DAC_LD_10_11			0x3ED6
#define MT9M034_DAC_LD_12_13			0x3ED8
#define MT9M034_DAC_LD_14_15			0x3EDA
#define MT9M034_DAC_LD_16_17			0x3EDC
#define MT9M034_DAC_LD_18_19			0x3EDE
#define MT9M034_DAC_LD_20_21			0x3EE0
#define MT9M034_DAC_LD_22_23			0x3EE2
#define MT9M034_DAC_LD_24_25			0x3EE4
#define MT9M034_DAC_LD_26_27			0x3EE6

#define MT9M034_HOR_AND_VER_BIN			0x0022
#define MT9M034_HOR_BIN				0x0011
#define MT9M034_DISABLE_BINNING			0x0000

#define MT9M034_RESET				0x00D9
#define MT9M034_STREAM_OFF			0x00D8
#define MT9M034_STREAM_ON			0x00DC


static uint16_t mt9m034_seq_data[] = {
  0x0025, 0x5050, 0x2D26, 0x0828, 0x0D17, 0x0926, 0x0028, 0x0526,
  0xA728, 0x0725, 0x8080, 0x2925, 0x0040, 0x2702, 0x1616, 0x2706,
  0x1F17, 0x3626, 0xA617, 0x0326, 0xA417, 0x1F28, 0x0526, 0x2028,
  0x0425, 0x2020, 0x2700, 0x171D, 0x2500, 0x2017, 0x1028, 0x0519,
  0x1703, 0x2706, 0x1703, 0x1741, 0x2660, 0x175A, 0x2317, 0x1122,
  0x1741, 0x2500, 0x9027, 0x0026, 0x1828, 0x002E, 0x2A28, 0x081C,
  0x1470, 0x7003, 0x1470, 0x7004, 0x1470, 0x7005, 0x1470, 0x7009,
  0x170C, 0x0014, 0x0020, 0x0014, 0x0050, 0x0314, 0x0020, 0x0314,
  0x0050, 0x0414, 0x0020, 0x0414, 0x0050, 0x0514, 0x0020, 0x2405,
  0x1400, 0x5001, 0x2550, 0x502D, 0x2608, 0x280D, 0x1709, 0x2600,
  0x2805, 0x26A7, 0x2807, 0x2580, 0x8029, 0x2500, 0x4027, 0x0216,
  0x1627, 0x0620, 0x1736, 0x26A6, 0x1703, 0x26A4, 0x171F, 0x2805,
  0x2620, 0x2804, 0x2520, 0x2027, 0x0017, 0x1D25, 0x0020, 0x1710,
  0x2805, 0x1A17, 0x0327, 0x0617, 0x0317, 0x4126, 0x6017, 0xAE25,
  0x0090, 0x2700, 0x2618, 0x2800, 0x2E2A, 0x2808, 0x1D05, 0x1470,
  0x7009, 0x1720, 0x1400, 0x2024, 0x1400, 0x5002, 0x2550, 0x502D,
  0x2608, 0x280D, 0x1709, 0x2600, 0x2805, 0x26A7, 0x2807, 0x2580,
  0x8029, 0x2500, 0x4027, 0x0216, 0x1627, 0x0617, 0x3626, 0xA617,
  0x0326, 0xA417, 0x1F28, 0x0526, 0x2028, 0x0425, 0x2020, 0x2700,
  0x171D, 0x2500, 0x2021, 0x1710, 0x2805, 0x1B17, 0x0327, 0x0617,
  0x0317, 0x4126, 0x6017, 0xAE25, 0x0090, 0x2700, 0x2618, 0x2800,
  0x2E2A, 0x2808, 0x1E17, 0x0A05, 0x1470, 0x7009, 0x1616, 0x1616,
  0x1616, 0x1616, 0x1616, 0x1616, 0x1616, 0x1616, 0x1616, 0x1616,
  0x1616, 0x1616, 0x1616, 0x1616, 0x1616, 0x1616, 0x1400, 0x2024,
  0x1400, 0x502B, 0x302C, 0x2C2C, 0x2C00, 0x0225, 0x5050, 0x2D26,
  0x0828, 0x0D17, 0x0926, 0x0028, 0x0526, 0xA728, 0x0725, 0x8080,
  0x2917, 0x0525, 0x0040, 0x2702, 0x1616, 0x2706, 0x1736, 0x26A6,
  0x1703, 0x26A4, 0x171F, 0x2805, 0x2620, 0x2804, 0x2520, 0x2027,
  0x0017, 0x1E25, 0x0020, 0x2117, 0x1028, 0x051B, 0x1703, 0x2706,
  0x1703, 0x1747, 0x2660, 0x17AE, 0x2500, 0x9027, 0x0026, 0x1828,
  0x002E, 0x2A28, 0x081E, 0x0831, 0x1440, 0x4014, 0x2020, 0x1410,
  0x1034, 0x1400, 0x1014, 0x0020, 0x1400, 0x4013, 0x1802, 0x1470,
  0x7004, 0x1470, 0x7003, 0x1470, 0x7017, 0x2002, 0x1400, 0x2002,
  0x1400, 0x5004, 0x1400, 0x2004, 0x1400, 0x5022, 0x0314, 0x0020,
  0x0314, 0x0050, 0x2C2C, 0x2C2C
};


static int      setControl ( oaCamera*, int, int );
static int	testControl ( oaCamera*, int, int );
static int	readControl ( oaCamera*, int );
static int      getFramePixelFormat ( oaCamera*, int );
static const FRAMESIZE* getFrameSizes ( oaCamera* );
static const char* getName ( oaCamera* );
static int      reset ( oaCamera* );
static int      stop ( oaCamera* );
static int      start ( oaCamera*, START_PARMS* );
static int      getControlRange ( oaCamera*, int, int*, int*, int*, int* );
static int      startReadFrame ( oaCamera*, unsigned int );
static int      finishReadFrame ( oaCamera* );
static int      readFrame ( oaCamera*, void** );
static int      closeCamera ( oaCamera* );

static void	_QHY5LIIInitFunctionPointers ( oaCamera* );
static void     _setGain ( oaCamera*, unsigned int );
static void     _setUSBTraffic ( oaCamera*, unsigned int );
static void     _setExposure ( oaCamera*, unsigned int );
static void     _setBitDepth ( oaCamera*, unsigned int );
static void     _setHighSpeed ( oaCamera*, unsigned int );
static void     _setHDR ( oaCamera*, unsigned int );
static void     _abortFrame ( oaCamera* );
static void     _setResolution ( oaCamera*, unsigned int, unsigned int );
static void     _initialiseRegisters ( oaCamera* );
static void     _setPLLRegister ( oaCamera*, unsigned int );
static int      _readTemperature ( oaCamera* );
static void     _setAllControls ( oaCamera* );
static void*    _capture ( void* );
//static void     _startExposure ( oaCamera*, unsigned int );
static int      _readExposure ( oaCamera*, void*, unsigned int );


/**
 * Initialise a given camera device
 */

int
_QHY5LIIInitCamera ( oaCamera* camera )
{
  int i, j;
  unsigned char buf[4];

  oacamDebugMsg ( DEBUG_CAM_INIT, "QHY5L-II: init: %s ()\n", __FUNCTION__ );

  OA_CLEAR ( camera->controls );
  OA_CLEAR ( camera->features );
  _QHY5LIIInitFunctionPointers ( camera );

  if ( cameraState.isColour ) {
    // FIX ME -- need to set any other video state info here?
    cameraState.videoCurrent = OA_PIX_FMT_GRBG8;
  } else {
    cameraState.videoGrey = 1;
    cameraState.videoCurrent = OA_PIX_FMT_GREY8;
  }
  cameraState.videoRGB24 = cameraState.videoGrey16 = 0;
  cameraState.currentBitDepth = 8;
  cameraState.longExposureMode = 0;

  cameraState.maxResolutionX = QHY5LII_IMAGE_WIDTH;
  cameraState.maxResolutionY = QHY5LII_IMAGE_HEIGHT;

  cameraState.resolutions[0].x = cameraState.maxResolutionX;
  cameraState.resolutions[0].y = cameraState.maxResolutionY;
  cameraState.resolutions[1].x = 1024;
  cameraState.resolutions[1].y = 768;
  cameraState.resolutions[2].x = 800;
  cameraState.resolutions[2].y = 600;
  cameraState.resolutions[3].x = 640;
  cameraState.resolutions[3].y = 480;
  cameraState.resolutions[4].x = 320;
  cameraState.resolutions[4].y = 240;
  cameraState.resolutions[5].x = 0;
  cameraState.resolutions[5].y = 0;

  cameraState.xSize = cameraState.maxResolutionX;
  cameraState.ySize = cameraState.maxResolutionY;
  cameraState.captureThreadStarted = 0;
  cameraState.captureThreadExit = 0;
  cameraState.haveDataToRead = 0;
  cameraState.nextBufferToRead = -1;
  cameraState.doSetParameters = 0;

  _usbControlMsg ( camera, QHY_REQUEST_WRITE, 0xc1, 0, 0, buf, 4, USB_TIMEOUT );

  camera->controls[ OA_CAM_CTRL_GAIN ] = OA_CAM_CTRL_TYPE_INT32;
  camera->min[ OA_CAM_CTRL_GAIN ] = 1;
  camera->max[ OA_CAM_CTRL_GAIN ] = 1000;
  camera->step[ OA_CAM_CTRL_GAIN ] = 1;
  camera->def[ OA_CAM_CTRL_GAIN ] = 500; // completely arbitrary
  cameraState.currentGain = 500;

  camera->controls[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = OA_CAM_CTRL_TYPE_INT32;
  camera->min[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 0;
  camera->max[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 10000000; // made up
  camera->step[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 1;
  // convert msec to usec
  camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = DEFAULT_EXPOSURE * 1000;
  cameraState.currentExposure = DEFAULT_EXPOSURE;

  camera->controls[ OA_CAM_CTRL_HIGHSPEED ] = OA_CAM_CTRL_TYPE_BOOLEAN;
  camera->min[ OA_CAM_CTRL_HIGHSPEED ] = 0;
  camera->max[ OA_CAM_CTRL_HIGHSPEED ] = 1;
  camera->step[ OA_CAM_CTRL_HIGHSPEED ] = 1;
  camera->def[ OA_CAM_CTRL_HIGHSPEED ] = DEFAULT_SPEED;
  cameraState.currentHighSpeed = 1;

  // FIX ME -- may be supported?
  // camera->controls[ OA_CAM_CTRL_BINNING ] = OA_CAM_CTRL_TYPE_DISCRETE;
  // camera->min[ OA_CAM_CTRL_BINNING ] = 1;
  // camera->max[ OA_CAM_CTRL_BINNING ] = 2;
  // camera->step[ OA_CAM_CTRL_BINNING ] = 1;
  // camera->def[ OA_CAM_CTRL_BINNING ] = 1;

  camera->controls[ OA_CAM_CTRL_USBTRAFFIC ] = OA_CAM_CTRL_TYPE_INT32;
  camera->min[ OA_CAM_CTRL_USBTRAFFIC ] = 0;
  camera->max[ OA_CAM_CTRL_USBTRAFFIC ] = 255;
  camera->step[ OA_CAM_CTRL_USBTRAFFIC ] = 1;
  camera->def[ OA_CAM_CTRL_USBTRAFFIC ] = DEFAULT_USBTRAFFIC;
  cameraState.currentUSBTraffic = DEFAULT_USBTRAFFIC;

  camera->controls[ OA_CAM_CTRL_HDR ] = OA_CAM_CTRL_TYPE_BOOLEAN;
  camera->min[ OA_CAM_CTRL_HDR ] = 0;
  camera->max[ OA_CAM_CTRL_HDR ] = 1;
  camera->step[ OA_CAM_CTRL_HDR ] = 1;
  camera->def[ OA_CAM_CTRL_HDR ] = 0;
  cameraState.currentHDR = 0;

  if ( cameraState.isColour ) {
    camera->controls[ OA_CAM_CTRL_RED_BALANCE ] = OA_CAM_CTRL_TYPE_INT32;
    camera->min[ OA_CAM_CTRL_RED_BALANCE ] = 0;
    camera->max[ OA_CAM_CTRL_RED_BALANCE ] = 200;
    camera->step[ OA_CAM_CTRL_RED_BALANCE ] = 1;
    camera->def[ OA_CAM_CTRL_RED_BALANCE ] = 100; // guess
    cameraState.currentRedBalance = 100;

    camera->controls[ OA_CAM_CTRL_BLUE_BALANCE ] = OA_CAM_CTRL_TYPE_INT32;
    camera->min[ OA_CAM_CTRL_BLUE_BALANCE ] = 0;
    camera->max[ OA_CAM_CTRL_BLUE_BALANCE ] = 200;
    camera->step[ OA_CAM_CTRL_BLUE_BALANCE ] = 1;
    camera->def[ OA_CAM_CTRL_BLUE_BALANCE ] = 100; // guess
    cameraState.currentBlueBalance = 100;

    camera->controls[ OA_CAM_CTRL_GREEN_BALANCE ] = OA_CAM_CTRL_TYPE_INT32;
    camera->min[ OA_CAM_CTRL_GREEN_BALANCE ] = 0;
    camera->max[ OA_CAM_CTRL_GREEN_BALANCE ] = 200;
    camera->step[ OA_CAM_CTRL_GREEN_BALANCE ] = 1;
    camera->def[ OA_CAM_CTRL_GREEN_BALANCE ] = 100; // guess
    cameraState.currentGreenBalance = 100;
  }

  camera->controls[ OA_CAM_CTRL_BIT_DEPTH ] = OA_CAM_CTRL_TYPE_DISCRETE;
  camera->controls[ OA_CAM_CTRL_TEMPERATURE ] = OA_CAM_CTRL_TYPE_READONLY;
  camera->controls[ OA_CAM_CTRL_DROPPED ] = OA_CAM_CTRL_TYPE_READONLY;
  camera->controls[ OA_CAM_CTRL_DROPPED_RESET ] = OA_CAM_CTRL_TYPE_BUTTON;

  cameraState.buffers = 0;
  cameraState.configuredBuffers = 0;

  cameraState.frameSize = cameraState.maxResolutionX *
      cameraState.maxResolutionY;
  cameraState.captureLength = cameraState.frameSize + 5;
  cameraState.imageBufferLength = 2 * ( cameraState.maxResolutionX *
      cameraState.maxResolutionY ) + QHY5LII_IMAGE_OFFSET;
  cameraState.buffers = calloc ( QHY5LII_BUFFERS, sizeof ( struct QHYbuffer ));
  for ( i = 0; i < QHY5LII_BUFFERS; i++ ) {
    void* m = malloc ( cameraState.imageBufferLength );
    if ( m ) {
      cameraState.buffers[i].start = m;
      cameraState.configuredBuffers++;
    } else {
      fprintf ( stderr, "%s malloc failed\n", __FUNCTION__ );
      if ( i ) {
        for ( j = 0; j < i; j++ ) {
          free (( void* ) cameraState.buffers[j].start );
        }
      }
      return 0;
    }
  }

  _setAllControls ( camera );

  return -1;
}


static void
_QHY5LIIInitFunctionPointers ( oaCamera* camera )
{
  // Set by QHYinit()
  // camera->funcs.initCamera = oaQHYInitCamera;
  camera->funcs.closeCamera = closeCamera;
  camera->funcs.resetCamera = reset;
  // camera->funcs.startCapture = start;
  camera->funcs.startCaptureExtended = start;
  camera->funcs.stopCapture = stop;
  // just uses versions in device entry
  // camera->funcs.hasLoadableFirmware = _cameraHasLoadableFirmware;
  // camera->funcs.isFirmwareLoaded = _cameraIsFirmwareLoaded;
  // camera->funcs.loadFirmware = _cameraLoadFirmware;
  camera->funcs.getName = getName;
  camera->funcs.getControlRange = getControlRange;
  camera->funcs.readControl = readControl;
  camera->funcs.setControl = setControl;
  camera->funcs.testControl = testControl;
  // camera->funcs.hasAuto = oaHasAuto;
  // camera->funcs.isAuto = _isAuto;
  // camera->funcs.setControlMulti = _setControlMulti;
  camera->funcs.startReadFrame = startReadFrame;
  camera->funcs.readFrame = readFrame;
  camera->funcs.finishReadFrame = finishReadFrame;
  camera->funcs.enumerateFrameSizes = getFrameSizes;
  // camera->funcs.enumerateFrameRates = _enumerateFrameRates;
  camera->funcs.getFramePixelFormat = getFramePixelFormat;
  // camera->funcs.testROISize = testROISize;
  // camera->funcs.getMenuString = _getMenuString;
  // camera->funcs.getAutoWBManualSetting = _getAutoWBManualSetting;
  // camera->funcs.hasFixedFrameRates = _hasFixedFrameRates;
}


static void
_setAllControls ( oaCamera* camera )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "QHY5L-II: control: %s()\n",
      __FUNCTION__ );

  pthread_mutex_lock ( &cameraState.captureMutex );
  _setBitDepth ( camera, cameraState.currentBitDepth );
  _setHDR ( camera, cameraState.currentHDR );
  _setUSBTraffic ( camera, cameraState.currentUSBTraffic );
  _setHighSpeed ( camera, cameraState.currentHighSpeed );
  _setResolution ( camera, cameraState.xSize, cameraState.ySize );
  _setExposure ( camera, cameraState.currentExposure );
  _setGain ( camera, cameraState.currentGain );
  _setUSBTraffic ( camera, cameraState.currentUSBTraffic );
  _setExposure ( camera, cameraState.currentExposure );
  pthread_mutex_unlock ( &cameraState.captureMutex );
}


static int
getControlRange ( oaCamera* camera, int control, int* min, int* max,
    int* step, int* def )
{
  if ( !camera->controls[ control ] ) {
    return -OA_ERR_INVALID_CONTROL;
  }

  *min = camera->min[ control ];
  *max = camera->max[ control ];
  *step = camera->step[ control ];
  *def = camera->def[ control ];
  return 0;
}


static const char*
getName ( oaCamera* camera )
{
  if ( cameraState.isColour ) {
    return "QHY5L-II colour";
  }
  return "QHY5L-II mono";
}


const static FRAMESIZE*
getFrameSizes ( oaCamera* camera )
{
  return cameraState.resolutions;
}


static int
closeCamera ( oaCamera* camera )
{
  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s()\n",
      __FUNCTION__ );

  libusb_release_interface ( cameraState.usbHandle, 0 );
  libusb_close ( cameraState.usbHandle );
  libusb_exit ( cameraState.usbContext );
  return 0;
}


static int
getFramePixelFormat ( oaCamera* camera, int depth )
{
  if ( 12 == depth || 16 == depth || ( 0 == depth &&
      cameraState.currentBitDepth > 8 ) ) {
    if ( cameraState.isColour ) {
      return OA_PIX_FMT_GRBG16BE;
    }
    return OA_PIX_FMT_GREY16BE;
  }
  if ( 8 == depth || 0 == depth ) {
    if ( cameraState.isColour ) {
      return OA_PIX_FMT_GRBG8;
    }
    return OA_PIX_FMT_GREY8;
  }
  return -OA_ERR_INVALID_BIT_DEPTH;
}


static int
readControl ( oaCamera* camera, int control )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "QHY5L-II: control: %s ( %d )\n",
      __FUNCTION__, control );

  if ( !camera->controls[ control ] ) {
    fprintf ( stderr, "Unimplemented control %d in QHY5L-II:%s\n", control,
        __FUNCTION__ );
    return -1;
  }

  switch ( control ) {

    case OA_CAM_CTRL_GAIN:
      return cameraState.currentGain;
      break;

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      return cameraState.currentExposure * 1000;
      break;

    case OA_CAM_CTRL_HIGHSPEED:
      return cameraState.currentHighSpeed;
      break;

    case OA_CAM_CTRL_USBTRAFFIC:
      return cameraState.currentUSBTraffic;
      break;

    case OA_CAM_CTRL_HDR:
      return cameraState.currentHDR;
      break;

    case OA_CAM_CTRL_RED_BALANCE:
      if ( cameraState.isColour ) {
        return cameraState.currentRedBalance;
      }
      break;

    case OA_CAM_CTRL_BLUE_BALANCE:
      if ( cameraState.isColour ) {
        return cameraState.currentBlueBalance;
      }
      break;

    case OA_CAM_CTRL_GREEN_BALANCE:
      if ( cameraState.isColour ) {
        return cameraState.currentGreenBalance;
      }
      break;

    case OA_CAM_CTRL_TEMPERATURE:
      return _readTemperature ( camera );
      break;

    case OA_CAM_CTRL_BIT_DEPTH:
      return cameraState.currentBitDepth;
      break;

    case OA_CAM_CTRL_DROPPED:
      return cameraState.droppedFrames;
      break;

    // FIX ME -- may be supported
    // case OA_CAM_CTRL_BINNING:
    //   fprintf ( stderr, "QHY5L-II: control %d not handled in %s\n", control,
    //       __FUNCTION__ );
    //   break;
  }
  fprintf ( stderr, "Unimplemented control %d in QHY5L-II:%s\n", control,
      __FUNCTION__ );
  return -1;
}


static int
testControl ( oaCamera* camera, int control, int value )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "QHY5L-II: control: %s ( %d, %d )\n",
      __FUNCTION__, control, value );

  switch ( control ) {

    case OA_CAM_CTRL_GAIN:
    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
    case OA_CAM_CTRL_USBTRAFFIC:
      if ( value >= camera->min[ control ] &&
          value <= camera->max[ control ] &&
          ( 0 == ( value - camera->min[ control ]) %
          camera->step[ control ])) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CAM_CTRL_RED_BALANCE:
    case OA_CAM_CTRL_BLUE_BALANCE:
    case OA_CAM_CTRL_GREEN_BALANCE:
      if ( camera->controls[ control ]) {
        if ( value >= camera->min[ control ] &&
            value <= camera->max[ control ] &&
            ( 0 == ( value - camera->min[ control ]) %
            camera->step[ control ])) {
          return OA_ERR_NONE;
        }
      }
      break;

    case OA_CAM_CTRL_BIT_DEPTH:
      if ( 8 == value || 12 == value || 16 == value ) {
        return OA_ERR_NONE;
      }
      break;

    // FIX ME -- may be supported?
    // case OA_CAM_CTRL_BINNING:
    //   if ( 1 == value || 2 == value ) {
    //     return OA_ERR_NONE;
    //   }
    //   break;

    case OA_CAM_CTRL_HIGHSPEED:
      return OA_ERR_NONE;
      break;

    case OA_CAM_CTRL_HDR:
      fprintf ( stderr, "QHY5L-II: %s not yet implemented for control %d\n",
          __FUNCTION__, control );
      return -OA_ERR_INVALID_CONTROL;
      break;

    case OA_CAM_CTRL_TEMPERATURE:
    default:
      return -OA_ERR_INVALID_CONTROL;
      break;
  }
  return -OA_ERR_OUT_OF_RANGE;
}


static int
setControl ( oaCamera* camera, int control, int value )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "QHY5L-II: control: %s ( %d, %d )\n",
      __FUNCTION__, control, value );

  if ( camera->controls[ control ] == OA_CAM_CTRL_TYPE_INT32 && ( value <
      camera->min[ control ] || value > camera->max[ control ] )) {
    return -1;
  }

  switch ( control ) {

    case OA_CAM_CTRL_GAIN:
      pthread_mutex_lock ( &cameraState.captureMutex );
      _setGain ( camera, value );
      pthread_mutex_unlock ( &cameraState.captureMutex );
      cameraState.currentGain = value;
      break;

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      // API exposure time is in units of usec, we use msec
      value /= 1000;
      if ( value < 1 ) { value = 1; }
      pthread_mutex_lock ( &cameraState.captureMutex );
      _setExposure ( camera, value );
      cameraState.currentExposure = value;
      cameraState.transferTime = cameraState.xSize * cameraState.ySize /
          PIXEL_RATE / 1000;
      pthread_mutex_unlock ( &cameraState.captureMutex );
      break;

    case OA_CAM_CTRL_BIT_DEPTH:
      if ( 8 == value || 12 == value || 16 == value ) {
        cameraState.currentBitDepth = value;
        _setAllControls ( camera );
        if ( cameraState.isColour ) {
          if ( 8 == value ) {
            cameraState.videoCurrent = OA_PIX_FMT_GRBG8;
          } else {
            cameraState.videoCurrent = OA_PIX_FMT_GRBG16BE;
          }
        } else {
          if ( 8 == value ) {
            cameraState.videoCurrent = OA_PIX_FMT_GREY8;
          } else {
            cameraState.videoCurrent = OA_PIX_FMT_GREY16BE;
          }
        }
      }
      break;

    case OA_CAM_CTRL_HIGHSPEED:
      pthread_mutex_lock ( &cameraState.captureMutex );
      _setHighSpeed ( camera, value );
      pthread_mutex_unlock ( &cameraState.captureMutex );
      cameraState.currentHighSpeed = value;
      break;

    case OA_CAM_CTRL_USBTRAFFIC:
      pthread_mutex_lock ( &cameraState.captureMutex );
      _setUSBTraffic ( camera, value );
      pthread_mutex_unlock ( &cameraState.captureMutex );
      cameraState.currentUSBTraffic = value;
      break;

    case OA_CAM_CTRL_BLUE_BALANCE:
      cameraState.currentBlueBalance = value;
      pthread_mutex_lock ( &cameraState.captureMutex );
      pthread_mutex_unlock ( &cameraState.captureMutex );
      _setGain ( camera, cameraState.currentGain );
      break;

    case OA_CAM_CTRL_RED_BALANCE:
      cameraState.currentRedBalance = value;
      pthread_mutex_lock ( &cameraState.captureMutex );
      _setGain ( camera, cameraState.currentGain );
      pthread_mutex_unlock ( &cameraState.captureMutex );
      break;

    case OA_CAM_CTRL_DROPPED_RESET:
      // FIX ME -- droppedFrames should probably be mutexed
      cameraState.droppedFrames = 0;
      break;

    case OA_CAM_CTRL_GREEN_BALANCE:
      fprintf ( stderr, "QHY5L-II: No idea how to set green balance\n" );
      break;

    default:
      fprintf ( stderr, "QHY5L-II: unrecognised control %d in %s\n", control,
          __FUNCTION__ );
      return 0;
      break;
  }
  return -1;
}


static int
reset ( oaCamera* camera )
{
  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s()\n",
      __FUNCTION__ );

  // The same as stop for the QHY5L-II

  stop ( camera );
  // _init ( camera );
  return 0;
}


static void
_setGain ( oaCamera* camera, unsigned int gain )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "QHY5L-II: control: %s ( %d )\n",
      __FUNCTION__, gain );

  static double C[8] = {10, 8, 5, 4, 2.5, 2, 1.25, 1};
  double S[8];
  int A[8], B[8];
  double error[8];
  double minValue, rebasedGain;
  unsigned short REG30B0, baseDGain, offset, data;
  int i, minValueIndex, limit;

  if ( cameraState.isColour ) {
    // This re-bases the gain from 0 to [ COLOUR_GAIN_MIN/10,
    // COLOUR_GAIN_MAX/10 ]
    if ( gain < 26 ) { gain = 26; }
    rebasedGain = (( COLOUR_GAIN_MAX - COLOUR_GAIN_MIN ) * gain / 1000.0 +
        COLOUR_GAIN_MIN  ) / 10.0;
  } else {
    // This re-bases the gain from 0 to [ MONO_GAIN_MIN/10, MONO_GAIN_MAX/10 ]
    rebasedGain = (( MONO_GAIN_MAX - MONO_GAIN_MIN ) * gain / 1000.0 +
        MONO_GAIN_MIN  ) / 10.0;
  }

  REG30B0 = ( cameraState.longExposureMode ? 0x5330 : 0x1330 ) & ~0x0030;
  limit = cameraState.isColour ? 3 : 7;

  for ( i = 0; i < 8; i++ ) {
    S[i] = rebasedGain / C[i];
    A[i] = S[i];
    B[i] = ( S[i] - A[i] ) / 0.03125;
    if ( !A[i] || A[i] > limit ) {
      A[i] = 10000;
    }
    error[i] = fabs (( A[i] + B[i] * 0.03125 ) * C[i] - rebasedGain );
    if (( 0 == i ) ||  minValue > error[i] ) {
      minValue = error[i];
      minValueIndex = i;
    }
  }

  offset = 0x30 - 0x10 * ( minValueIndex >> 1 );
  data = 0xd308 - 0x100 * ( minValueIndex & 1 );

  if ( cameraState.longExposureMode ) {
    _setExposure ( camera, 1 );
  }

  _i2cWrite16 ( camera, MT9M034_DIGITAL_TEST, REG30B0 + offset );
  _i2cWrite16 ( camera, MT9M034_DAC_LD_24_25, data );

  baseDGain = B[ minValueIndex ] + A[ minValueIndex ] * 32;

  if ( cameraState.isColour ) {
    _i2cWrite16 ( camera, MT9M034_BLUE_GAIN, baseDGain *
        cameraState.currentBlueBalance / 100 );
    _i2cWrite16 ( camera, MT9M034_RED_GAIN, baseDGain *
        cameraState.currentRedBalance / 100 );
    _i2cWrite16 ( camera, MT9M034_GREEN2_GAIN, baseDGain );
    _i2cWrite16 ( camera, MT9M034_GREEN1_GAIN, baseDGain );
  } else {
    _i2cWrite16 ( camera, MT9M034_GLOBAL_GAIN, baseDGain );
  }

  if ( cameraState.longExposureMode ) {
    usleep ( 500000 );
    _setExposure ( camera, cameraState.currentExposure );
  }
}


static void
_setHighSpeed ( oaCamera* camera, unsigned int value )
{
  unsigned char buf;

  oacamDebugMsg ( DEBUG_CAM_CTRL, "QHY5L-II: control: %s ( %d )\n",
      __FUNCTION__, value );

  if ( value ) {
    if ( 8 == cameraState.currentBitDepth ) {
      cameraState.CMOSClock = 48;
      buf = 2;
    } else {
      cameraState.CMOSClock = 24;
      buf = 1;
    }
  } else {
    if ( 8 == cameraState.currentBitDepth ) {
      cameraState.CMOSClock = 24;
      buf = 1;
    } else {
      cameraState.CMOSClock = 12;
      buf = 0;
    }
  }
  _usbControlMsg ( camera, QHY_REQUEST_WRITE, 0xc8, 0, 0, &buf, 1, 0 );
}


static void
_setUSBTraffic ( oaCamera* camera, unsigned int value )
{
  unsigned int base;

  oacamDebugMsg ( DEBUG_CAM_CTRL, "QHY5L-II: control: %s ( %d )\n",
      __FUNCTION__, value );

  base = ( QHY5LII_IMAGE_WIDTH == cameraState.xSize ) ? 1650 : 1388;
  _i2cWrite16 ( camera, MT9M034_LINE_LENGTH_PCK, base + value * 50 );
}


static void
_setExposure ( oaCamera* camera, unsigned int value )
{
  double pixelTime, rowTime, maxShortExposureTime;
  unsigned short rowLength, shortExposureTime, coarseIntegrationTime;
  unsigned long newTimeMillisec, newTimeMicrosec;
  unsigned char buf[4];

  oacamDebugMsg ( DEBUG_CAM_CTRL, "QHY5L-II: control: %s ( %d )\n",
      __FUNCTION__, value );

  pixelTime = 1.0 / cameraState.CMOSClock / cameraState.PLLRatio;

  rowLength = _i2cRead16 ( camera, MT9M034_LINE_LENGTH_PCK );
  rowTime = rowLength * pixelTime;
  maxShortExposureTime = 65000 * rowTime;
  newTimeMicrosec = value * 1000;


  if ( newTimeMicrosec > maxShortExposureTime ) {
     fprintf ( stderr, "%s: long exposure mode may not work\n", __FUNCTION__ );
     cameraState.longExposureMode = 1;
     shortExposureTime = 65000;
     _i2cWrite16 ( camera, MT9M034_COARSE_INTEGRATION_TIME, shortExposureTime );
     newTimeMillisec = ( newTimeMicrosec - maxShortExposureTime ) / 1000;
     buf[0] = 0;
     buf[1] = ( newTimeMillisec >> 16 ) & 0xff;
     buf[2] = ( newTimeMillisec >> 8 ) & 0xff;
     buf[3] = newTimeMillisec & 0xff;
     _usbControlMsg ( camera, QHY_REQUEST_WRITE, 0xc1, 0, 0, buf, 4, 0 );
  } else {
     _abortFrame ( camera );
     cameraState.longExposureMode = 0;
     if (( coarseIntegrationTime = newTimeMicrosec / rowTime ) < 1 ) {
       coarseIntegrationTime = 1;
     }
     _i2cWrite16 ( camera, MT9M034_COARSE_INTEGRATION_TIME,
         coarseIntegrationTime );
  }
}


static void
_setBitDepth ( oaCamera* camera, unsigned int depth )
{
  unsigned char buf;

  oacamDebugMsg ( DEBUG_CAM_CTRL, "QHY5L-II: control: %s ( %d )\n",
      __FUNCTION__, depth );

  buf = ( depth <= 8 ) ? 0 : 1;
  _usbControlMsg ( camera, QHY_REQUEST_WRITE, 0xcd, 0, 0, &buf, 1, 0 );
}


static void
_abortFrame ( oaCamera* camera )
{
  unsigned char buf[4] = { 0, 0, 0, 0 };

  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s()\n",
      __FUNCTION__ );

  _usbControlMsg ( camera, QHY_REQUEST_WRITE, 0xc1, 0, 0, buf, 4, 0 );
  usleep ( 100 );
}


static void
_setResolution ( oaCamera* camera, unsigned int x, unsigned int y )
{
  unsigned int clock, xStart, yStart, xEnd, yEnd, frameLength, rowLength;

  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s ( %d, %d )\n",
      __FUNCTION__, x, y );

  frameLength = y + 26;
  rowLength = 1388;

  if ( x < 320 && y < 240 ) {
    clock = 1;
  } else if ( x < 640 && y < 480 ) {
    clock = 1;
  } else if ( x < 800 && y < 600 ) {
    clock = 2;
  } else if ( x < 1024 && y < 768 ) {
    clock = 0;
  } else {
    clock = 0;
    frameLength = 990;
    rowLength = 1650;
  }
  rowLength += cameraState.currentUSBTraffic * 50;

  _initialiseRegisters ( camera );
  _setPLLRegister ( camera, clock );

  xStart = 4 + ( QHY5LII_IMAGE_WIDTH - x ) / 2;
  yStart = 4 + ( QHY5LII_IMAGE_HEIGHT - y ) / 2;
  xEnd = xStart + x - 1;
  yEnd = yStart + y - 1;

  _i2cWrite16 ( camera, MT9M034_Y_ADDR_START, yStart );
  _i2cWrite16 ( camera, MT9M034_X_ADDR_START, xStart );
  _i2cWrite16 ( camera, MT9M034_Y_ADDR_END, yEnd );
  _i2cWrite16 ( camera, MT9M034_X_ADDR_END, xEnd );
  _i2cWrite16 ( camera, MT9M034_FRAME_LENGTH_LINES, frameLength );
  _i2cWrite16 ( camera, MT9M034_LINE_LENGTH_PCK, rowLength );
  _i2cWrite16 ( camera, MT9M034_RESET_REGISTER, 0x10DC );

  cameraState.frameSize = x * y *
      (( cameraState.currentBitDepth <= 8 ) ? 1 : 2 );
  cameraState.captureLength = cameraState.frameSize + QHY5LII_IMAGE_OFFSET;
}


static void
_setHDR ( oaCamera* camera, unsigned int state )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "QHY5L-II: control: %s ( %d, %d )\n",
      __FUNCTION__, state );

  if ( state ) {
    fprintf ( stderr, "QHY5L-II: %s: function unimplemented\n", __FUNCTION__ );
  }
}


static int
start ( oaCamera* camera, START_PARMS* params )
{
  unsigned char buf[1];

  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s ( %d, %d )\n",
      __FUNCTION__, params->size.x, params->size.y );

  if ( cameraState.captureThreadStarted ) {
    fprintf ( stderr, "capture thread already exists\n" );
    return -1;
  }
  cameraState.captureThreadStarted = 0;
  cameraState.captureThreadExit = 0;
  cameraState.nextBufferToRead = 0;
  if ( cameraState.xSize != params->size.x ||
      cameraState.ySize != params->size.y ) {
    cameraState.xSize = params->size.x;
    cameraState.ySize = params->size.y;
    _setAllControls ( camera );
  }
  buf[0] = 100;
  _usbControlMsg ( camera, QHY_REQUEST_WRITE, 0xb3, 0, 0, buf, 1, 0 );
  pthread_create ( &cameraState.captureThread, 0, &_capture, camera );
  cameraState.captureThreadStarted = 1;
  return 0;
}


static int
stop ( oaCamera* camera )
{
  unsigned char buf[4];

  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s()\n",
      __FUNCTION__ );

  if ( !cameraState.captureThreadStarted ) {
    return 0;
  }

  pthread_mutex_lock ( &cameraState.captureMutex );
  cameraState.captureThreadExit = 1;
  pthread_mutex_unlock ( &cameraState.captureMutex );
  pthread_join ( cameraState.captureThread, 0 );
  cameraState.captureThreadStarted = 0;

  _usbControlMsg ( camera, QHY_REQUEST_WRITE, 0xc1, 0, 0, buf, 4, 0 );
  return 0;
}


static void*
_capture ( void* param )
{
  oaCamera*     camera = param;
  int           quitThread, nextBuffer, readState;
  unsigned int  waitTime;
  void*         data;

  pthread_mutex_lock ( &cameraState.captureMutex );
  cameraState.haveDataToRead = 0;
  cameraState.nextBufferToRead = -1;
  pthread_mutex_unlock ( &cameraState.captureMutex );
  nextBuffer = 0;

  while ( 1 ) {

    oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s() in capture loop\n",
        __FUNCTION__ );

    pthread_mutex_lock ( &cameraState.captureMutex );
    quitThread = cameraState.captureThreadExit;
    if ( nextBuffer == cameraState.nextBufferToRead ) {
      nextBuffer = ( nextBuffer + 1 ) % cameraState.configuredBuffers;
    }
    // show we precalculate this?
    waitTime = cameraState.currentExposure * 500;
    waitTime = ( cameraState.currentExposure - 50 ) * 1000;
    if ( waitTime < 100 ) { waitTime = 100; }
    pthread_mutex_unlock ( &cameraState.captureMutex );
    if ( quitThread ) {
      break;
    }

    data = cameraState.buffers[ nextBuffer ].start;
    // pthread_mutex_lock ( &cameraState.captureMutex );
    // _startExposure ( camera, cameraState.currentExposure );
    // pthread_mutex_unlock ( &cameraState.captureMutex );
    usleep ( 500 );
    readState = _readExposure ( camera, data, cameraState.captureLength );

    oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s() read exposure, "
        "waiting for lock\n", __FUNCTION__ );

    pthread_mutex_lock ( &cameraState.captureMutex );

    oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s() got lock\n",
        __FUNCTION__ );

    if ( !readState ) {
      cameraState.nextBufferToRead = nextBuffer;
      cameraState.haveDataToRead = 1;
      oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s(), notifying that "
          "frame is available\n", __FUNCTION__ );
      pthread_cond_signal ( &cameraState.frameAvailable );
    }
    pthread_mutex_unlock ( &cameraState.captureMutex );
    nextBuffer = ( nextBuffer + 1 ) % cameraState.configuredBuffers;
  }
  return 0;
}


/*
static void
_startExposure ( oaCamera* camera, unsigned int frameTime )
{
  cameraState.frameTime = frameTime;
  _setExposure ( camera, frameTime );
}
*/

static int
_readExposure ( oaCamera* camera, void* buffer, unsigned int readLength )
{
  int xferred, ret;

  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s()\n",
      __FUNCTION__ );

  // Great weirdness is here.  Sometimes the camera will return the amount
  // of data requested.  Sometimes it's five bytes short and the next
  // transfer request gets those remaining bytes.  It's utterly unclear to
  // me what should be going on, so I've just attempted to deal with the
  // mess as tidily as possible

  // first have a go at reading all the data

  // FIX ME -- references to droppedFrames here should probably be mutexed

  ret = _usbBulkTransfer ( camera, 0x82, buffer, readLength, &xferred,
      USB_TIMEOUT );
  if ( ret < 0 ) {
    fprintf ( stderr, "QHY5L-II::%s, usbBulkTransfer returns %d\n",
        __FUNCTION__, ret );
    cameraState.droppedFrames++;
    return ret;
  }
  // now, if we've read five bytes less than expected, read the remainder
  // into a dummy buffer as we won't use them anyhow

  if ( xferred != readLength ) {
    int ret2, xfer2;
    uint8_t dummy[ QHY5LII_IMAGE_OFFSET ];

    if ( xferred == ( readLength - QHY5LII_IMAGE_OFFSET )) {

      oacamDebugMsg ( DEBUG_CAM_USB, "QHY5L-II: USB: %s, usbBulkTransfer "
          "off by %d, doing dummy read\n", __FUNCTION__, QHY5LII_IMAGE_OFFSET );

      ret2 = _usbBulkTransfer ( camera, 0x82, dummy, QHY5LII_IMAGE_OFFSET,
          &xfer2, USB_TIMEOUT );
      if ( ret2 < 0 ) {
        fprintf ( stderr, "QHY5L-II::%s, usbBulkTransfer(2) returns %d\n",
            __FUNCTION__, ret );
        cameraState.droppedFrames++;
        return ret2;
      }
      xferred += xfer2;
    }
  }

  if ( xferred != readLength ) {
    oacamDebugMsg ( DEBUG_CAM_USB, "QHY5L-II: USB: %s, usbBulkTransfer "
        "returns %d bytes, expected %d\n", __FUNCTION__, xferred, readLength );
    if ( xferred < readLength ) {
      cameraState.droppedFrames++;
      return -1;
    }
  }

  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s() returns %d\n",
      __FUNCTION__, 0 );

  return 0;
}


static int
startReadFrame ( oaCamera* camera, unsigned int frameTime )
{
  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s()\n",
      __FUNCTION__ );

  return 0;
}


static int
readFrame ( oaCamera* camera, void** buffer )
{
  int haveData = 0;
  int bufferToRead;
  int ret;
  struct timespec waitUntil;
  struct timeval now;
  long frameSec = 0;
  long frameNanoSec = 0;

  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s()\n",
      __FUNCTION__ );

  frameNanoSec = cameraState.frameTime * 1000000;
  frameSec = cameraState.frameTime / 1000 + 5;

  pthread_mutex_lock ( &cameraState.captureMutex );

  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s(), waiting for "
      "frame available signal\n", __FUNCTION__ );

  while ( !cameraState.haveDataToRead ) {

    gettimeofday ( &now, 0 );
    waitUntil.tv_sec = now.tv_sec + frameSec;
    if (( now.tv_usec * 1000 + frameNanoSec ) > 1000000000 ) {
      waitUntil.tv_sec++;
    }
    waitUntil.tv_nsec = ( now.tv_usec * 1000 + frameNanoSec ) % 1000000000;

    oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s(), waiting...\n",
        __FUNCTION__ );

    if (( ret = pthread_cond_timedwait ( &cameraState.frameAvailable,
        &cameraState.captureMutex, &waitUntil )) && ret != ETIMEDOUT ) {
      oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s(), "
          "pthread_cond_timedwait returned error %d\n", __FUNCTION__, ret );
    }
  }

  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s(), frame "
      "available signal received & data present\n", __FUNCTION__ );

  haveData = cameraState.haveDataToRead;
  bufferToRead = cameraState.nextBufferToRead;
  cameraState.haveDataToRead = 0;
  pthread_mutex_unlock ( &cameraState.captureMutex );

  *buffer = haveData ? cameraState.buffers[ bufferToRead ].start : 0;
  return ( haveData ? cameraState.frameSize : 0 );
}


static int
finishReadFrame ( oaCamera* camera )
{
  oacamDebugMsg ( DEBUG_CAM_CMD, "QHY5L-II: command: %s()\n",
      __FUNCTION__ );
  return 0;
}


static void
_initialiseRegisters ( oaCamera* camera )
{
  int i;

  _i2cWrite16 ( camera, MT9M034_RESET_REGISTER, 0x0001);
  _i2cWrite16 ( camera, MT9M034_RESET_REGISTER, 0x10D8);
  usleep(100000);

  // Linear sequencer

  _i2cWrite16 ( camera, MT9M034_SEQ_CTRL_PORT, 0x8000 );

  for ( i = 0; i < 276; i++ ) {
    _i2cWrite16 ( camera, MT9M034_SEQ_DATA_PORT, mt9m034_seq_data[i] );
  }

  // RESERVED_MFR_309E
  _i2cWrite16 ( camera, MT9M034_ERS_PROG_START_ADDR, 0x018A );


  _i2cWrite16 ( camera, MT9M034_RESET_REGISTER, 0x10D8 );
  _i2cWrite16 ( camera, MT9M034_MODE_CTRL, 0x0029 );
  _i2cWrite16 ( camera, MT9M034_DATA_PEDESTAL, 0x00C8 );
  _i2cWrite16 ( camera, 0x3EDA, 0x0F03); // RESERVED_MFR_3EDA
  _i2cWrite16 ( camera, 0x3EDE, 0xC007); // RESERVED_MFR_3EDE
  _i2cWrite16 ( camera, 0x3ED8, 0x01EF); // RESERVED_MFR_3ED8
  _i2cWrite16 ( camera, 0x3EE2, 0xA46B); // RESERVED_MFR_3EE2
  _i2cWrite16 ( camera, 0x3EE0, 0x067D); // RESERVED_MFR_3EE0
  _i2cWrite16 ( camera, 0x3EDC, 0x0070); // RESERVED_MFR_3EDC
  _i2cWrite16 ( camera, MT9M034_DARK_CONTROL, 0x0404 );
  _i2cWrite16 ( camera, 0x3EE6, 0x4303); // RESERVED_MFR_3EE6
  _i2cWrite16 ( camera, MT9M034_DAC_LD_24_25, 0xD208 );
  _i2cWrite16 ( camera, 0x3ED6, 0x00BD); // RESERVED_MFR_3ED6
  _i2cWrite16 ( camera, 0x3EE6, 0x8303); // RESERVED_MFR_3EE6
  _i2cWrite16 ( camera, 0x30E4, 0x6372); // RESERVED_MFR_30E4
  _i2cWrite16 ( camera, 0x30E2, 0x7253); // RESERVED_MFR_30E2
  _i2cWrite16 ( camera, 0x30E0, 0x5470); // RESERVED_MFR_30E0
  _i2cWrite16 ( camera, 0x30E6, 0xC4CC); // RESERVED_MFR_30E6
  _i2cWrite16 ( camera, 0x30E8, 0x8050); // RESERVED_MFR_30E8
  usleep(200);
  _i2cWrite16 ( camera, MT9M034_VT_SYS_CLK_DIV, 14 );
  _i2cWrite16 ( camera, MT9M034_VT_PIX_CLK_DIV, 1 );
  _i2cWrite16 ( camera, MT9M034_PRE_PLL_CLK_DIV, 3 );
  _i2cWrite16 ( camera, MT9M034_PLL_MULTIPLIER, 65 );
  _i2cWrite16 ( camera, MT9M034_MODE_CTRL, 0x0029 );
  _i2cWrite16 ( camera, MT9M034_DIGITAL_TEST, 0x5330 );
  _i2cWrite16 ( camera, MT9M034_GLOBAL_GAIN, 0x00ff );
  _i2cWrite16 ( camera, MT9M034_COARSE_INTEGRATION_TIME, 0x0020 );
  _i2cWrite16 ( camera, MT9M034_EMBEDDED_DATA_CTRL, 0x1802 );
}


static void
_setPLLRegister ( oaCamera* camera, unsigned int clock )
{
  static unsigned short multiplier[] = { 42, 65, 57 };
  double pllRatio = 1.0;

   if ( clock && !cameraState.longExposureMode ) {
     pllRatio = multiplier[ clock ] / 42.0;
   }
   cameraState.PLLRatio = pllRatio;

   _i2cWrite16 ( camera, MT9M034_VT_SYS_CLK_DIV, 14 );
   _i2cWrite16 ( camera, MT9M034_VT_PIX_CLK_DIV, 1 );
   _i2cWrite16 ( camera, MT9M034_PRE_PLL_CLK_DIV, 3 );

   _i2cWrite16 ( camera, MT9M034_PLL_MULTIPLIER, multiplier[ clock ]);
   _i2cWrite16 ( camera, MT9M034_MODE_CTRL, 0x0029 );

   _i2cWrite16 ( camera, MT9M034_DIGITAL_TEST, cameraState.longExposureMode ?
        0x5330 : 0x1330 ); // 5370: PLL BYPASS   1370  USE PLL

   _i2cWrite16 ( camera, MT9M034_GLOBAL_GAIN, 0x00ff );
   _i2cWrite16 ( camera, MT9M034_COARSE_INTEGRATION_TIME, 0x0020 );

   _i2cWrite16 ( camera, MT9M034_EMBEDDED_DATA_CTRL, 0x1802);
}


static int
_readTemperature ( oaCamera* camera )
{
  double gradient, baseline;
  uint16_t value, calibrationVal1, calibrationVal2;

  // start measuring
  _i2cWrite16 ( camera, MT9M034_TEMPERATURE_CONTROL, 0x0011 );

  calibrationVal1 = _i2cRead16 ( camera, MT9M034_TEMPERATURE_CALIB_1 );
  calibrationVal2 = _i2cRead16 ( camera, MT9M034_TEMPERATURE_CALIB_2 );

  // stop measuring
  _i2cWrite16 ( camera, MT9M034_TEMPERATURE_CONTROL, 0x0000 );
  value = _i2cRead16 ( camera, MT9M034_TEMPERATURE );

  gradient = ( 70.0 - 55.0 ) / ( calibrationVal1 - calibrationVal2 );
  baseline = gradient * calibrationVal1 - 70.0;
  return ( int )(( gradient * value - baseline ) * 10 );
}
