/*****************************************************************************
 *
 * UVCinit.c -- Initialise UVC cameras
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#if HAVE_LIBUVC

#include <openastro/oacam.h>

#include <libuvc/libuvc.h>

#include "UVCoacam.h"
#include "unimplemented.h"
#include "oacamprivate.h"

#define	cameraState	camera->_uvc

// FIX ME -- move this somewhere more sensible
// This is required because the bitmap of known controls in the
// processing units does not match the id of the control.  Gah!!!

struct puCtrl controlData[] = {
  {
    .uvcControl		= UVC_PU_BRIGHTNESS_CONTROL,
    .oaControl		= OA_CAM_CTRL_BRIGHTNESS,
    .oaControlType	= OA_CAM_CTRL_TYPE_INT32,
    .size		= 2
  }, {
    .uvcControl		= UVC_PU_CONTRAST_CONTROL,
    .oaControl		= OA_CAM_CTRL_CONTRAST,
    .oaControlType	= OA_CAM_CTRL_TYPE_INT32,
    .size		= 2
  }, {
    .uvcControl		= UVC_PU_HUE_CONTROL,
    .oaControl		= OA_CAM_CTRL_HUE,
    .oaControlType	= OA_CAM_CTRL_TYPE_INT32,
    .size		= 2
  }, {
    .uvcControl		= UVC_PU_SATURATION_CONTROL,
    .oaControl		= OA_CAM_CTRL_SATURATION,
    .oaControlType	= OA_CAM_CTRL_TYPE_INT32,
    .size		= 2
  }, {
    .uvcControl		= UVC_PU_SHARPNESS_CONTROL,
    .oaControl		= OA_CAM_CTRL_SHARPNESS,
    .oaControlType	= OA_CAM_CTRL_TYPE_INT32,
    .size		= 2
  }, {
    .uvcControl		= UVC_PU_GAMMA_CONTROL,
    .oaControl		= OA_CAM_CTRL_GAMMA,
    .oaControlType	= OA_CAM_CTRL_TYPE_INT32,
    .size		= 2
  }, {
    .uvcControl		= UVC_PU_WHITE_BALANCE_TEMPERATURE_CONTROL,
    .oaControl		= OA_CAM_CTRL_WHITE_BALANCE_TEMP,
    .oaControlType	= OA_CAM_CTRL_TYPE_INT32,
    .size		= 2
  }, {
    .uvcControl		= UVC_PU_WHITE_BALANCE_COMPONENT_CONTROL,
    .oaControl		= 0, // odd one.  R&B combined?
    .oaControlType	= OA_CAM_CTRL_TYPE_INT32,
    .size		= 4
  }, {
    .uvcControl		= UVC_PU_BACKLIGHT_COMPENSATION_CONTROL,
    .oaControl		= 0,
    .oaControlType	= OA_CAM_CTRL_TYPE_INT32,
    .size		= 2
  }, {
    .uvcControl		= UVC_PU_GAIN_CONTROL,
    .oaControl		= OA_CAM_CTRL_GAIN,
    .oaControlType	= OA_CAM_CTRL_TYPE_INT32,
    .size		= 2
  }, {
    .uvcControl		= UVC_PU_POWER_LINE_FREQUENCY_CONTROL,
    .oaControl		= 0,
    .oaControlType	= OA_CAM_CTRL_TYPE_MENU,
    .size		= 1
  }, {
    .uvcControl		= UVC_PU_HUE_AUTO_CONTROL,
    .oaControl		= OA_CAM_CTRL_HUE_AUTO,
    .oaControlType	= OA_CAM_CTRL_TYPE_BOOLEAN,
    .size		= 1
  }, {
    .uvcControl		= UVC_PU_WHITE_BALANCE_TEMPERATURE_AUTO_CONTROL,
    .oaControl		= OA_CAM_CTRL_AUTO_WHITE_BALANCE_TEMP,
    .oaControlType	= OA_CAM_CTRL_TYPE_BOOLEAN,
    .size		= 1
  }, {
    .uvcControl		= UVC_PU_WHITE_BALANCE_COMPONENT_AUTO_CONTROL,
    .oaControl		= OA_CAM_CTRL_AUTO_WHITE_BALANCE,
    .oaControlType	= OA_CAM_CTRL_TYPE_BOOLEAN,
    .size		= 1
  }, {
    .uvcControl		= UVC_PU_DIGITAL_MULTIPLIER_CONTROL,
    .oaControl		= 0,
    .oaControlType	= OA_CAM_CTRL_TYPE_INT32,
    .size		= 2
  }, {
    .uvcControl		= UVC_PU_DIGITAL_MULTIPLIER_LIMIT_CONTROL,
    .oaControl		= 0,
    .oaControlType	= OA_CAM_CTRL_TYPE_INT32,
    .size		= 2
  }, {
    .uvcControl		= UVC_PU_ANALOG_VIDEO_STANDARD_CONTROL,
    .oaControl		= 0,
    .oaControlType	= OA_CAM_CTRL_TYPE_READONLY,
    .size		= 1
  }, {
    .uvcControl		= UVC_PU_ANALOG_LOCK_STATUS_CONTROL,
    .oaControl		= 0,
    .oaControlType	= OA_CAM_CTRL_TYPE_READONLY,
    .size		= 1
  }
};

unsigned int numPUControls = sizeof ( controlData ) / sizeof ( struct puCtrl );

static void _UVCInitFunctionPointers ( oaCamera* );
static void _getUVCControlValues ( oaCamera*, uvc_device_handle_t*,
    uint8_t, int );


/**
 * Initialise a given camera device
 */

oaCamera*
oaUVCInitCamera ( oaDevice* device )
{
  oaCamera*				camera;
  unsigned int                          ret, matched, haveCamera, haveUnit;
  unsigned int                          numPUControls, oaControl, i, j;
  int					deviceAddr, deviceBus, numUVCDevices;
  int					multiplier;
  unsigned short                        busNum, addr;
  uint64_t				flags;
  uvc_device_t**                        devlist;
  uvc_device_t*                         uvcDevice;
  uvc_device_descriptor_t*              desc;
  uvc_device_handle_t*			uvcHandle;
  const uvc_input_terminal_t*		inputTerminals;
  const uvc_input_terminal_t*		cameraTerminal;
  enum uvc_ct_ctrl_selector		control;
  enum uvc_pu_ctrl_selector		uvcControl;
  const uvc_processing_unit_t*		processingUnits;
  const uvc_processing_unit_t*		unit;
  const uvc_format_desc_t*		formatDescs;
  const uvc_format_desc_t*		format;
  const uvc_frame_desc_t*		frame;
  int					allFramesHaveFixedRates;

  if (!( camera = ( oaCamera* ) malloc ( sizeof ( oaCamera )))) {
    perror ( "malloc oaCamera failed" );
    return 0;
  }

  OA_CLEAR ( cameraState );
  OA_CLEAR ( camera->controls );
  OA_CLEAR ( camera->features );
  _oaInitCameraFunctionPointers ( camera );
  _UVCInitFunctionPointers ( camera );

  ( void ) strcpy ( camera->deviceName, device->deviceName );
  cameraState.initialised = 0;
  cameraState.index = -1;

  // FIX ME -- This is a bit ugly.  Much of it is repeated from the
  // getCameras function.  I should join the two together somehow.

  if ( uvc_init ( &cameraState.uvcContext, 0 ) != UVC_SUCCESS ) {
    fprintf ( stderr, "uvc_init failed\n" );
    return 0;
  }

  if ( uvc_get_device_list ( cameraState.uvcContext, &devlist ) !=
      UVC_SUCCESS ) {
    fprintf ( stderr, "uvc_get_device_list failed\n" );
    return 0;
  }
  numUVCDevices = 0;
  while ( devlist[numUVCDevices] ) { numUVCDevices++; }
  if ( numUVCDevices < 1 ) {
    uvc_free_device_list ( devlist, 1 );
    uvc_exit ( cameraState.uvcContext );
    if ( numUVCDevices ) {
      fprintf ( stderr, "Can't see any UVC devices now (list returns -1)\n" );
      free ( camera );
      return 0;
    }
    fprintf ( stderr, "Can't see any UVC devices now\n" );
    free ( camera );
    return 0;
  }

  matched = 0;
  deviceAddr = device->_devIndex & 0xff;
  deviceBus = device->_devIndex >> 8;

  for ( i = 0; i < numUVCDevices && !matched; i++ ) {
    uvcDevice = devlist[i];
    if ( uvc_get_device_descriptor ( uvcDevice, &desc )) {
      uvc_free_device_list ( devlist, 1 );
      uvc_exit ( cameraState.uvcContext );
      fprintf ( stderr, "get UVC device descriptor failed\n" );
      free ( camera );
      return 0;
    }
    if ( uvc_get_bus_number ( uvcDevice ) == deviceBus &&
        uvc_get_device_address ( uvcDevice ) == deviceAddr ) {
      // this looks like the one!
      matched = 1;
      uvc_open ( uvcDevice, &uvcHandle );
    }
  }
  uvc_free_device_list ( devlist, 1 );

  if ( !matched ) {
    fprintf ( stderr, "No matching UVC device found!\n" );
    uvc_exit ( cameraState.uvcContext );
    free ( camera );
    return 0;
  }
  if ( !uvcHandle ) {
    fprintf ( stderr, "Unable to open UVC device!\n" );
    uvc_exit ( cameraState.uvcContext );
    free ( camera );
    return 0;
  }

  if (!( inputTerminals = uvc_get_input_terminals ( uvcHandle ))) {
    fprintf ( stderr, "No input terminals found!\n" );
    uvc_exit ( cameraState.uvcContext );
    free ( camera );
    return 0;
  }

  cameraTerminal = inputTerminals;
  haveCamera = 0;
  do {
    // fprintf ( stderr, "terminal id: %d, type %d, fmax %d, fmin %d, f %d, flags %llx\n", cameraTerminal->bTerminalID, cameraTerminal->wTerminalType, cameraTerminal->wObjectiveFocalLengthMin, cameraTerminal->wObjectiveFocalLengthMax, cameraTerminal->wOcularFocalLength, cameraTerminal->bmControls );
    if ( UVC_ITT_CAMERA == cameraTerminal->wTerminalType ) {
      haveCamera = 1;
    } else {
      cameraTerminal = cameraTerminal->next;
    }
  } while ( !haveCamera && cameraTerminal );

  if ( !haveCamera ) {
    uvc_close ( uvcHandle );
    fprintf ( stderr, "Device doesn't actually look like a camera\n" );
    uvc_exit ( cameraState.uvcContext );
    free ( camera );
    return 0;
  }

  control = 1;
  flags = cameraTerminal->bmControls;
  while ( flags ) {
    if ( flags & 1 ) {

      // FIX ME -- need to know what the units are for all of these

      switch ( control ) {

        case UVC_CT_SCANNING_MODE_CONTROL:
          fprintf ( stderr, "Unsupported control CT_SCANNING_MODE_CONTROL\n" );
          break;

        case UVC_CT_AE_MODE_CONTROL:
          // FIX ME
          // UVC auto exposure mode is messy -- a bitfield of:
          // 1 = manual, 2 = auto, 4 = shutter priority, 8 = aperture priority
          // fortunately the exponents of the bit values correspond to the
          // menu values we're using
          camera->controls[ OA_CAM_CTRL_AUTO_EXPOSURE ] =
              OA_CAM_CTRL_TYPE_MENU;
          camera->min[ OA_CAM_CTRL_AUTO_EXPOSURE ] = 0;
          camera->max[ OA_CAM_CTRL_AUTO_EXPOSURE ] = 3;
          camera->step[ OA_CAM_CTRL_AUTO_EXPOSURE ] = 1;
          camera->def[ OA_CAM_CTRL_AUTO_EXPOSURE ] = 1; // an assumption
          break;

        case UVC_CT_EXPOSURE_TIME_ABSOLUTE_CONTROL:
          camera->controls[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] =
              OA_CAM_CTRL_TYPE_INT32;
          if ( uvc_get_exposure_abs ( uvcHandle,
              &( camera->min[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ]), UVC_GET_MIN ) !=
              UVC_SUCCESS ) {
            fprintf ( stderr, "failed to get min value for AE setting\n" );
          }
          if ( uvc_get_exposure_abs ( uvcHandle,
              &( camera->max[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ]), UVC_GET_MAX ) !=
              UVC_SUCCESS ) {
            fprintf ( stderr, "failed to get max value for AE setting\n" );
          }
          if ( uvc_get_exposure_abs ( uvcHandle,
              &( camera->step[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ]), UVC_GET_RES ) !=
              UVC_SUCCESS ) {
            fprintf ( stderr, "failed to get resolution for AE setting\n" );
          }
          if ( uvc_get_exposure_abs ( uvcHandle,
              &( camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ]), UVC_GET_RES ) !=
              UVC_SUCCESS ) {
            fprintf ( stderr, "failed to get default for AE setting\n" );
          }

          // now convert the values from msec to usec.

          camera->min[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] *= 1000;
          camera->max[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] *= 1000;
          camera->step[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] *= 1000;
          camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] *= 1000;

          break;

        case UVC_CT_ZOOM_ABSOLUTE_CONTROL:
        case UVC_CT_EXPOSURE_TIME_RELATIVE_CONTROL:
        case UVC_CT_AE_PRIORITY_CONTROL:
        case UVC_CT_FOCUS_ABSOLUTE_CONTROL:
        case UVC_CT_FOCUS_RELATIVE_CONTROL:
        case UVC_CT_FOCUS_AUTO_CONTROL:
        case UVC_CT_IRIS_ABSOLUTE_CONTROL:
        case UVC_CT_IRIS_RELATIVE_CONTROL:
        case UVC_CT_ZOOM_RELATIVE_CONTROL:
        case UVC_CT_PANTILT_ABSOLUTE_CONTROL:
        case UVC_CT_PANTILT_RELATIVE_CONTROL:
        case UVC_CT_ROLL_ABSOLUTE_CONTROL:
        case UVC_CT_ROLL_RELATIVE_CONTROL:
        case UVC_CT_PRIVACY_CONTROL:
          fprintf ( stderr, "Unsupported UVC control %d\n", control );
          break;

        default:
          fprintf ( stderr, "unknown UVC control %d\n", control );
      }
    }
    flags >>= 1;
    control++;
  }

  if (!( processingUnits = uvc_get_processing_units ( uvcHandle ))) {
    fprintf ( stderr, "No processing units found!\n" );
    uvc_exit ( cameraState.uvcContext );
    free ( camera );
    return 0;
  }

  unit = processingUnits;
  haveUnit = 0;
  do {
    // FIX ME -- find out how this should actually work
    // if ( cameraTerminal->bTerminalID == unit->bSourceID ) {
      haveUnit = 1;
    // } else {
    //   unit = unit->next;
    // }
  } while ( !haveUnit && unit );

  if ( !haveUnit ) {
    uvc_close ( uvcHandle );
    fprintf ( stderr, "Can't find any processing units for the camera\n" );
    uvc_exit ( cameraState.uvcContext );
    free ( camera );
    return 0;
  }

  flags = unit->bmControls;
  numPUControls = sizeof ( controlData ) / sizeof ( struct puCtrl );
  for ( i = 0; i < numPUControls; i++ ) {
    if ( flags & 1 ) {
      // FIX ME -- remove these two temp variables
      uvcControl = controlData[ i ].uvcControl;
      oaControl = controlData[ i ].oaControl;
      // if oaControl == 0 we don't support this yet, but white balance
      // component is a special case because it is red and blue balance
      // combined
      if ( oaControl || ( UVC_PU_WHITE_BALANCE_COMPONENT_CONTROL ==
          uvcControl )) {
        // FIX ME -- need to know what the default units are for all of these
        _getUVCControlValues ( camera, uvcHandle, unit->bUnitID, i );
      } else {
        fprintf ( stderr, "Unsupported UVC processing unit control %d\n",
            uvcControl );
      }
    }
    flags >>= 1;
  }

  if ( flags ) {
    fprintf ( stderr, "unknown UVC processing unit controls are present\n" );
  }

  // Now process the format descriptions...

  cameraState.videoGrey = cameraState.videoBY8 = 0;
  cameraState.videoCurrent = 0;
  cameraState.isColour = 0;
  camera->features.rawMode = camera->features.demosaicMode = 0;

  formatDescs = uvc_get_format_descs ( uvcHandle );
  format = formatDescs;
  do {
    // we're only interested (for the moment) in frame formats
    if ( UVC_VS_FORMAT_FRAME_BASED == format->bDescriptorSubtype ) {

      // and of the frame formats, we're only interested in Y800 or BY8
      // for the time being

      if ( !memcmp ( format->fourccFormat, "BY8 ", 4 )) {
        camera->features.rawMode = 1;
        cameraState.videoBY8 = 1;
        cameraState.videoCurrent = format;
        cameraState.isColour = 1;
      }

      if ( !memcmp ( format->fourccFormat, "Y800", 4 )) {
        if ( !cameraState.videoCurrent ) {
          cameraState.videoCurrent = format;
        }
      }
    }
    format = format->next;
  } while ( format );

  if ( !cameraState.videoCurrent ) {
    fprintf ( stderr, "No suitable video format found on %s",
      camera->deviceName );
    free (( void* ) camera );
    return 0;
  }

  cameraState.maxResolutionX = cameraState.maxResolutionY = 0;
  frame = cameraState.videoCurrent->frame_descs;
  allFramesHaveFixedRates = 1;
  i = 0;
  do {
    if (( cameraState.resolutions[i].x = frame->wWidth ) >
        cameraState.maxResolutionX ) {
      cameraState.maxResolutionX = cameraState.resolutions[i].x;
    }
    if (( cameraState.resolutions[i].y = frame->wHeight ) >
        cameraState.maxResolutionY ) {
      cameraState.maxResolutionY = cameraState.resolutions[i].y;
    }
    if ( !frame->bFrameIntervalType ) {
      allFramesHaveFixedRates = 0;
    }
    i++;
    frame = frame->next;
  } while ( frame );

  camera->features.frameRates = allFramesHaveFixedRates;

  cameraState.resolutions[i].x = 0;
  cameraState.resolutions[i].y = 0;

  camera->interface = device->interface;
  cameraState.uvcHandle = uvcHandle;
  cameraState.index = device->_devIndex;
  cameraState.unitId = unit->bUnitID;
  cameraState.buffers = 0;
  cameraState.configuredBuffers = 0;

  // Save a local copy for the values of red and blue balance here to
  // save having to read the combined value every time we want to
  // change it.

  cameraState.componentBalance = getUVCControl ( cameraState.uvcHandle,
      cameraState.unitId, UVC_PU_WHITE_BALANCE_COMPONENT_CONTROL, 4,
      UVC_GET_CUR );

  // The largest buffer size we should need

  multiplier = 1;
  cameraState.imageBufferLength = cameraState.maxResolutionX *
      cameraState.maxResolutionY * multiplier;
  cameraState.buffers = calloc ( UVC_BUFFERS, sizeof ( struct UVCbuffer ));
  for ( i = 0; i < UVC_BUFFERS; i++ ) {
    void* m = malloc ( cameraState.imageBufferLength );
    if ( m ) {
      cameraState.buffers[i].start = m;
      cameraState.configuredBuffers++;
    } else {
      fprintf ( stderr, "%s malloc failed\n", __FUNCTION__ );
      if ( i ) {
        for ( j = 0; j < i; j++ ) {
          free (( void* ) cameraState.buffers[j].start );
        }
      }
      return 0;
    }
  }

  pthread_mutex_init ( &cameraState.captureMutex, 0 );
  pthread_cond_init ( &cameraState.frameAvailable, 0 );

  return camera;
}


static void
_UVCInitFunctionPointers ( oaCamera* camera )
{
  camera->funcs.initCamera = oaUVCInitCamera;
  camera->funcs.closeCamera = oaUVCCloseCamera;
  camera->funcs.resetCamera = oaUVCCameraReset;
  // camera->funcs.startCapture = oaUVCCameraStart;
  camera->funcs.startCaptureExtended = oaUVCCameraStart;
  camera->funcs.stopCapture = oaUVCCameraStop;
  camera->funcs.hasLoadableFirmware = oaUVCCameraHasLoadableFirmware;
  // camera->funcs.isFirmwareLoaded = _cameraIsFirmwareLoaded;
  // camera->funcs.loadFirmware = _cameraLoadFirmware;
  camera->funcs.getName = oaUVCCameraGetName;
  camera->funcs.getControlRange = oaUVCCameraGetControlRange;
  camera->funcs.readControl = oaUVCCameraReadControl;
  camera->funcs.setControl = oaUVCCameraSetControl;
  camera->funcs.testControl = oaUVCCameraTestControl;
  // camera->funcs.hasAuto = oaHasAuto;
  // camera->funcs.isAuto = _isAuto;
  // camera->funcs.setControlMulti = _setControlMulti;
  camera->funcs.startReadFrame = oaUVCCameraStartReadFrame;
  camera->funcs.readFrame = oaUVCCameraReadFrame;
  camera->funcs.finishReadFrame = oaUVCCameraFinishReadFrame;
  camera->funcs.enumerateFrameSizes = oaUVCCameraGetFrameSizes;
  camera->funcs.enumerateFrameRates = oaUVCCameraGetFrameRates;
  camera->funcs.getFramePixelFormat = oaUVCCameraGetFramePixelFormat;
  // camera->funcs.testROISize = oaUVCCameraTestROISize;
  camera->funcs.getMenuString = oaUVCCameraGetMenuString;
  // camera->funcs.getAutoWBManualSetting = oaUVCCameraGetAutoWBManualSetting;
  camera->funcs.hasFixedFrameRates = oaUVCCameraHasFixedFrameRates;
}


static void
_getUVCControlValues ( oaCamera* camera, uvc_device_handle_t* uvcHandle,
    uint8_t unitId, int index )
{
  int len;
  unsigned int oaControl;
  enum uvc_pu_ctrl_selector uvcControl;

  uvcControl = controlData[ index ].uvcControl;
  oaControl = controlData[ index ].oaControl;
  len = controlData[ index ].size;

  // special case for white component control
  if ( UVC_PU_WHITE_BALANCE_COMPONENT_CONTROL == uvcControl ) {
    int val;

    if (( val = getUVCControl ( uvcHandle, unitId,
        UVC_PU_WHITE_BALANCE_COMPONENT_CONTROL, len, UVC_GET_MIN )) < 0 ) {
      return;
    }
    camera->min[ OA_CAM_CTRL_BLUE_BALANCE ] = val & 0xffff;
    camera->min[ OA_CAM_CTRL_RED_BALANCE ] = val >> 16;

    if (( val = getUVCControl ( uvcHandle, unitId,
        UVC_PU_WHITE_BALANCE_COMPONENT_CONTROL, len, UVC_GET_MAX )) < 0 ) {
      return;
    }
    camera->max[ OA_CAM_CTRL_BLUE_BALANCE ] = val & 0xffff;
    camera->max[ OA_CAM_CTRL_RED_BALANCE ] = val >> 16;

    if (( val = getUVCControl ( uvcHandle, unitId,
        UVC_PU_WHITE_BALANCE_COMPONENT_CONTROL, len, UVC_GET_RES )) < 0 ) {
      return;
    }
    camera->step[ OA_CAM_CTRL_BLUE_BALANCE ] = val & 0xffff;
    camera->step[ OA_CAM_CTRL_RED_BALANCE ] = val >> 16;

    if (( val = getUVCControl ( uvcHandle, unitId,
        UVC_PU_WHITE_BALANCE_COMPONENT_CONTROL, len, UVC_GET_DEF )) < 0 ) {
      return;
    }
    camera->def[ OA_CAM_CTRL_BLUE_BALANCE ] = val & 0xffff;
    camera->def[ OA_CAM_CTRL_RED_BALANCE ] = val >> 16;

    camera->controls[ OA_CAM_CTRL_BLUE_BALANCE ] = OA_CAM_CTRL_TYPE_INT32;
    camera->controls[ OA_CAM_CTRL_RED_BALANCE ] = OA_CAM_CTRL_TYPE_INT32;

  } else {

    camera->controls[ oaControl ] = controlData[ index ].oaControlType;

    switch ( camera->controls[ oaControl ] ) {
      case OA_CAM_CTRL_TYPE_INT32:
      case OA_CAM_CTRL_TYPE_MENU:
        if (( camera->min[ oaControl ] = getUVCControl ( uvcHandle, unitId,
            uvcControl, len, UVC_GET_MIN )) < 0 ) {
          return;
        }
        if (( camera->max[ oaControl ] = getUVCControl ( uvcHandle, unitId,
            uvcControl, len, UVC_GET_MAX )) < 0 ) {
          return;
        }
        if (( camera->step[ oaControl ] = getUVCControl ( uvcHandle, unitId,
            uvcControl, len, UVC_GET_RES )) < 0 ) {
          return;
        }
        if (( camera->def[ oaControl ] = getUVCControl ( uvcHandle, unitId,
            uvcControl, len, UVC_GET_DEF )) < 0 ) {
          return;
        }
        break;

      case OA_CAM_CTRL_TYPE_BOOLEAN:
        camera->min[ oaControl ] = 0;
        camera->max[ oaControl ] = 1;
        camera->step[ oaControl ] = 1;
        if (( camera->def[ oaControl ] = getUVCControl ( uvcHandle, unitId,
            uvcControl, len, UVC_GET_DEF )) < 0 ) {
          return;
        }
        break;

      case OA_CAM_CTRL_TYPE_READONLY:
        break;

      default:
        fprintf ( stderr, "unhandled control type %d in %s\n",
            camera->controls[ oaControl ], __FUNCTION__ );
        break;
    }
  }
}


int
getUVCControl ( uvc_device_handle_t* uvcHandle, uint8_t unitId,
    enum uvc_ct_ctrl_selector ctrl, int len, enum uvc_req_code req )
{
  int ret, i;
  uint8_t data[4] = { 0xde, 0xad, 0xbe, 0xef };
  unsigned int val;

  if (( ret = uvc_get_ctrl ( uvcHandle, unitId, ctrl, data, len, req )) !=
      len ) {
    fprintf ( stderr, "%s requested %d for control %d, got %d\n",
        __FUNCTION__, len, ctrl, ret );
    return ret;
  }

  val = 0;
  i = len - 1;
  do {
    val <<= 8;
    val += data[i--];
  } while ( i >= 0 );

  return val;
}


int
setUVCControl ( uvc_device_handle_t* uvcHandle, uint8_t unitId,
    enum uvc_ct_ctrl_selector ctrl, int len, int value )
{
  int ret, i;
  uint8_t data[4] = { 0x00, 0x00, 0x00, 0x00 };

  for ( i = 0; i < len; i++ ) {
    data[i] = value & 0xff;
    value >>= 8;
  }

  if (( ret = uvc_set_ctrl ( uvcHandle, unitId, ctrl, data, len )) !=
      len ) {
    fprintf ( stderr, "%s requested %d for control %d, got %d\n",
        __FUNCTION__, len, ctrl, ret );
    return ret;
  }
  return ret;
}


int
oaUVCCloseCamera ( oaCamera* camera )
{
  if ( camera ) {
    uvc_close ( cameraState.uvcHandle );
    // uvc_unref_device ( device );
  }
  return 0;
}


#endif /* HAVE_LIBUVC */
