/*****************************************************************************
 *
 * V4L2getState.c -- state querying for V4L2 cameras
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#if HAVE_LIBV4L2

#include <openastro/oacam.h>

#include "V4L2oacam.h"
#include "v4l2ioctl.h"
#include "oacamprivate.h"

#define	cameraState		camera->_v4l2

int
oaV4L2CameraGetControlRange ( oaCamera* camera, int control, int* min,
    int* max, int* step, int* def )
{
  *min = camera->min [ control ];
  *max = camera->max [ control ];
  *step = camera->step [ control ];
  *def = camera->def [ control ];
  return 0;
}


const FRAMESIZE*
oaV4L2CameraGetFrameSizes ( oaCamera* camera )
{
  return cameraState.resolutions;
}


const FRAMERATE*
oaV4L2CameraGetFrameRates ( oaCamera* camera, int resX, int resY )
{
  struct v4l2_frmivalenum fint;
  int j = 0, k;

  OA_CLEAR ( cameraState.framerates );
  for ( k = 0; j < OA_MAX_FRAMERATES; k++ ) { 
    OA_CLEAR( fint );
    fint.index = k;
    fint.pixel_format = cameraState.videoCurrent;
    fint.width = resX;
    fint.height = resY;
    if ( -1 == v4l2ioctl ( cameraState.fd, VIDIOC_ENUM_FRAMEINTERVALS,
        &fint )) { 
      if ( EINVAL == errno) { 
        break;
      } else { 
        perror("VIDIOC_ENUM_FRAMEINTERVALS");
      } 
    } 
    if ( V4L2_FRMIVAL_TYPE_DISCRETE == fint.type ) { 
      cameraState.framerates[j].numerator = fint.discrete.numerator;
      cameraState.framerates[j].denominator = fint.discrete.denominator;
      j++;
    } 
  } 

  cameraState.framerates[j].numerator = 0;
  cameraState.framerates[j].denominator = 0;

  return cameraState.framerates;
}


const char*
oaV4L2CameraGetName ( oaCamera* camera )
{
  return camera->deviceName;
}


int
oaV4L2CameraGetFramePixelFormat ( oaCamera* camera, int depth )
{
  // the driver lies for the colour TIS DxK cameras, so we fix it here
  if ( cameraState.colourDxK ) {
    return OA_PIX_FMT_GBRG8;
  }

  switch ( cameraState.videoCurrent ) {
    case V4L2_PIX_FMT_RGB24:
      return OA_PIX_FMT_RGB24;
      break;
    case V4L2_PIX_FMT_GREY:
      return OA_PIX_FMT_GREY8;
      break;
    case V4L2_PIX_FMT_SBGGR8:
      return OA_PIX_FMT_BGGR8;
    case V4L2_PIX_FMT_SRGGB8:
      return OA_PIX_FMT_RGGB8;
    case V4L2_PIX_FMT_SGRBG8:
      return OA_PIX_FMT_GRBG8;
    case V4L2_PIX_FMT_SGBRG8:
      return OA_PIX_FMT_GBRG8;
    case V4L2_PIX_FMT_YUV420:
    case V4L2_PIX_FMT_YUYV:
    case V4L2_PIX_FMT_Y16:
    default:
      fprintf ( stderr, "oaV4L2CameraGetFramePixelFormat can't handle pixel format %d\n", cameraState.videoCurrent );
      return OA_PIX_FMT_RGB24;
      break;
  }
}

#endif /* HAVE_LIBV4L2 */
