/*****************************************************************************
 *
 * ZWASIcontrol.c -- control functions for ZW ASI cameras
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>
#include <openastro/oacam.h>

#include <ASICamera.h>

#include "ZWASIoacam.h"

#define	cameraState		camera->_zwasi


int
oaZWASICameraSetControl ( oaCamera* camera, int control, int arg1 )
{
   // FIX ME -- change ranges & return OOB or unsupported control error

  oacamDebugMsg ( DEBUG_CAM_CTRL, "ZWASI: control: %s ( %d, %d )\n",
      __FUNCTION__, control, arg1 );

  switch ( control ) {

    case OA_CAM_CTRL_BRIGHTNESS:
      setValue ( CONTROL_BRIGHTNESS, arg1, cameraState.autoBrightness );
      cameraState.currentBrightness = arg1;
      break;

    case OA_CAM_CTRL_BLUE_BALANCE:
      setValue ( CONTROL_WB_B, arg1, cameraState.autoBlueBalance );
      cameraState.currentBlueBalance = arg1;
      break;

    case OA_CAM_CTRL_RED_BALANCE:
      setValue ( CONTROL_WB_R, arg1, cameraState.autoRedBalance );
      cameraState.currentRedBalance = arg1;
      break;

    case OA_CAM_CTRL_GAMMA:
      setValue ( CONTROL_GAMMA, arg1, cameraState.autoGamma );
      cameraState.currentGamma = arg1;
      break;

    case OA_CAM_CTRL_GAIN:
      setValue ( CONTROL_GAIN, arg1, cameraState.autoGain );
      cameraState.currentGain = arg1;
      break;

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      setValue ( CONTROL_EXPOSURE, arg1, cameraState.autoExposure );
      cameraState.currentAbsoluteExposure = arg1;
      break;

    case OA_CAM_CTRL_USBTRAFFIC:
      setValue ( CONTROL_BANDWIDTHOVERLOAD, arg1, cameraState.autoUSBTraffic );
      cameraState.currentUSBTraffic = arg1;
      break;

    case OA_CAM_CTRL_BINNING:
      // This will take effect when the resolution is switched
      cameraState.binMode = arg1;
      if ( arg1 < OA_BIN_MODE_2x2 && 8 == cameraState.currentBitDepth ) {
        cameraState.videoCurrent = cameraState.next8BitMode;
      }
      break;

    case OA_CAM_CTRL_HFLIP:
      cameraState.currentHFlip = arg1;
      SetMisc ( cameraState.currentHFlip, cameraState.currentVFlip );
      break;

    case OA_CAM_CTRL_VFLIP:
      cameraState.currentVFlip = arg1;
      SetMisc ( cameraState.currentHFlip, cameraState.currentVFlip );
      break;

    case OA_CAM_CTRL_BIT_DEPTH:
      // if we're a colour camera here and we're going to 8-bit then
      // unless we're in raw mode we really want to go back to RGB24.
      // Unfortunately because theres no 16-bit RGB mode we don't
      // know if we were using raw mode or not.  So we have to save
      // the value to come back to later
      if ( 16 == arg1 || 12 == arg1 ) {
        cameraState.currentBitDepth = 16;
        cameraState.next8BitMode = cameraState.videoCurrent;
        cameraState.videoCurrent = IMG_RAW16;
      } else {
        if ( 8 == arg1 ) {
          cameraState.currentBitDepth = 8;
          if ( IMG_RGB24 != cameraState.videoCurrent ) {
            cameraState.videoCurrent = cameraState.next8BitMode;
          }
        }
      }
      break;

    case OA_CAM_CTRL_COLOUR_MODE:
      // This is a bit messy.  If we're in 16-bit mode then you get RAW
      // anyhow.  There's no 16-bit RGB.  So if we're setting NONRAW
      // and we're in 16-bit, just ignore it.  Also see the comments for
      // OA_CAM_CTRL_BIT_DEPTH regarding 8-bit mode.
      if ( OA_COLOUR_MODE_RAW == arg1 ) {
        cameraState.next8BitMode = IMG_RAW8;
        if ( 16 == cameraState.currentBitDepth ) {
          cameraState.videoCurrent = IMG_RAW16;
        } else {
          cameraState.videoCurrent = IMG_RAW8;
        }
      } else {
        cameraState.next8BitMode = IMG_RGB24;
        if ( 16 != cameraState.currentBitDepth ) {
          cameraState.videoCurrent = IMG_RGB24;
        }
      }
      break;

    case OA_CAM_CTRL_AUTO_GAIN:
      setValue ( CONTROL_GAIN, cameraState.currentGain, arg1 );
      cameraState.autoGain = arg1;
      break;

    case OA_CAM_CTRL_AUTO_GAMMA:
      setValue ( CONTROL_GAMMA, cameraState.currentGamma, arg1 );
      cameraState.autoGamma = arg1;
      break;

    case OA_CAM_CTRL_AUTO_BRIGHTNESS:
      setValue ( CONTROL_BRIGHTNESS, cameraState.currentBrightness, arg1 );
      cameraState.autoBrightness = arg1;
      break;

    case OA_CAM_CTRL_AUTO_EXPOSURE:
    {
      int val = ( OA_EXPOSURE_AUTO == arg1 ? 1 : 0 );
      setValue ( CONTROL_EXPOSURE, cameraState.currentAbsoluteExposure, val );
      cameraState.autoExposure = val;
      break;
    }

    case OA_CAM_CTRL_AUTO_RED_BALANCE:
      setValue ( CONTROL_WB_R, cameraState.currentRedBalance, arg1 );
      cameraState.autoRedBalance = arg1;
      break;

    case OA_CAM_CTRL_AUTO_BLUE_BALANCE:
      setValue ( CONTROL_WB_B, cameraState.currentBlueBalance, arg1 );
      cameraState.autoBlueBalance = arg1;
      break;

    case OA_CAM_CTRL_AUTO_USBTRAFFIC:
      setValue ( CONTROL_BANDWIDTHOVERLOAD, cameraState.currentUSBTraffic,
          arg1 );
      cameraState.autoUSBTraffic = arg1;
      break;

    default:
      fprintf ( stderr, "Unrecognised control %d in %s\n", control,
          __FUNCTION__ );
      return 0;
      break;
  }
  return -1;
}


int
oaZWASICameraReadControl ( oaCamera* camera, int control )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "ZWASI: control: %s ( %d )\n",
      __FUNCTION__, control );

  switch ( control ) {

    case OA_CAM_CTRL_BRIGHTNESS:
      return cameraState.currentBrightness;
      break;

    case OA_CAM_CTRL_BLUE_BALANCE:
      return cameraState.currentBlueBalance;
      break;

    case OA_CAM_CTRL_RED_BALANCE:
      return cameraState.currentRedBalance;
      break;

    case OA_CAM_CTRL_GAMMA:
      return cameraState.currentGamma;
      break;

    case OA_CAM_CTRL_GAIN:
      return cameraState.currentGain;
      break;

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      return cameraState.currentAbsoluteExposure;
      break;

    case OA_CAM_CTRL_USBTRAFFIC:
      return cameraState.currentUSBTraffic;
      break;

    case OA_CAM_CTRL_BINNING:
      return cameraState.binMode;
      break;

    case OA_CAM_CTRL_HFLIP:
      return cameraState.currentHFlip;
      break;

    case OA_CAM_CTRL_VFLIP:
      return cameraState.currentVFlip;
      break;

    case OA_CAM_CTRL_BIT_DEPTH:
      return cameraState.currentBitDepth;
      break;

    case OA_CAM_CTRL_COLOUR_MODE:
      if ( IMG_RAW16 == cameraState.videoCurrent || IMG_RAW8 ==
          cameraState.videoCurrent ) {
        return OA_COLOUR_MODE_RAW;
      }
      return OA_COLOUR_MODE_NONRAW;
      break;

    case OA_CAM_CTRL_TEMPERATURE:
    {
      float temp = getSensorTemp();
      int tempx10 = ( temp * 10.0 );
      return tempx10;
      break;
    }

    case OA_CAM_CTRL_AUTO_GAIN:
      return cameraState.autoGain;
      break;

    case OA_CAM_CTRL_AUTO_GAMMA:
      return cameraState.autoGamma;
      break;

    case OA_CAM_CTRL_AUTO_BRIGHTNESS:
      return cameraState.autoBrightness;
      break;

    case OA_CAM_CTRL_AUTO_EXPOSURE:
      return cameraState.autoExposure;
      break;

    case OA_CAM_CTRL_AUTO_RED_BALANCE:
      return cameraState.autoRedBalance;
      break;

    case OA_CAM_CTRL_AUTO_BLUE_BALANCE:
      return cameraState.autoBlueBalance;
      break;

    case OA_CAM_CTRL_AUTO_USBTRAFFIC:
      return cameraState.autoUSBTraffic;
      break;

    case OA_CAM_CTRL_DROPPED:
      return getDroppedFrames();
      break;

    default:
      fprintf ( stderr,
          "Unrecognised control %d in %s\n", control, __FUNCTION__ );
      return 0;
      break;
  }
  return -1;
}


int
oaZWASICameraTestControl ( oaCamera* camera, int control, int arg1 )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "ZWASI: control: %s ( %d, %d )\n",
      __FUNCTION__, control, arg1 );

  switch ( control ) {

    case OA_CAM_CTRL_BRIGHTNESS:
    case OA_CAM_CTRL_BLUE_BALANCE:
    case OA_CAM_CTRL_RED_BALANCE:
    case OA_CAM_CTRL_GAMMA:
    case OA_CAM_CTRL_GAIN:
    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
    case OA_CAM_CTRL_USBTRAFFIC:
      if ( arg1 >= camera->min[ control ] &&
          arg1 <= camera->max[ control ] &&
          ( 0 == ( arg1 - camera->min[ control ]) %
          camera->step[ control ])) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CAM_CTRL_BINNING:
      if ( isBinSupported ( arg1 )) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CAM_CTRL_BIT_DEPTH:

      // This may be a bit of an ugly assumption, but I think it should hold
      // for the time being

      if ( camera->controls[ OA_CAM_CTRL_BIT_DEPTH ] ) {
        if ( 16 == arg1 || 12 == arg1 || 8 == arg1 ) {
          return OA_ERR_NONE;
        }
      }
      if ( 8 == arg1 ) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CAM_CTRL_COLOUR_MODE:
      if (( camera->features.rawMode && OA_COLOUR_MODE_RAW == arg1 ) ||
          ( camera->features.demosaicMode && OA_COLOUR_MODE_NONRAW == arg1 )) {
        return OA_ERR_NONE;
      }
      break;

    // This lot are all boolean, so we'll take any value
    case OA_CAM_CTRL_HFLIP:
    case OA_CAM_CTRL_VFLIP:
    case OA_CAM_CTRL_AUTO_GAIN:
    case OA_CAM_CTRL_AUTO_GAMMA:
    case OA_CAM_CTRL_AUTO_BRIGHTNESS:
    case OA_CAM_CTRL_AUTO_EXPOSURE:
    case OA_CAM_CTRL_AUTO_RED_BALANCE:
    case OA_CAM_CTRL_AUTO_BLUE_BALANCE:
    case OA_CAM_CTRL_AUTO_USBTRAFFIC:
      return OA_ERR_NONE;
      break;

    default:
      // If we reach here it's because we don't recognise the control
      // FIX ME -- make this debug only
      fprintf ( stderr, "Unrecognised control %d in oaZWASICameraSetControl\n",
          control );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }
  // And if we reach here it's because the value wasn't valid
  return -OA_ERR_OUT_OF_RANGE;
}


int
oaZWASICameraSetROI ( oaCamera* camera, int x, int y )
{
fprintf ( stderr, "%s called unexpectedly\n", __FUNCTION__ );
  cameraState.xSize = x;
  cameraState.ySize = y;
  return 0;
}


int
oaZWASICameraStart ( oaCamera* camera, START_PARMS* parms )
{
  int multiplier;

  oacamDebugMsg ( DEBUG_CAM_CTRL, "ZWASI: control: %s ([ %d, %d ])\n",
      __FUNCTION__, parms->size.x, parms->size.y );

  cameraState.xSize = parms->size.x;
  cameraState.ySize = parms->size.y;
  // RGB colour is 3 bytes per pixel, mono one for 8-bit, two for 16-bit,
  // RAW is one for 8-bit, 2 for 16-bit
  multiplier = ( IMG_RGB24 == cameraState.videoCurrent ) ? 3 :
      ( IMG_RAW16 == cameraState.videoCurrent ) ? 2 : 1;
  cameraState.imageBufferLength = parms->size.x * parms->size.y * multiplier;
  cameraState.lastUsedBuffer = -1;

  /*
   * This is now done by assigning the largest possible buffers when
   * the camera is initialised, but could be changed back to here to
   * use meory more efficiently.
   *
  cameraState.configuredBuffers = 0;
  cameraState.buffers = calloc ( ZWASI_BUFFERS, sizeof ( struct ZWASIbuffer ));
  for ( i = 0; i < ZWASI_BUFFERS; i++ ) {
    void* m = malloc ( cameraState.imageBufferLength );
    if ( m ) {
      cameraState.buffers[i].start = m;
      cameraState.configuredBuffers++;
    } else {
      fprintf ( stderr, "oaZWASICameraStart malloc failed\n" );
      if ( i ) {
        for ( j = 0; i < i; j++ ) {
          free (( void* ) cameraState.buffers[j].start );
        }
      }
      return -1;
    }
  }
  */

  setImageFormat ( cameraState.xSize, cameraState.ySize, cameraState.binMode,
      cameraState.videoCurrent );
  if ( OA_BIN_MODE_NONE == cameraState.binMode &&
      ( cameraState.xSize != cameraState.maxResolutionX ||
      cameraState.ySize != cameraState.maxResolutionY )) {
    setStartPos (( cameraState.maxResolutionX - cameraState.xSize ) / 2,
        ( cameraState.maxResolutionY - cameraState.ySize ) / 2 );
  }
  setValue ( CONTROL_BRIGHTNESS, cameraState.currentBrightness, 0 );
  setValue ( CONTROL_WB_B, cameraState.currentBlueBalance, 0 );
  setValue ( CONTROL_WB_R, cameraState.currentRedBalance, 0 );
  setValue ( CONTROL_GAMMA, cameraState.currentGamma, 0 );
  setValue ( CONTROL_GAIN, cameraState.currentGain, 0 );
  setValue ( CONTROL_EXPOSURE, cameraState.currentAbsoluteExposure, 0 );
  setValue ( CONTROL_BANDWIDTHOVERLOAD, cameraState.currentUSBTraffic, 0 );

  SetMisc ( cameraState.currentHFlip, cameraState.currentVFlip );

  startCapture();
  return 0;
}


int
oaZWASICameraStop ( oaCamera* camera )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "ZWASI: command: %s ()\n", __FUNCTION__ );

  stopCapture();

  /*
   * This will be needed if the buffer assignment is restored in
   * CameraStart()
   *
  for ( i = 0; i < cameraState.configuredBuffers; i++ ) {
    free (( void* ) cameraState.buffers[i].start );
  }
  cameraState.configuredBuffers = 0;
  cameraState.lastUsedBuffer = -1;
  free (( void* ) cameraState.buffers );
  cameraState.buffers = 0;
   */

  return 0;
}


// FIX ME -- should this return value be the other way around?
int
oaZWASICameraReset ( oaCamera* camera )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "ZWASI: control: %s ()\n", __FUNCTION__ );

  stopCapture();
  closeCamera();
  if ( !openCamera ( cameraState.index )) {
    fprintf ( stderr, "cannot open video device %d\n", cameraState.index );
    return -1;
  }
  // FIX ME -- check return code
  initCamera();
  return 0;
}
