/*****************************************************************************
 *
 * ZWASIinit.c -- Initialise ZW ASI cameras
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>
#include <openastro/oacam.h>

#include <ASICamera.h>

#include "ZWASIoacam.h"
#include "oacamprivate.h"
#include "unimplemented.h"

#define	cameraState		camera->_zwasi

static void _ZWASIInitFunctionPointers ( oaCamera* );

/**
 * Initialise a given camera device
 */

oaCamera*
oaZWASIInitCamera ( oaDevice* device )
{
  oaCamera*			camera;
  int                   	i, j, n, multiplier;

  oacamDebugMsg ( DEBUG_CAM_INIT, "ZWASI: init: %s ()\n", __FUNCTION__ );

  if (!( camera = ( oaCamera* ) malloc ( sizeof ( oaCamera )))) {
    perror ( "malloc oaCamera failed" );
    return 0;
  }
  OA_CLEAR ( cameraState );

  ( void ) strcpy ( camera->deviceName, device->deviceName );
  cameraState.initialised = 0;
  cameraState.index = -1;

  if ( !openCamera ( device->_devIndex )) {
    fprintf ( stderr, "open of camera %ld failed\n", device->_devIndex );
    return 0;
  }

  camera->interface = device->interface;
  cameraState.index = device->_devIndex;
  cameraState.cameraType = device->_devType;

  // FIX ME -- should check from scratch for the genuine default values
  // as the current code isn't reliable for warm-started cameras

  OA_CLEAR ( camera->controls );
  OA_CLEAR ( camera->features );
  _oaInitCameraFunctionPointers ( camera );
  _ZWASIInitFunctionPointers ( camera );

  // bool here comes from the ASI header file
  if ( isAvailable ( CONTROL_GAIN )) {
    bool autoEnabled;
    camera->controls[ OA_CAM_CTRL_GAIN ] = OA_CAM_CTRL_TYPE_INT32;
    camera->min[ OA_CAM_CTRL_GAIN ] = getMin ( CONTROL_GAIN );
    camera->max[ OA_CAM_CTRL_GAIN ] = getMax ( CONTROL_GAIN );
    camera->step[ OA_CAM_CTRL_GAIN ] = 1;
    camera->def[ OA_CAM_CTRL_GAIN ] = getValue ( CONTROL_GAIN,
        &autoEnabled );
    if ( camera->def[ OA_CAM_CTRL_GAIN ] <
        camera->min[ OA_CAM_CTRL_GAIN ] ) {
      camera->def[ OA_CAM_CTRL_GAIN ] =
          camera->min[ OA_CAM_CTRL_GAIN ];
    }
    if ( camera->def[ OA_CAM_CTRL_GAIN ] >
        camera->max[ OA_CAM_CTRL_GAIN ] ) {
      camera->def[ OA_CAM_CTRL_GAIN ] =
          camera->max[ OA_CAM_CTRL_GAIN ];
    }
    cameraState.currentGain = camera->def[ OA_CAM_CTRL_GAIN ];
    if ( isAutoSupported ( CONTROL_GAIN )) {
      camera->controls[ OA_CAM_CTRL_AUTO_GAIN ] = OA_CAM_CTRL_TYPE_BOOLEAN;
      camera->min[ OA_CAM_CTRL_AUTO_GAIN ] = 0;
      camera->max[ OA_CAM_CTRL_AUTO_GAIN ] = 1;
      camera->step[ OA_CAM_CTRL_AUTO_GAIN ] = 1;
      camera->def[ OA_CAM_CTRL_AUTO_GAIN ] = cameraState.autoGain =
          autoEnabled;
    }
  }

  if ( isAvailable ( CONTROL_EXPOSURE )) {
    bool autoEnabled;
    camera->controls[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = OA_CAM_CTRL_TYPE_INT32;
    camera->min[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] =
        getMin ( CONTROL_EXPOSURE );
    camera->max[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] =
        getMax ( CONTROL_EXPOSURE );
    camera->step[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 1;
    camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] =
        getValue ( CONTROL_EXPOSURE, &autoEnabled );
    if ( camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] <
        camera->min[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] ) {
      camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] =
          camera->min[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ];
    }
    if ( camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] >
        camera->max[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] ) {
      camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] =
          camera->max[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ];
    }
    cameraState.currentAbsoluteExposure =
        camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ];
    if ( isAutoSupported ( CONTROL_EXPOSURE )) {
      camera->controls[ OA_CAM_CTRL_AUTO_EXPOSURE ] = OA_CAM_CTRL_TYPE_BOOLEAN;
      camera->min[ OA_CAM_CTRL_AUTO_EXPOSURE ] = 0;
      camera->max[ OA_CAM_CTRL_AUTO_EXPOSURE ] = 1;
      camera->step[ OA_CAM_CTRL_AUTO_EXPOSURE ] = 1;
      camera->def[ OA_CAM_CTRL_AUTO_EXPOSURE ] = cameraState.autoExposure =
          autoEnabled;
    }
  }

  if ( isAvailable ( CONTROL_GAMMA )) {
    bool autoEnabled;
    camera->controls[ OA_CAM_CTRL_GAMMA ] = OA_CAM_CTRL_TYPE_INT32;
    camera->min[ OA_CAM_CTRL_GAMMA ] = getMin ( CONTROL_GAMMA );
    camera->max[ OA_CAM_CTRL_GAMMA ] = getMax ( CONTROL_GAMMA );
    camera->step[ OA_CAM_CTRL_GAMMA ] = 1;
    camera->def[ OA_CAM_CTRL_GAMMA ] = getValue ( CONTROL_GAMMA,
        &autoEnabled );
    if ( camera->def[ OA_CAM_CTRL_GAMMA ] <
        camera->min[ OA_CAM_CTRL_GAMMA ] ) {
      camera->def[ OA_CAM_CTRL_GAMMA ] =
          camera->min[ OA_CAM_CTRL_GAMMA ];
    }
    if ( camera->def[ OA_CAM_CTRL_GAMMA ] >
        camera->max[ OA_CAM_CTRL_GAMMA ] ) {
      camera->def[ OA_CAM_CTRL_GAMMA ] =
          camera->max[ OA_CAM_CTRL_GAMMA ];
    }
    cameraState.currentGamma = camera->def[ OA_CAM_CTRL_GAMMA ];
    if ( isAutoSupported ( CONTROL_GAMMA )) {
      camera->controls[ OA_CAM_CTRL_AUTO_GAMMA ] = OA_CAM_CTRL_TYPE_BOOLEAN;
      camera->min[ OA_CAM_CTRL_AUTO_GAMMA ] = 0;
      camera->max[ OA_CAM_CTRL_AUTO_GAMMA ] = 1;
      camera->step[ OA_CAM_CTRL_AUTO_GAMMA ] = 1;
      camera->def[ OA_CAM_CTRL_AUTO_GAMMA ] = cameraState.autoGamma =
          autoEnabled;
    }
  }

  if ( isAvailable ( CONTROL_WB_R )) {
    bool autoEnabled;
    camera->controls[ OA_CAM_CTRL_RED_BALANCE ] = OA_CAM_CTRL_TYPE_INT32;
    camera->min[ OA_CAM_CTRL_RED_BALANCE ] = getMin ( CONTROL_WB_R );
    camera->max[ OA_CAM_CTRL_RED_BALANCE ] = getMax ( CONTROL_WB_R );
    camera->step[ OA_CAM_CTRL_RED_BALANCE ] = 1;
    camera->def[ OA_CAM_CTRL_RED_BALANCE ] = getValue ( CONTROL_WB_R,
        &autoEnabled );
    if ( camera->def[ OA_CAM_CTRL_RED_BALANCE ] <
        camera->min[ OA_CAM_CTRL_RED_BALANCE ] ) {
      camera->def[ OA_CAM_CTRL_RED_BALANCE ] =
          camera->min[ OA_CAM_CTRL_RED_BALANCE ];
    }
    if ( camera->def[ OA_CAM_CTRL_RED_BALANCE ] >
        camera->max[ OA_CAM_CTRL_RED_BALANCE ] ) {
      camera->def[ OA_CAM_CTRL_RED_BALANCE ] =
          camera->max[ OA_CAM_CTRL_RED_BALANCE ];
    }
    cameraState.currentRedBalance = camera->def[ OA_CAM_CTRL_RED_BALANCE ];
    if ( isAutoSupported ( CONTROL_WB_R )) {
      camera->controls[ OA_CAM_CTRL_AUTO_RED_BALANCE ] =
          OA_CAM_CTRL_TYPE_BOOLEAN;
      camera->min[ OA_CAM_CTRL_AUTO_RED_BALANCE ] = 0;
      camera->max[ OA_CAM_CTRL_AUTO_RED_BALANCE ] = 1;
      camera->step[ OA_CAM_CTRL_AUTO_RED_BALANCE ] = 1;
      camera->def[ OA_CAM_CTRL_AUTO_RED_BALANCE ] =
          cameraState.autoRedBalance = autoEnabled;
    }
  }

  if ( isAvailable ( CONTROL_WB_B )) {
    bool autoEnabled;
    camera->controls[ OA_CAM_CTRL_BLUE_BALANCE ] = OA_CAM_CTRL_TYPE_INT32;
    camera->min[ OA_CAM_CTRL_BLUE_BALANCE ] = getMin ( CONTROL_WB_B );
    camera->max[ OA_CAM_CTRL_BLUE_BALANCE ] = getMax ( CONTROL_WB_B );
    camera->step[ OA_CAM_CTRL_BLUE_BALANCE ] = 1;
    camera->def[ OA_CAM_CTRL_BLUE_BALANCE ] = getValue ( CONTROL_WB_B,
        &autoEnabled );
    if ( camera->def[ OA_CAM_CTRL_BLUE_BALANCE ] <
        camera->min[ OA_CAM_CTRL_BLUE_BALANCE ] ) {
      camera->def[ OA_CAM_CTRL_BLUE_BALANCE ] =
          camera->min[ OA_CAM_CTRL_BLUE_BALANCE ];
    }
    if ( camera->def[ OA_CAM_CTRL_BLUE_BALANCE ] >
        camera->max[ OA_CAM_CTRL_BLUE_BALANCE ] ) {
      camera->def[ OA_CAM_CTRL_BLUE_BALANCE ] =
          camera->max[ OA_CAM_CTRL_BLUE_BALANCE ];
    }
    cameraState.currentBlueBalance =
        camera->def[ OA_CAM_CTRL_BLUE_BALANCE ];
    if ( isAutoSupported ( CONTROL_WB_B )) {
      camera->controls[ OA_CAM_CTRL_AUTO_BLUE_BALANCE ] =
        OA_CAM_CTRL_TYPE_BOOLEAN;
      camera->min[ OA_CAM_CTRL_AUTO_BLUE_BALANCE ] = 0;
      camera->max[ OA_CAM_CTRL_AUTO_BLUE_BALANCE ] = 1;
      camera->step[ OA_CAM_CTRL_AUTO_BLUE_BALANCE ] = 1;
      camera->def[ OA_CAM_CTRL_AUTO_BLUE_BALANCE ] =
          cameraState.autoBlueBalance = autoEnabled;
    }
  }

  if ( isAvailable ( CONTROL_BRIGHTNESS )) {
    bool autoEnabled;
    camera->controls[ OA_CAM_CTRL_BRIGHTNESS ] = OA_CAM_CTRL_TYPE_INT32;
    camera->min[ OA_CAM_CTRL_BRIGHTNESS ] = getMin ( CONTROL_BRIGHTNESS );
    camera->max[ OA_CAM_CTRL_BRIGHTNESS ] = getMax ( CONTROL_BRIGHTNESS );
    camera->step[ OA_CAM_CTRL_BRIGHTNESS ] = 1;
    camera->def[ OA_CAM_CTRL_BRIGHTNESS ] = getValue ( CONTROL_BRIGHTNESS,
        &autoEnabled );
    if ( camera->def[ OA_CAM_CTRL_BRIGHTNESS ] <
        camera->min[ OA_CAM_CTRL_BRIGHTNESS ] ) {
      camera->def[ OA_CAM_CTRL_BRIGHTNESS ] =
        camera->min[ OA_CAM_CTRL_BRIGHTNESS ];
    }
    if ( camera->def[ OA_CAM_CTRL_BRIGHTNESS ] >
        camera->max[ OA_CAM_CTRL_BRIGHTNESS ] ) {
      camera->def[ OA_CAM_CTRL_BRIGHTNESS ] =
          camera->max[ OA_CAM_CTRL_BRIGHTNESS ];
    }
    cameraState.currentBrightness = camera->def[ OA_CAM_CTRL_BRIGHTNESS ];
    if ( isAutoSupported ( CONTROL_BRIGHTNESS )) {
      camera->controls[ OA_CAM_CTRL_AUTO_BRIGHTNESS ] =
          OA_CAM_CTRL_TYPE_BOOLEAN;
      camera->min[ OA_CAM_CTRL_AUTO_BRIGHTNESS ] = 0;
      camera->max[ OA_CAM_CTRL_AUTO_BRIGHTNESS ] = 1;
      camera->step[ OA_CAM_CTRL_AUTO_BRIGHTNESS ] = 1;
      camera->def[ OA_CAM_CTRL_AUTO_BRIGHTNESS ] =
          cameraState.autoBrightness = autoEnabled;
    }
  }

  if ( isAvailable ( CONTROL_BANDWIDTHOVERLOAD )) {
    bool autoEnabled;
    camera->controls[ OA_CAM_CTRL_USBTRAFFIC ] = OA_CAM_CTRL_TYPE_INT32;
    camera->min[ OA_CAM_CTRL_USBTRAFFIC ] =
        getMin ( CONTROL_BANDWIDTHOVERLOAD );
    camera->max[ OA_CAM_CTRL_USBTRAFFIC ] =
        getMax ( CONTROL_BANDWIDTHOVERLOAD );
    camera->step[ OA_CAM_CTRL_USBTRAFFIC ] = 1;
    camera->def[ OA_CAM_CTRL_USBTRAFFIC ] =
        getValue ( CONTROL_BANDWIDTHOVERLOAD, &autoEnabled );
    if ( camera->def[ OA_CAM_CTRL_USBTRAFFIC ] <
        camera->min[ OA_CAM_CTRL_USBTRAFFIC ] ) {
      camera->def[ OA_CAM_CTRL_USBTRAFFIC ] =
        camera->min[ OA_CAM_CTRL_USBTRAFFIC ];
    }
    if ( camera->def[ OA_CAM_CTRL_USBTRAFFIC ] >
        camera->max[ OA_CAM_CTRL_USBTRAFFIC ] ) {
      camera->def[ OA_CAM_CTRL_USBTRAFFIC ] =
          camera->max[ OA_CAM_CTRL_USBTRAFFIC ];
    }
    cameraState.currentUSBTraffic = camera->def[ OA_CAM_CTRL_USBTRAFFIC ];
    if ( isAutoSupported ( CONTROL_BANDWIDTHOVERLOAD )) {
      camera->controls[ OA_CAM_CTRL_AUTO_USBTRAFFIC ] =
          OA_CAM_CTRL_TYPE_BOOLEAN;
      camera->min[ OA_CAM_CTRL_AUTO_USBTRAFFIC ] = 0;
      camera->max[ OA_CAM_CTRL_AUTO_USBTRAFFIC ] = 1;
      camera->step[ OA_CAM_CTRL_AUTO_USBTRAFFIC ] = 1;
      camera->def[ OA_CAM_CTRL_AUTO_USBTRAFFIC ] =
          cameraState.autoUSBTraffic = autoEnabled;
    }
  }

  cameraState.maxResolutionX = getMaxWidth();
  cameraState.maxResolutionY = getMaxHeight();

  // FIX ME -- could merge this with the resolution settings
  if ( isBinSupported ( 2 ) || isBinSupported ( 3 ) || isBinSupported ( 4 )) {
    camera->controls[ OA_CAM_CTRL_BINNING ] = OA_CAM_CTRL_TYPE_DISCRETE;
  }
  camera->controls[ OA_CAM_CTRL_DROPPED ] = OA_CAM_CTRL_TYPE_READONLY;

  // These appear to be supported by all cameras (I think )

  camera->controls[ OA_CAM_CTRL_HFLIP ] = OA_CAM_CTRL_TYPE_BOOLEAN;
  camera->min[ OA_CAM_CTRL_HFLIP ] = 0;
  camera->max[ OA_CAM_CTRL_HFLIP ] = 1;
  camera->step[ OA_CAM_CTRL_HFLIP ] = 1;
  camera->def[ OA_CAM_CTRL_HFLIP ] = cameraState.currentHFlip = 0;

  camera->controls[ OA_CAM_CTRL_VFLIP ] = OA_CAM_CTRL_TYPE_BOOLEAN;
  camera->min[ OA_CAM_CTRL_VFLIP ] = 0;
  camera->max[ OA_CAM_CTRL_VFLIP ] = 1;
  camera->step[ OA_CAM_CTRL_VFLIP ] = 1;
  camera->def[ OA_CAM_CTRL_VFLIP ] = cameraState.currentVFlip = 0;

  camera->controls[ OA_CAM_CTRL_TEMPERATURE ] = OA_CAM_CTRL_TYPE_READONLY;

  // All cameras support ROI according to Sam@ZWO
  camera->features.ROI = 1;

  // Now there are other controls and features that we need to support too

  // FIX ME
  // some of these could probably be set generically. eg.
  // if a colour cam supports RAW8 or RAW16 then rawMode = 1;
  // if a colour cam supports RGB24 then demosaicMode = 1;
  // if a colour cam supports both then set up OA_CAM_CTRL_COLOUR_MODE
  // if both RAW8 & RAW16 are supported set up OA_CAM_CTRL_BIT_DEPTH

  switch ( cameraState.cameraType ) {

    case ZWOCAM_ASI130MM:
      break;

    case ZWOCAM_ASI120MC:
      camera->controls[ OA_CAM_CTRL_COLOUR_MODE ] = OA_CAM_CTRL_TYPE_DISCRETE;
      camera->features.rawMode = 1;
      camera->features.demosaicMode = 1;
    case ZWOCAM_ASI120MM:
      camera->controls[ OA_CAM_CTRL_BIT_DEPTH ] = OA_CAM_CTRL_TYPE_DISCRETE;
      break;

    case ZWOCAM_ASI035MM:
      break;

    case ZWOCAM_ASI035MC:
      camera->controls[ OA_CAM_CTRL_COLOUR_MODE ] = OA_CAM_CTRL_TYPE_DISCRETE;
      camera->features.rawMode = 1;
      camera->features.demosaicMode = 1;
      break;

    case ZWOCAM_ASI030MC:
      camera->controls[ OA_CAM_CTRL_COLOUR_MODE ] = OA_CAM_CTRL_TYPE_DISCRETE;
      camera->features.rawMode = 1;
      camera->features.demosaicMode = 1;
      break;

    case ZWOCAM_ASI034MC:
      camera->controls[ OA_CAM_CTRL_COLOUR_MODE ] = OA_CAM_CTRL_TYPE_DISCRETE;
      camera->features.rawMode = 1;
      camera->features.demosaicMode = 1;
      break;

    default:
      fprintf ( stderr, "unknown camera type %d. Guessing feature support\n",
          cameraState.cameraType );
      if ( isColorCam()) {
        camera->controls[ OA_CAM_CTRL_COLOUR_MODE ] = OA_CAM_CTRL_TYPE_DISCRETE;
        camera->features.rawMode = 1;
        camera->features.demosaicMode = 1;
      }
      if (( isImgTypeSupported ( IMG_RGB24 ) ||
          isImgTypeSupported ( IMG_RAW8 )) &&
          isImgTypeSupported ( IMG_RAW16 )) {
        camera->controls[ OA_CAM_CTRL_BIT_DEPTH ] = OA_CAM_CTRL_TYPE_DISCRETE;
      }
      break;
  }

  // Ok, now we need to find out what frame formats are supported and
  // which one we want to use

  cameraState.videoRGB24 = cameraState.videoGrey16 = cameraState.videoGrey = 0;
  cameraState.videoCurrent = -1;
  // The mono ASI120MM will do RGB24 as a greyscale RGB image if we ask it
  // to, but that's rather wasteful, so we only support this for colour
  // cameras
  if ( isColorCam()) {
    if ( isImgTypeSupported ( IMG_RGB24 )) {
      cameraState.videoCurrent = IMG_RGB24;
      cameraState.videoRGB24 = 1;
    }
  } else {
    if ( isImgTypeSupported ( IMG_RAW8 )) { // prefer 8-bit to 16-bit initially
      cameraState.videoCurrent = IMG_RAW8;
      cameraState.videoGrey = 1;
    }
    if ( isImgTypeSupported ( IMG_RAW16 )) {
      if ( cameraState.videoCurrent < 0 ) {
        cameraState.videoCurrent = IMG_RAW16;
      }
      cameraState.videoGrey16 = 1;
    }
    if ( isImgTypeSupported ( IMG_Y8 )) {
      if ( cameraState.videoCurrent < 0 ) {
        cameraState.videoCurrent = IMG_Y8;
      }
      cameraState.videoGrey = 1;
    }
  }

  if ( -1 == cameraState.videoCurrent ) {
    fprintf ( stderr, "No suitable video format found on camera %d",
        cameraState.index );
    free (( void* ) camera );
    return 0;
  }

  cameraState.next8BitMode = cameraState.videoCurrent;
  cameraState.binMode = OA_BIN_MODE_NONE;

  // These resolutions are only valid without binning

  n = 0;
  switch ( cameraState.cameraType ) {
    case ZWOCAM_ASI130MM:
      cameraState.resolutions[n].x = 1280;
      cameraState.resolutions[n++].y = 1024;
      cameraState.resolutions[n].x = 1280;
      cameraState.resolutions[n++].y = 600;
      cameraState.resolutions[n].x = 1280;
      cameraState.resolutions[n++].y = 400;
      cameraState.resolutions[n].x = 800;
      cameraState.resolutions[n++].y = 600;
      cameraState.resolutions[n].x = 800;
      cameraState.resolutions[n++].y = 400;
      cameraState.resolutions[n].x = 640;
      cameraState.resolutions[n++].y = 480;
      cameraState.resolutions[n].x = 600;
      cameraState.resolutions[n++].y = 400;
      cameraState.resolutions[n].x = 400;
      cameraState.resolutions[n++].y = 400;
      cameraState.resolutions[n].x = 480;
      cameraState.resolutions[n++].y = 320;
      cameraState.resolutions[n].x = 320;
      cameraState.resolutions[n++].y = 240;
      cameraState.resolutions[n].x = 0;
      cameraState.resolutions[n++].y = 0;

      cameraState.binMode2Resolutions[0].x = 640;
      cameraState.binMode2Resolutions[0].y = 512;
      cameraState.binMode2Resolutions[1].x = 0;
      cameraState.binMode2Resolutions[1].y = 0;

      cameraState.binMode3Resolutions[0].x = 0;
      cameraState.binMode3Resolutions[0].y = 0;

      cameraState.binMode4Resolutions[0].x = 320;
      cameraState.binMode4Resolutions[0].y = 256;
      cameraState.binMode4Resolutions[1].x = 0;
      cameraState.binMode4Resolutions[1].y = 0;

      break;

    case ZWOCAM_ASI120MM:
    case ZWOCAM_ASI120MC:
      cameraState.resolutions[n].x = 1280;
      cameraState.resolutions[n++].y = 960;
      cameraState.resolutions[n].x = 1280;
      cameraState.resolutions[n++].y = 720;
      cameraState.resolutions[n].x = 1280;
      cameraState.resolutions[n++].y = 600;
      cameraState.resolutions[n].x = 1280;
      cameraState.resolutions[n++].y = 400;
      cameraState.resolutions[n].x = 960;
      cameraState.resolutions[n++].y = 960;
      cameraState.resolutions[n].x = 1024;
      cameraState.resolutions[n++].y = 768;
      cameraState.resolutions[n].x = 1024;
      cameraState.resolutions[n++].y = 600;
      cameraState.resolutions[n].x = 1024;
      cameraState.resolutions[n++].y = 400;
      cameraState.resolutions[n].x = 800;
      cameraState.resolutions[n++].y = 800;
      cameraState.resolutions[n].x = 800;
      cameraState.resolutions[n++].y = 640;
      cameraState.resolutions[n].x = 800;
      cameraState.resolutions[n++].y = 512;
      cameraState.resolutions[n].x = 800;
      cameraState.resolutions[n++].y = 320;
      cameraState.resolutions[n].x = 640;
      cameraState.resolutions[n++].y = 560;
      cameraState.resolutions[n].x = 640;
      cameraState.resolutions[n++].y = 480;
      cameraState.resolutions[n].x = 512;
      cameraState.resolutions[n++].y = 440;
      cameraState.resolutions[n].x = 512;
      cameraState.resolutions[n++].y = 400;
      cameraState.resolutions[n].x = 480;
      cameraState.resolutions[n++].y = 320;
      cameraState.resolutions[n].x = 320;
      cameraState.resolutions[n++].y = 240;
      cameraState.resolutions[n].x = 0;
      cameraState.resolutions[n++].y = 0;

      cameraState.binMode2Resolutions[0].x = 640;
      cameraState.binMode2Resolutions[0].y = 480;
      cameraState.binMode2Resolutions[1].x = 0;
      cameraState.binMode2Resolutions[1].y = 0;

      cameraState.binMode3Resolutions[0].x = 0;
      cameraState.binMode3Resolutions[0].y = 0;

      cameraState.binMode4Resolutions[0].x = 0;
      cameraState.binMode4Resolutions[0].y = 0;

      break;

    case ZWOCAM_ASI035MM:
    case ZWOCAM_ASI035MC:
      cameraState.resolutions[0].x = 752;
      cameraState.resolutions[0].y = 480;
      cameraState.resolutions[1].x = 640;
      cameraState.resolutions[1].y = 480;
      cameraState.resolutions[2].x = 600;
      cameraState.resolutions[2].y = 400;
      cameraState.resolutions[3].x = 400;
      cameraState.resolutions[3].y = 400;
      cameraState.resolutions[4].x = 320;
      cameraState.resolutions[4].y = 240;
      cameraState.resolutions[5].x = 0;
      cameraState.resolutions[6].y = 0;

      if ( ZWOCAM_ASI035MM == cameraState.cameraType ) {
        cameraState.binMode2Resolutions[0].x = 376;
        cameraState.binMode2Resolutions[0].y = 240;
        cameraState.binMode2Resolutions[1].x = 0;
        cameraState.binMode2Resolutions[1].y = 0;
      } else {
        cameraState.binMode2Resolutions[0].x = 0;
        cameraState.binMode2Resolutions[0].y = 0;
      }

      cameraState.binMode3Resolutions[0].x = 0;
      cameraState.binMode3Resolutions[0].y = 0;

      cameraState.binMode4Resolutions[0].x = 0;
      cameraState.binMode4Resolutions[0].y = 0;

      break;

    case ZWOCAM_ASI030MC:
      cameraState.resolutions[0].x = 640;
      cameraState.resolutions[0].y = 480;
      cameraState.resolutions[1].x = 480;
      cameraState.resolutions[1].y = 320;
      cameraState.resolutions[2].x = 320;
      cameraState.resolutions[2].y = 240;
      cameraState.resolutions[3].x = 0;
      cameraState.resolutions[3].y = 0;

      cameraState.binMode2Resolutions[0].x = 0;
      cameraState.binMode2Resolutions[0].y = 0;

      cameraState.binMode3Resolutions[0].x = 0;
      cameraState.binMode3Resolutions[0].y = 0;

      cameraState.binMode4Resolutions[0].x = 0;
      cameraState.binMode4Resolutions[0].y = 0;

      break;

    case ZWOCAM_ASI034MC:
      cameraState.resolutions[0].x = 728;
      cameraState.resolutions[0].y = 512;
      cameraState.resolutions[1].x = 640;
      cameraState.resolutions[1].y = 480;
      cameraState.resolutions[2].x = 480;
      cameraState.resolutions[2].y = 320;
      cameraState.resolutions[3].x = 320;
      cameraState.resolutions[3].y = 240;
      cameraState.resolutions[4].x = 0;
      cameraState.resolutions[4].y = 0;

      cameraState.binMode2Resolutions[0].x = 0;
      cameraState.binMode2Resolutions[0].y = 0;

      cameraState.binMode3Resolutions[0].x = 0;
      cameraState.binMode3Resolutions[0].y = 0;

      cameraState.binMode4Resolutions[0].x = 0;
      cameraState.binMode4Resolutions[0].y = 0;

      break;

    default:
      fprintf ( stderr, "unknown camera type %d. Using limited resolutions\n",
          cameraState.cameraType );
      cameraState.resolutions[0].x = cameraState.maxResolutionX;
      cameraState.resolutions[0].y = cameraState.maxResolutionY;
      cameraState.resolutions[1].x = 0;
      cameraState.resolutions[1].y = 0;

      n = 0;
      if ( isBinSupported ( 2 )) {
        cameraState.binMode2Resolutions[n].x = 0;
        cameraState.binMode2Resolutions[n].y = 0;
        n++;
      }
      cameraState.binMode2Resolutions[n].x = 0;
      cameraState.binMode2Resolutions[n].y = 0;

      n = 0;
      if ( isBinSupported ( 3 )) {
        cameraState.binMode3Resolutions[n].x = 0;
        cameraState.binMode3Resolutions[n].y = 0;
        n++;
      }
      cameraState.binMode3Resolutions[n].x = 0;
      cameraState.binMode3Resolutions[n].y = 0;

      n = 0;
      if ( isBinSupported ( 3 )) {
        cameraState.binMode4Resolutions[n].x = 0;
        cameraState.binMode4Resolutions[n].y = 0;
        n++;
      }
      cameraState.binMode4Resolutions[n].x = 0;
      cameraState.binMode4Resolutions[n].y = 0;

      break;
  }

  cameraState.xSize = cameraState.maxResolutionX;
  cameraState.ySize = cameraState.maxResolutionY;
  cameraState.buffers = 0;
  cameraState.configuredBuffers = 0;
  cameraState.frameRateSupport = 0;

  // FIX ME -- check return code?
  initCamera();

  setImageFormat ( cameraState.xSize, cameraState.ySize, cameraState.binMode,
      cameraState.videoCurrent );

  // The largest buffer size we should need
  // RGB colour is 3 bytes per pixel, mono one for 8-bit, two for 16-bit,
  // RAW is one for 8-bit, 2 for 16-bit
  multiplier = ( IMG_RGB24 == cameraState.videoCurrent ) ? 3 :
      ( IMG_RAW16 == cameraState.videoCurrent ) ? 2 : 1;
  cameraState.imageBufferLength = cameraState.maxResolutionX *
      cameraState.maxResolutionY * multiplier;
  cameraState.buffers = calloc ( ZWASI_BUFFERS, sizeof ( struct ZWASIbuffer ));
  for ( i = 0; i < ZWASI_BUFFERS; i++ ) {
    void* m = malloc ( cameraState.imageBufferLength );
    if ( m ) {
      cameraState.buffers[i].start = m;
      cameraState.configuredBuffers++;
    } else {
      fprintf ( stderr, "%s malloc failed\n", __FUNCTION__ );
      if ( i ) {
        for ( j = 0; j < i; j++ ) {
          free (( void* ) cameraState.buffers[j].start );
        }
      }
      return 0;
    }
  }

  return camera;
}


static void
_ZWASIInitFunctionPointers ( oaCamera* camera )
{
  camera->funcs.initCamera = oaZWASIInitCamera;
  camera->funcs.closeCamera = oaZWASICloseCamera;
  camera->funcs.resetCamera = oaZWASICameraReset;
  // camera->funcs.startCapture = oaZWASICameraStart;
  camera->funcs.startCaptureExtended = oaZWASICameraStart;
  camera->funcs.stopCapture = oaZWASICameraStop;
  camera->funcs.hasLoadableFirmware = oaZWASICameraHasLoadableFirmware;
  // camera->funcs.isFirmwareLoaded = _cameraIsFirmwareLoaded;
  // camera->funcs.loadFirmware = _cameraLoadFirmware;
  camera->funcs.getName = oaZWASICameraGetName;
  camera->funcs.getControlRange = oaZWASICameraGetControlRange;
  camera->funcs.readControl = oaZWASICameraReadControl;
  camera->funcs.setControl = oaZWASICameraSetControl;
  camera->funcs.testControl = oaZWASICameraTestControl;
  // camera->funcs.hasAuto = oaHasAuto;
  // camera->funcs.isAuto = _isAuto;
  // camera->funcs.setControlMulti = _setControlMulti;
  camera->funcs.startReadFrame = oaZWASICameraStartReadFrame;
  camera->funcs.readFrame = oaZWASICameraReadFrame;
  camera->funcs.finishReadFrame = oaZWASICameraFinishReadFrame;
  camera->funcs.enumerateFrameSizes = oaZWASICameraGetFrameSizes;
  // camera->funcs.enumerateFrameRates = _enumerateFrameRates;
  camera->funcs.getFramePixelFormat = oaZWASICameraGetFramePixelFormat;
  camera->funcs.testROISize = oaZWASICameraTestROISize;
  // camera->funcs.getMenuString = _getMenuString;
  // camera->funcs.getAutoWBManualSetting = _getAutoWBManualSetting;
  // camera->funcs.hasFixedFrameRates = _hasFixedFrameRates;
}


