/*****************************************************************************
 *
 * generalSettings.cc -- class for general settings in the settings UI
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#include "generalSettings.h"

#include "configuration.h"
#include "state.h"


GeneralSettings::GeneralSettings ( QWidget* parent ) : QWidget ( parent )
{
  saveBox = new QCheckBox ( tr ( "Load and save settings automatically" ),
      this );
  saveBox->setChecked ( config.saveSettings );

  reticleButtons = new QButtonGroup ( this );
  circleButton = new QRadioButton ( tr ( "Use circular reticle" ));
  crossButton = new QRadioButton ( tr ( "Use cross reticle" ));
  tramlineButton = new QRadioButton ( tr ( "Use tramline reticle" ));
  circleButton->setChecked ( config.reticleStyle == RETICLE_CIRCLE ? 1 : 0 );
  crossButton->setChecked ( config.reticleStyle == RETICLE_CROSS ? 1 : 0 );
  tramlineButton->setChecked ( config.reticleStyle == RETICLE_TRAMLINES ?
      1 : 0 );
  reticleButtons->addButton ( circleButton );
  reticleButtons->addButton ( crossButton );
  reticleButtons->addButton ( tramlineButton );

  recentreButton = new QPushButton ( tr ( "Recentre Reticle" ));
  derotateButton = new QPushButton ( tr ( "Derotate Reticle" ));
  reticlebox = new QHBoxLayout();
  reticlebox->addWidget ( recentreButton );
  reticlebox->addWidget ( derotateButton );
  reticlebox->addStretch ( 1 );

  tempButtons = new QButtonGroup ( this );
  degCButton = new QRadioButton ( tr ( "Display temperatures in degrees C" ));
  degFButton = new QRadioButton ( tr ( "Display temperatures in degrees F" ));
  degCButton->setChecked ( config.tempsInC );
  degFButton->setChecked ( !config.tempsInC );
  tempButtons->addButton ( degCButton );
  tempButtons->addButton ( degFButton );

  saveCaptureSettings = new QCheckBox ( tr (
      "Write capture settings to file" ));
  saveCaptureSettings->setChecked ( config.saveCaptureSettings );

  connectSole = new QCheckBox ( tr (
      "Connect single camera on startup" ));
  connectSole->setChecked ( config.connectSoleCamera );

  fpsLabel = new QLabel ( tr ( "Display FPS" ), this );
  fpsCountLabel = new QLabel ( QString::number ( config.displayFPS ), this );
  fpsSlider = new QSlider ( Qt::Horizontal, this );
  fpsSlider->setRange ( 1, 30 );
  fpsSlider->setSingleStep ( 1 );
  fpsSlider->setValue ( config.displayFPS );
  fpsSlider->setTickPosition ( QSlider::TicksBelow );
  fpsSlider->setTickInterval ( 1 );

  fpsbox = new QHBoxLayout();
  fpsbox->addWidget ( fpsLabel );
  fpsbox->addWidget ( fpsCountLabel );
  fpsbox->addWidget ( fpsSlider );

  box = new QVBoxLayout ( this );
  box->addWidget ( saveBox );
  box->addSpacing ( 15 );
  box->addWidget ( circleButton );
  box->addWidget ( crossButton );
  box->addWidget ( tramlineButton );
  box->addSpacing ( 15 );
  box->addLayout ( reticlebox );
  box->addSpacing ( 15 );
  box->addWidget ( degCButton );
  box->addWidget ( degFButton );
  box->addSpacing ( 15 );
  box->addWidget ( saveCaptureSettings );
  box->addSpacing ( 15 );
  box->addWidget ( connectSole );
  box->addSpacing ( 15 );
  box->addLayout ( fpsbox );
  box->addStretch ( 1 );
  setLayout ( box );

  connect ( saveBox, SIGNAL ( stateChanged ( int )), parent,
      SLOT ( dataChanged()));
  connect ( reticleButtons, SIGNAL ( buttonClicked ( int )), parent,
      SLOT ( dataChanged()));
  connect ( saveCaptureSettings, SIGNAL ( stateChanged ( int )), parent,
      SLOT ( dataChanged()));
  connect ( connectSole, SIGNAL ( stateChanged ( int )), parent,
      SLOT ( dataChanged()));
  connect ( fpsSlider, SIGNAL ( valueChanged ( int )), parent,
      SLOT ( dataChanged()));
  connect ( fpsSlider, SIGNAL ( valueChanged ( int )), this,
      SLOT ( updateFPSLabel ( int )));
  connect ( recentreButton, SIGNAL ( clicked()), state.previewWidget,
      SLOT ( recentreReticle()));
  connect ( derotateButton, SIGNAL ( clicked()), state.previewWidget,
      SLOT ( derotateReticle()));
}


GeneralSettings::~GeneralSettings()
{
  state.mainWindow->destroyLayout (( QLayout* ) box );
}


void
GeneralSettings::storeSettings ( void )
{
  config.saveSettings = saveBox->isChecked() ? 1 : 0;
  if ( circleButton->isChecked()) {
    config.reticleStyle = RETICLE_CIRCLE;
  }
  if ( crossButton->isChecked()) {
    config.reticleStyle = RETICLE_CROSS;
  }
  if ( tramlineButton->isChecked()) {
    config.reticleStyle = RETICLE_TRAMLINES;
  }
  config.tempsInC = degCButton->isChecked();
  state.mainWindow->resetTemperatureLabel();
  config.saveCaptureSettings = saveCaptureSettings->isChecked() ? 1 : 0;
  config.connectSoleCamera = connectSole->isChecked() ? 1 : 0;

  config.displayFPS = fpsSlider->value();
  state.previewWidget->setDisplayFPS ( config.displayFPS );
}


void
GeneralSettings::updateFPSLabel ( int value )
{
  fpsCountLabel->setText ( QString::number ( value ));
}
