/*****************************************************************************
 *
 * mainWindow.cc -- the main controlling window class
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#include <QtGui>

extern "C" {
#include <openastro/oacam.h>
}

#include "version.h"
#include "configuration.h"
#include "displayWindow.h"
#include "mainWindow.h"
#include "state.h"

CONFIG		config;
STATE		state;

static const char* styleGroupBoxBorders =
  "QGroupBox { "
    "margin-top: 1ex;"
    "border: 2px groove grey;"
//  "border-radius: 5px;"
  "} "
  "QGroupBox::title { "
    "subcontrol-origin: margin;"
    "subcontrol-position: left top;"
    "left: 1em;"
  "}";


MainWindow::MainWindow()
{
  signalMapper = 0;
  rescan = disconnect = 0;
  connectedCameras = cameraMenuCreated = 0;
  doingQuit = 0;
  state.histogramOn = 0;
  state.histogramWidget = 0;
  state.needGroupBoxBorders = 0;

  QString currentStyle = QApplication::style()->objectName();
  if ( currentStyle.toLower() == "gtk+" ) {
    state.needGroupBoxBorders = 1;
    this->setStyleSheet ( styleGroupBoxBorders );
  }

  readConfig();
  createStatusBar();
  createMenus();
  setWindowTitle( APPLICATION_NAME " " VERSION_STR );

  state.mainWindow = this;
  state.controlWidget = 0;
  state.libavStarted = 0;
  state.camera = new Camera;
  oldHistogramState = -1;
  state.lastRecordedFile = "";
  updateTemperatureLabel = 0;
  state.captureIndex = 0;
  state.settingsWidget = 0;

  displayWindow = new DisplayWindow ( this );
  setCentralWidget ( displayWindow );

#if USE_APP_PATH
  QString path = QCoreApplication::applicationDirPath();
  if ( path.endsWith ( "/MacOS" )) {
    path.chop ( 6 );
  }
  oaSetRootPath ( path.toStdString().c_str());
#endif

  char d[ PATH_MAX ];
  state.currentDirectory = QString::fromAscii ( getcwd ( d, PATH_MAX ));

  if ( connectedCameras == 1 && config.connectSoleCamera ) {
    connectCamera ( 0 );
  }
}


MainWindow::~MainWindow()
{
  // FIX ME -- delete cameras[]

  state.histogramOn = 0;

  delete about;
  delete histogram;
  delete autorun;
  delete demosaic;
  delete filters;
  delete profiles;
  delete capture;
  delete cameraOpt;
  delete general;
  delete demosaicOpt;
  delete flipY;
  delete flipX;
  delete darkframe;
  delete focusaid;
  delete cutout;
  delete reticle;
  delete autoguide;
  delete alignbox;
  delete autoalign;
  delete histogramOpt;
  if ( signalMapper ) {
    delete signalMapper;
  }
  delete exit;
  if ( rescan ) {
    delete rescan;
  }
  if ( disconnect ) {
    delete disconnect;
  }
  // delete saveConfigAs;
  delete saveConfig;
  delete loadConfig;
  delete capturedValue;
  delete fpsActualValue;
  delete fpsMaxValue;
  delete progressBar;
  delete capturedLabel;
  delete fpsActualLabel;
  delete fpsMaxLabel;
  delete displayWindow;
  if ( state.camera ) {
    delete state.camera;
  }
  if ( state.histogramWidget ) {
    delete state.histogramWidget;
  }
}


void
MainWindow::readConfig ( void )
{
  QSettings settings ( ORGANISATION_NAME_SETTINGS, APPLICATION_NAME );

  // -1 means we don't have a config file.  We change it to 1 later in the
  // function
  config.saveSettings = settings.value ( "saveSettings", -1 ).toInt();

  if ( !config.saveSettings ) {

    config.tempsInC = 1;
    config.reticleStyle = 1;
    config.displayFPS = 15;

    config.showHistogram = 0;
    config.autoAlign = 0;
    config.showReticle = 0;
    config.cutout = 0;
    config.focusAid = 0;
    config.darkFrame = 0;
    config.flipX = 0;
    config.flipY = 0;
    config.demosaic = 0;

    config.sixteenBit = 0;
    config.binning2x2 = 0;
    config.rawMode = 0;

    config.useROI = 0;
    config.imageSizeX = 0;
    config.imageSizeY = 0;

    config.zoomButton1Option = 1;
    config.zoomButton2Option = 3;
    config.zoomButton3Option = 5;
    config.zoomValue = 100;

    config.controlValues [ OA_CAM_CTRL_GAIN ] = 50;
    config.controlValues [ OA_CAM_CTRL_EXPOSURE ] = 10;
    config.controlValues [ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 100;
    config.controlValues [ OA_CAM_CTRL_GAMMA ] = -1;
    config.controlValues [ OA_CAM_CTRL_BRIGHTNESS ] = -1;
    config.exposureMenuOption = 3;
    config.frameRateNumerator = 0;
    config.frameRateDenominator = 1;
    config.selectableControl[0] = OA_CAM_CTRL_GAMMA;
    config.selectableControl[1] = OA_CAM_CTRL_BRIGHTNESS;

    config.profileOption = 0;
    config.filterOption = 0;
    config.fileTypeOption = 1;
    config.limitEnabled = 0;
    config.framesLimitValue = 0;
    config.secondsLimitValue = 0;
    config.limitType = 0;
    config.fileNameTemplate = QString ( "oaCapture-%DATE-%TIME" );
    config.captureDirectory = QString ( "" );

    config.autorunCount = 0;
    config.autorunDelay = 0;
    config.saveCaptureSettings = 1;

    config.preview = 1;
    config.nightMode = 0;

    config.splitHistogram = 0;
    config.histogramOnTop = 1;

    config.demosaicPreview = 0;
    config.demosaicOutput = 0;
    config.cfaPattern = 0;
    config.demosaicMethod = 1;

    config.numProfiles = 0;
    config.numFilters = 0;

  } else {

    int version = settings.value ( "configVersion", CONFIG_VERSION ).toInt();

    restoreGeometry ( settings.value ( "geometry").toByteArray());

    // FIX ME -- how to handle this?
    // config.cameraDevice = settings.value ( "device/camera", -1 ).toInt();

    config.tempsInC = settings.value ( "tempsInCentigrade", 1 ).toInt();
    config.connectSoleCamera = settings.value ( "connectSoleCamera",
        1 ).toInt();
    config.saveCaptureSettings = settings.value ( "saveCaptureSettings",
        1 ).toInt();

    config.showHistogram = settings.value ( "options/showHistogram",
        0 ).toInt();
    config.autoAlign = settings.value ( "options/autoAlign", 0 ).toInt();
    config.showReticle = settings.value ( "options/showReticle", 0 ).toInt();
    config.cutout = settings.value ( "options/cutout", 0 ).toInt();
    config.focusAid = settings.value ( "options/focusAid", 0 ).toInt();
    config.darkFrame = settings.value ( "options/darkFrame", 0 ).toInt();
    config.flipX = settings.value ( "options/flipX", 0 ).toInt();
    config.flipY = settings.value ( "options/flipY", 0 ).toInt();
    config.demosaic = settings.value ( "options/demosaic", 0 ).toInt();

    config.sixteenBit = settings.value ( "camera/sixteenBit", 0 ).toInt();
    config.binning2x2 = settings.value ( "camera/binning2x2", 0 ).toInt();
    config.rawMode = settings.value ( "camera/raw", 0 ).toInt();
    // FIX ME -- reset these temporarily.  needs fixing properly
    config.sixteenBit = 0;
    config.binning2x2 = 0;
    config.rawMode = 0;

    config.useROI = settings.value ( "image/useROI", 0 ).toInt();
    config.imageSizeX = settings.value ( "image/imageSizeX", 0 ).toInt();
    config.imageSizeY = settings.value ( "image/imageSizeY", 0 ).toInt();

    config.zoomButton1Option = settings.value ( "image/zoomButton1Option",
        1 ).toInt();
    config.zoomButton2Option = settings.value ( "image/zoomButton2Option",
        3 ).toInt();
    config.zoomButton3Option = settings.value ( "image/zoomButton3Option",
        5 ).toInt();
    config.zoomValue = settings.value ( "image/zoomValue", 100 ).toInt();

    if ( version < 3 ) {
      config.controlValues[ OA_CAM_CTRL_GAIN ] = settings.value (
          "control/gainValue", 50 ).toInt();
      config.controlValues[ OA_CAM_CTRL_EXPOSURE ] = settings.value (
          "control/exposureValue", 10 ).toInt();
      config.controlValues[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = settings.value (
          "control/exposureAbsoluteValue", 10 ).toInt();
      config.controlValues[ OA_CAM_CTRL_GAMMA ] = settings.value (
          "control/gammaValue", -1 ).toInt();
      config.controlValues[ OA_CAM_CTRL_BRIGHTNESS ] = settings.value (
          "control/brightnessValue", -1 ).toInt();
      settings.remove( "control/gainValue" );
      settings.remove( "control/exposureValue" );
      settings.remove( "control/exposureAbsoluteValue" );
      settings.remove( "control/gammaValue" );
      settings.remove( "control/brightnessValue" );
    }

    config.exposureMenuOption = settings.value ( "control/exposureMenuOption",
        3 ).toInt();
    config.frameRateNumerator = settings.value ( "control/frameRateNumerator",
        0 ).toInt();
    config.frameRateDenominator = settings.value (
        "control/frameRateDenominator", 1 ).toInt();
    config.selectableControl[0] = settings.value (
        "control/selectableControl1", OA_CAM_CTRL_GAMMA ).toInt();
    config.selectableControl[1] = settings.value (
        "control/selectableControl2", OA_CAM_CTRL_BRIGHTNESS ).toInt();
    if ( config.selectableControl[1] == config.selectableControl[0] ) {
      config.selectableControl[1] = -1;
    }

    config.profileOption = settings.value ( "control/profileOption",
        0 ).toInt();
    config.filterOption = settings.value ( "control/filterOption", 0 ).toInt();
    config.fileTypeOption = settings.value ( "control/fileTypeOption",
        1 ).toInt();
    config.limitEnabled = settings.value ( "control/limitEnabled", 0 ).toInt();
    config.framesLimitValue = settings.value ( "control/framesLimitValue",
        0 ).toInt();
    config.secondsLimitValue = settings.value ( "control/secondsLimitValue",
        0 ).toInt();
    if ( version < 3 ) {
      settings.remove( "control/limitValue" );
    }
    config.limitType = settings.value ( "control/limitType", 0 ).toInt();
    config.fileNameTemplate = settings.value ( "control/fileNameTemplate",
        "oaCapture-%DATE-%TIME" ).toString();
    config.captureDirectory = settings.value ( "control/captureDirectory",
        "" ).toString();

    config.autorunCount = settings.value ( "autorun/count", 0 ).toInt();
    config.autorunDelay = settings.value ( "autorun/delay", 0 ).toInt();

    config.preview = settings.value ( "display/preview", 1 ).toInt();
    config.nightMode = settings.value ( "display/nightMode", 0 ).toInt();
    config.displayFPS = settings.value ( "display/displayFPS", 15 ).toInt();
    // fix a problem with existing configs
    if ( !config.displayFPS ) { config.displayFPS = 15; }

    config.splitHistogram = settings.value ( "histogram/split", 0 ).toInt();
    config.histogramOnTop = settings.value ( "histogram/onTop", 1 ).toInt();

    config.demosaicPreview = settings.value ( "demosaic/preview", 0 ).toInt();
    config.demosaicOutput = settings.value ( "demosaic/output", 0 ).toInt();
    config.demosaicMethod = settings.value ( "demosaic/method", 1 ).toInt();
    config.cfaPattern = settings.value ( "demosaic/cfaPattern", 1 ).toInt();

    config.reticleStyle = settings.value ( "reticle/style",
        RETICLE_CIRCLE ).toInt();

    if ( version > 2 ) {
      int numControls = settings.beginReadArray ( "controls" );
      if ( numControls ) {
        for ( int i = 1; i <= numControls; i++ ) {
          settings.setArrayIndex ( i-1 );
          config.controlValues[ i ] = settings.value ( "controlValue",
              0 ).toInt();
        }
      }
      settings.endArray();
    }

    config.numProfiles = settings.beginReadArray ( "profiles" );
    if ( config.numProfiles ) {
      for ( int i = 0; i < config.numProfiles; i++ ) {
        settings.setArrayIndex ( i );
        PROFILE p;
        p.profileName = settings.value ( "name", "" ).toString();
        p.sixteenBit = settings.value ( "sixteenBit", 0 ).toInt();
        p.binning2x2 = settings.value ( "binning2x2", 0 ).toInt();
        p.rawMode = settings.value ( "raw", 0 ).toInt();
        p.useROI = settings.value ( "useROI", 0 ).toInt();
        p.imageSizeX = settings.value ( "imageSizeX", 0 ).toInt();
        p.imageSizeY = settings.value ( "imageSizeY", 0 ).toInt();
        if ( version < 3 ) {
          p.controls[ OA_CAM_CTRL_GAIN ] = settings.value (
              "gainValue", 50 ).toInt();
          p.controls[ OA_CAM_CTRL_EXPOSURE ] = settings.value (
              "exposureValue", 10 ).toInt();
          p.controls[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = settings.value (
              "exposureAbsoluteValue", 10 ).toInt();
          p.controls[ OA_CAM_CTRL_GAMMA ] = settings.value (
              "gammaValue", -1 ).toInt();
          p.controls[ OA_CAM_CTRL_BRIGHTNESS ] = settings.value (
              "brightnessValue", -1 ).toInt();
          settings.remove( "gainValue" );
          settings.remove( "exposureValue" );
          settings.remove( "exposureAbsoluteValue" );
          settings.remove( "gammaValue" );
          settings.remove( "brightnessValue" );
        } else {
          int numControls = settings.beginReadArray ( "controls" );
          bzero ( p.controls, sizeof ( p.controls ));
          for ( int j = 0; j < numControls; j++ ) {
            settings.setArrayIndex ( j );
            p.controls[ j ] = settings.value ( "controlValue", 0 ).toInt();
          }
          settings.endArray();
        }
        p.frameRateNumerator = settings.value ( "frameRateNumerator",
            0 ).toInt();
        p.frameRateDenominator = settings.value ( "frameRateDenominator",
            1 ).toInt();
        p.filterOption = settings.value ( "filterOption", 0 ).toInt();
        p.fileTypeOption = settings.value ( "fileTypeOption", 1 ).toInt();
        p.fileNameTemplate = settings.value ( "fileNameTemplate",
            "oaCapture-%DATE-%TIME" ).toString();
        p.limitEnabled = settings.value ( "limitEnabled", 0 ).toInt();
        p.framesLimitValue = settings.value ( "framesLimitValue", 0 ).toInt();
        p.secondsLimitValue = settings.value ( "secondsLimitValue", 0 ).toInt();
        if ( version < 3 ) {
          settings.remove( "limitValue" );
        }
        p.limitType = settings.value ( "limitType", -1 ).toInt();
        config.profiles.append ( p );
      }
    }
    settings.endArray();

    config.numFilters = settings.beginReadArray ( "filters" );
    if ( config.numFilters ) {
      for ( int i = 0; i < config.numFilters; i++ ) {
        settings.setArrayIndex ( i );
        FILTER f;
        f.filterName = settings.value ( "name", "" ).toString();
        config.filters.append ( f );
      }
    }
    settings.endArray();
  }

  if ( !config.saveSettings || config.saveSettings == -1 ) {
    config.saveSettings = -config.saveSettings;
    // FIX ME -- these should probably be configured elsewhere
    QList<QString> defaults;
    defaults << "L" << "R" << "G" << "B" << "IR" << "UV" << "Ha" << "Hb" <<
        "S2" << "O3" << "CH4";
    config.numFilters = defaults.count();
    for ( int i = 0; i < config.numFilters; i++ ) {
      FILTER f;
      f.filterName = defaults[i];
      config.filters.append ( f );
    }
  }
}


void
MainWindow::writeConfig ( void )
{
  if ( !config.saveSettings ) {
    return;
  }

  QSettings settings ( ORGANISATION_NAME_SETTINGS, APPLICATION_NAME );

  settings.setValue ( "saveSettings", config.saveSettings );

  settings.setValue ( "configVersion", CONFIG_VERSION );
  settings.setValue ( "geometry", geometry());

  settings.setValue ( "tempsInCentigrade", config.tempsInC );
  settings.setValue ( "connectSoleCamera", config.connectSoleCamera );
  settings.setValue ( "saveCaptureSettings", config.saveCaptureSettings );

  // FIX ME -- how to handle this?
  // settings.setValue ( "device/camera", -1 ).toInt();

  settings.setValue ( "options/showHistogram", config.showHistogram );
  settings.setValue ( "options/autoAlign", config.autoAlign );
  settings.setValue ( "options/showReticle", config.showReticle );
  settings.setValue ( "options/cutout", config.cutout );
  settings.setValue ( "options/focusAid", config.focusAid );
  settings.setValue ( "options/darkFrame", config.darkFrame );
  settings.setValue ( "options/flipX", config.flipX );
  settings.setValue ( "options/flipY", config.flipY );
  settings.setValue ( "options/demosaic", config.demosaic );

  settings.setValue ( "camera/sixteenBit", config.sixteenBit );
  settings.setValue ( "camera/binning2x2", config.binning2x2 );
  settings.setValue ( "camera/raw", config.rawMode );

  settings.setValue ( "image/useROI", config.useROI );
  settings.setValue ( "image/imageSizeX", config.imageSizeX );
  settings.setValue ( "image/imageSizeY", config.imageSizeY );

  settings.setValue ( "image/zoomButton1Option", config.zoomButton1Option );
  settings.setValue ( "image/zoomButton2Option", config.zoomButton2Option );
  settings.setValue ( "image/zoomButton3Option", config.zoomButton3Option );
  settings.setValue ( "image/zoomValue", config.zoomValue );

  settings.setValue ( "control/exposureMenuOption", config.exposureMenuOption );
  settings.setValue ( "control/frameRateNumerator", config.frameRateNumerator );
  settings.setValue ( "control/frameRateDenominator",
      config.frameRateDenominator );
  settings.setValue ( "control/selectableControl1",
      config.selectableControl[0] );
  settings.setValue ( "control/selectableControl2",
      config.selectableControl[1] );

  settings.setValue ( "control/profileOption", config.profileOption );
  settings.setValue ( "control/filterOption", config.filterOption );
  settings.setValue ( "control/fileTypeOption", config.fileTypeOption );
  settings.setValue ( "control/limitEnabled", config.limitEnabled );
  settings.setValue ( "control/framesLimitValue", config.framesLimitValue );
  settings.setValue ( "control/secondsLimitValue", config.secondsLimitValue );
  settings.setValue ( "control/limitType", config.limitType );
  settings.setValue ( "control/fileNameTemplate", config.fileNameTemplate );
  settings.setValue ( "control/captureDirectory", config.captureDirectory );

  settings.setValue ( "autorun/count", config.autorunCount );
  settings.setValue ( "autorun/delay", config.autorunDelay );

  settings.setValue ( "display/preview", config.preview );
  settings.setValue ( "display/nightMode", config.nightMode );
  settings.setValue ( "display/displayFPS", config.displayFPS );

  settings.setValue ( "histogram/split", config.splitHistogram );
  settings.setValue ( "histogram/onTop", config.histogramOnTop );

  settings.setValue ( "demosaic/preview", config.demosaicPreview );
  settings.setValue ( "demosaic/output", config.demosaicOutput );
  settings.setValue ( "demosaic/method", config.demosaicMethod );
  settings.setValue ( "demosaic/cfaPattern", config.cfaPattern );

  settings.setValue ( "reticle/style", config.reticleStyle );

  settings.beginWriteArray ( "controls" );
  for ( int i = 1; i < OA_CAM_CTRL_LAST_P1; i++ ) {
    settings.setArrayIndex ( i-1 );
    settings.setValue ( "controlValue", config.controlValues[i] );
  }
  settings.endArray();

  settings.beginWriteArray ( "profiles" );
  if ( config.numProfiles ) {
    for ( int i = 0; i < config.numProfiles; i++ ) {
      settings.setArrayIndex ( i );
      settings.setValue ( "name", config.profiles[i].profileName );
      settings.setValue ( "sixteenBit", config.profiles[i].sixteenBit );
      settings.setValue ( "binning2x2", config.profiles[i].binning2x2 );
      settings.setValue ( "raw", config.profiles[i].rawMode );
      settings.setValue ( "useROI", config.profiles[i].useROI );
      settings.setValue ( "imageSizeX", config.profiles[i].imageSizeX );
      settings.setValue ( "imageSizeY", config.profiles[i].imageSizeY );
      settings.beginWriteArray ( "controls" );
      for ( int j = 0; j < OA_CAM_CTRL_LAST_P1; j++ ) {
        settings.setArrayIndex ( j );
        settings.setValue ( "controlValue", config.profiles[i].controls[ j ]);
      }
      settings.endArray();
      settings.setValue ( "frameRateNumerator",
          config.profiles[i].frameRateNumerator );
      settings.setValue ( "frameRateDenominator",
          config.profiles[i].frameRateDenominator );
      settings.setValue ( "filterOption", config.profiles[i].filterOption );
      settings.setValue ( "fileTypeOption", config.profiles[i].fileTypeOption );
      settings.setValue ( "fileNameTemplate",
          config.profiles[i].fileNameTemplate );
      settings.setValue ( "limitEnabled", config.profiles[i].limitEnabled );
      settings.setValue ( "framesLimitValue",
          config.profiles[i].framesLimitValue );
      settings.setValue ( "secondsLimitValue",
          config.profiles[i].secondsLimitValue );
      settings.setValue ( "limitType", config.profiles[i].limitType );
    }
  }
  settings.endArray();

  settings.beginWriteArray ( "filters" );
  if ( config.numFilters ) {
    for ( int i = 0; i < config.numFilters; i++ ) {
      settings.setArrayIndex ( i );
      settings.setValue ( "name", config.filters[i].filterName );
    }
  }
  settings.endArray();

}


void
MainWindow::createStatusBar ( void )
{
  statusLine = statusBar();
  setStatusBar ( statusLine );

  tempLabel = new QLabel();
  if ( config.tempsInC ) {
    tempLabel->setText ( tr ( "Temp (C)" ));
  } else {
    tempLabel->setText ( tr ( "Temp (F)" ));
  }
  tempLabel->setFixedWidth ( 60 );
  fpsMaxLabel = new QLabel ( tr ( "FPS (max)" ));
  fpsMaxLabel->setFixedWidth ( 65 );
  fpsActualLabel = new QLabel ( tr ( "FPS (actual)" ));
  fpsActualLabel->setFixedWidth ( 80 );
  capturedLabel = new QLabel ( tr ( "Captured" ));
  capturedLabel->setFixedWidth ( 60 );
  droppedLabel = new QLabel ( tr ( "Dropped" ));
  droppedLabel->setFixedWidth ( 55 );
  progressBar = new QProgressBar;
  progressBar->setFixedWidth ( 200 );
  progressBar->setRange ( 0, 100 );
  progressBar->setTextVisible ( true );

  tempValue = new QLabel ( "" );
  tempValue->setFixedWidth ( 30 );
  fpsMaxValue = new QLabel ( "0" );
  fpsMaxValue->setFixedWidth ( 30 );
  fpsActualValue = new QLabel ( "0" );
  fpsActualValue->setFixedWidth ( 30 );
  capturedValue = new QLabel ( "0" );
  capturedValue->setFixedWidth ( 40 );
  droppedValue = new QLabel ( "0" );
  droppedValue->setFixedWidth ( 40 );

  statusLine->addPermanentWidget ( tempLabel );
  statusLine->addPermanentWidget ( tempValue );
  statusLine->addPermanentWidget ( fpsMaxLabel );
  statusLine->addPermanentWidget ( fpsMaxValue );
  statusLine->addPermanentWidget ( fpsActualLabel );
  statusLine->addPermanentWidget ( fpsActualValue );
  statusLine->addPermanentWidget ( capturedLabel );
  statusLine->addPermanentWidget ( capturedValue );
  statusLine->addPermanentWidget ( droppedLabel );
  statusLine->addPermanentWidget ( droppedValue );
  statusLine->addPermanentWidget ( progressBar );
  
  statusLine->showMessage ( tr ( "started" ));
}


void
MainWindow::createMenus ( void )
{
  // FIX ME -- add "restore program defaults" option

  // File menu
  loadConfig = new QAction ( tr ( "Re&load Config" ), this );
  loadConfig->setStatusTip ( tr ( "Load default configuration" ));
  // FIX ME - set up slots

  saveConfig = new QAction ( tr ( "&Save Config" ), this );
  saveConfig->setShortcut ( QKeySequence::Save );
  saveConfig->setStatusTip ( tr ( "Save default configuration" ));
  // FIX ME - set up slots

  exit = new QAction ( tr ( "&Quit" ), this );
  exit->setShortcut ( QKeySequence::Quit );
  connect ( exit, SIGNAL( triggered()), this, SLOT( quit()));

  fileMenu = menuBar()->addMenu( tr ( "&File" ));
  fileMenu->addAction ( loadConfig );
  fileMenu->addAction ( saveConfig );
  fileMenu->addSeparator();
  fileMenu->addAction ( exit );

  // Camera device menu

  cameraMenu = menuBar()->addMenu ( tr ( "&Device" ));
  doCameraMenu(0);

  // Options menu

  histogramOpt = new QAction ( QIcon ( ":/icons/barchart.png" ),
      tr ( "Histogram" ), this );
  histogramOpt->setStatusTip ( tr ( "Open window for image histogram" ));
  histogramOpt->setCheckable ( true );
  connect ( histogramOpt, SIGNAL( changed()), this, SLOT( enableHistogram()));
  histogramOpt->setChecked ( config.showHistogram );

  autoalign = new QAction ( tr ( "Auto Align" ), this );
  autoalign->setCheckable ( true );
  // FIX ME - set up slots

  alignbox = new QAction ( tr ( "Align Box" ), this );
  alignbox->setCheckable ( true );
  // FIX ME - set up slots

  autoguide = new QAction ( tr ( "Auto Guide" ), this );
  autoguide->setCheckable ( true );
  // FIX ME - set up slots

  reticle = new QAction ( QIcon ( ":/icons/reticle.png" ),
      tr ( "Reticle" ), this );
  reticle->setStatusTip ( tr ( "Overlay a reticle on the preview image" ));
  reticle->setCheckable ( true );
  connect ( reticle, SIGNAL( changed()), this, SLOT( enableReticle()));
  reticle->setChecked ( config.showReticle );

  cutout = new QAction ( tr ( "Cut Out" ), this );
  cutout->setCheckable ( true );
  // FIX ME - set up slots

  focusaid = new QAction ( tr ( "Focus Aid" ), this );
  focusaid->setCheckable ( true );
  // FIX ME - set up slots

  darkframe = new QAction ( tr ( "Dark Frame" ), this );
  darkframe->setCheckable ( true );
  // FIX ME - set up slots

  flipX = new QAction ( QIcon ( ":/icons/object-flip-horizontal.png" ), 
      tr ( "Flip X" ), this );
  flipX->setStatusTip ( tr ( "Flip image left<->right" ));
  flipX->setCheckable ( true );
  flipX->setChecked ( config.flipX );
  connect ( flipX, SIGNAL( changed()), this, SLOT( enableFlipX()));

  flipY = new QAction ( QIcon ( ":/icons/object-flip-vertical.png" ),
      tr ( "Flip Y" ), this );
  flipY->setStatusTip ( tr ( "Flip image top<->bottom" ));
  flipY->setCheckable ( true );
  flipY->setChecked ( config.flipY );
  connect ( flipY, SIGNAL( changed()), this, SLOT( enableFlipY()));

  demosaicOpt = new QAction ( QIcon ( ":/icons/mosaic.png" ),
      tr ( "Demosaic" ), this );
  demosaicOpt->setCheckable ( true );
  demosaicOpt->setChecked ( config.demosaic );
  connect ( demosaicOpt, SIGNAL( changed()), this, SLOT( enableDemosaic()));

  optionsMenu = menuBar()->addMenu ( tr ( "&Options" ));
  optionsMenu->addAction ( histogramOpt );
#ifdef ENABLE_AUTOALIGN
  optionsMenu->addAction ( autoalign );
#endif
#ifdef ENABLE_ALIGNBOX
  optionsMenu->addAction ( alignbox );
#endif
#ifdef ENABLE_AUTOGUIDE
  optionsMenu->addAction ( autoguide );
#endif
  optionsMenu->addAction ( reticle );
#ifdef ENABLE_CUTOUT
  optionsMenu->addAction ( cutout );
#endif
#ifdef ENABLE_FOCUSAID
  optionsMenu->addAction ( focusaid );
#endif
#ifdef ENABLE_DARKFRAME
  optionsMenu->addAction ( darkframe );
#endif
  optionsMenu->addAction ( flipX );
  optionsMenu->addAction ( flipY );
  optionsMenu->addAction ( demosaicOpt );

  // settings menu

  general = new QAction ( QIcon ( ":/icons/cog.png" ),
      tr ( "General" ), this );
  general->setStatusTip ( tr ( "General configuration" ));
  connect ( general, SIGNAL( triggered()), this, SLOT( doGeneralSettings()));

  capture = new QAction ( QIcon ( ":/icons/capture.png" ),
      tr ( "Capture" ), this );
  connect ( capture, SIGNAL( triggered()), this, SLOT( doCaptureSettings()));

  cameraOpt = new QAction ( QIcon ( ":/icons/planetary-camera.png" ),
      tr ( "Camera" ), this );
  connect ( cameraOpt, SIGNAL( triggered()), this, SLOT( doCameraSettings()));
  // not enabled until we connect a camera
  cameraOpt->setEnabled ( 0 );

  profiles = new QAction ( QIcon ( ":/icons/jupiter.png" ),
      tr ( "Profiles" ), this );
  profiles->setStatusTip ( tr ( "Edit saved profiles" ));
  connect ( profiles, SIGNAL( triggered()), this, SLOT( doProfileSettings()));

  filters = new QAction ( QIcon ( ":/icons/filter-wheel.png" ),
      tr ( "Filters" ), this );
  filters->setStatusTip ( tr ( "Configuration for filters" ));
  connect ( filters, SIGNAL( triggered()), this, SLOT( doFilterSettings()));

  demosaic = new QAction ( QIcon ( ":/icons/mosaic.png" ),
      tr ( "Demosaic" ), this );
  demosaic->setStatusTip ( tr ( "Configuration for demosaicking" ));
  connect ( demosaic, SIGNAL( triggered()), this, SLOT( doDemosaicSettings()));

  autorun = new QAction ( QIcon ( ":/icons/clicknrun.png" ),
      tr ( "Autorun" ), this );
  autorun->setStatusTip ( tr ( "Configuration for repeat captures" ));
  connect ( autorun, SIGNAL( triggered()), this, SLOT( doAutorunSettings()));

  histogram = new QAction ( QIcon ( ":/icons/barchart.png" ),
      tr ( "Histogram" ), this );
  histogram->setStatusTip ( tr ( "Configuration for histogram" ));
  connect ( histogram, SIGNAL( triggered()), this,
      SLOT( doHistogramSettings()));

  settingsMenu = menuBar()->addMenu ( tr ( "&Settings" ));
  settingsMenu->addAction ( general );
  settingsMenu->addAction ( capture );
  settingsMenu->addAction ( cameraOpt );
  settingsMenu->addAction ( profiles );
  settingsMenu->addAction ( filters );
  settingsMenu->addAction ( demosaic );
  settingsMenu->addAction ( autorun );
  settingsMenu->addAction ( histogram );

  // help menu

  about = new QAction ( tr ( "About" ), this );
  connect ( about, SIGNAL( triggered()), this, SLOT( aboutDialog()));

  helpMenu = menuBar()->addMenu ( tr ( "&Help" ));
  helpMenu->addAction ( about );
}


void
MainWindow::connectCamera ( int deviceIndex )
{
  int v, ret, attempt;

  if ( -1 == oldHistogramState ) {
    oldHistogramState = state.histogramOn;
  }
  state.histogramOn = 0;
  doDisconnect();

  for ( attempt = 0, ret = 1; ret == 1 && attempt < 2; attempt++ ) {
    if (( ret = state.camera->initialise ( devs[ deviceIndex ] ))) {
      if ( !attempt && ret == 1 ) {
        if ( connectedCameras == 1 ) {
          // we think a rescan should be sufficient to identify the camera
          // as there's only one that we know about.  deviceIndex should
          // be 0 by definition for this
          int retries = 5, haveCamera = 0;
          do {
            doCameraMenu(1);
            if ( connectedCameras ) {
              haveCamera = 1;
            } else {
              sleep(1);
            }
            retries--;
          } while ( !connectedCameras && retries );
          if ( haveCamera && connectedCameras == 1 ) {
            continue;
          }
          QMessageBox::warning ( this, APPLICATION_NAME,
              tr ( "The firmware has loaded, but a rescan is required "
              "and the camera must be selected again." ));
        } else {
          QMessageBox::warning ( this, APPLICATION_NAME,
              tr ( "The firmware has loaded, but a rescan is required "
              "and the camera must be selected again." ));
        }
      } else {
        QMessageBox::warning ( this, APPLICATION_NAME,
            tr ( "Unable to connect camera" ));
      }
      state.histogramOn = oldHistogramState;
      return;
    }
  }

  disconnect->setEnabled( 1 );
  rescan->setEnabled( 0 );
  displayWindow->configure();
  statusLine->showMessage ( state.camera->name() + tr ( " connected" ), 5000 );
  clearTemperature();
  clearDroppedFrames();
  state.captureWidget->enableStartButton ( 1 );
  state.captureWidget->enableProfileSelect ( 1 );
  // FIX ME -- should these happen in the "configure" functions for each
  // widget?
  state.previewWidget->setVideoFramePixelFormat (
      state.camera->videoFramePixelFormat());
  state.cameraWidget->enableBinningControl ( state.camera->hasBinning ( 2 ));
  v = state.camera->hasControl ( OA_CAM_CTRL_TEMPERATURE );
  state.previewWidget->enableTempDisplay ( v );
  styleStatusBarTemp ( v );
  v = state.camera->hasControl ( OA_CAM_CTRL_DROPPED );
  state.previewWidget->enableDroppedDisplay ( v );
  styleStatusBarDroppedFrames ( v );
  if ( state.settingsWidget ) {
    state.settingsWidget->enableTab ( state.cameraSettingsIndex, 1 );
  }
  cameraOpt->setEnabled ( 1 );

  enableFlipX();
  enableFlipY();

  // start regardless of whether we're displaying or capturing the
  // data
  state.camera->start();
  state.histogramOn = oldHistogramState;
  oldHistogramState = -1;

  state.captureWidget->enableTIFFCapture ((
      !OA_ISBAYER( state.camera->videoFramePixelFormat()) ||
      ( config.demosaic && config.demosaicOutput )) ? 1 : 0 );
  state.previewWidget->enableScreenUpdates(1);
}


void
MainWindow::disconnectCamera ( void )
{
  if ( state.settingsWidget ) {
    state.settingsWidget->enableTab ( state.cameraSettingsIndex, 0 );
  }
  cameraOpt->setEnabled ( 0 );
  oldHistogramState = state.histogramOn;
  state.histogramOn = 0;
  state.captureWidget->enableStartButton ( 0 );
  state.captureWidget->enableProfileSelect ( 0 );
  doDisconnect();
  statusLine->showMessage ( tr ( "Camera disconnected" ));
}


void
MainWindow::doDisconnect ( void )
{
  if ( state.camera && state.camera->isInitialised()) {
    state.previewWidget->enableScreenUpdates(0);
    if ( state.captureWidget ) {
      state.captureWidget->closeOutputHandler();
    }
    state.camera->stop();
    state.camera->disconnect();
    disconnect->setEnabled( 0 );
    rescan->setEnabled( 1 );
  }
}


void
MainWindow::rescanCameras ( void )
{
  doCameraMenu(0);
}


void
MainWindow::setCapturedFrames ( unsigned int newVal )
{
  QString stringVal;

  stringVal.setNum ( newVal );
  capturedValue->setText ( stringVal );
}


void
MainWindow::setActualFrameRate ( unsigned int count )
{
  QString stringVal;

  stringVal.setNum ( count );
  fpsActualValue->setText ( stringVal );
}


void
MainWindow::setTemperature ( float temp )
{
  QString stringVal;

  if ( updateTemperatureLabel == 1 ) {
    if ( config.tempsInC ) {
      tempLabel->setText ( tr ( "Temp (C)" ));
    } else {
      tempLabel->setText ( tr ( "Temp (F)" ));
    }
    updateTemperatureLabel = 0;
  }

  if ( !config.tempsInC ) {
    temp = temp * 9 / 5 + 32;
  }
  stringVal.setNum ( temp, 'g', 3 );
  tempValue->setText ( stringVal );
}


void
MainWindow::setDroppedFrames ( unsigned int dropped )
{
  QString stringVal;

  stringVal.setNum ( dropped );
  droppedValue->setText ( stringVal );
}


void
MainWindow::resetTemperatureLabel()
{
  updateTemperatureLabel = 1;
} 
  

void
MainWindow::clearTemperature ( void )
{
  tempValue->setText ( "" );
}


void
MainWindow::clearDroppedFrames ( void )
{
  droppedValue->setText ( "" );
}

void
MainWindow::quit ( void )
{
  doingQuit = 1;
  doDisconnect();
  writeConfig();
  qApp->quit();
}


void
MainWindow::showFPSMaxValue ( int value )
{
  fpsMaxValue->setText ( QString::number ( value ));
}


void
MainWindow::clearFPSMaxValue ( void )
{
  fpsMaxValue->setText ( "" );
}


void
MainWindow::showStatusMessage ( QString message )
{
  statusLine->showMessage ( message );
}


void
MainWindow::setNightMode ( int mode )
{
  // FIX ME -- need to set flag so subwindows can be started with the
  // correct stylesheet
  if ( mode ) {
    setNightStyleSheet ( this );
    if ( state.histogramWidget ) {
      setNightStyleSheet ( state.histogramWidget );
    }
    config.nightMode = 1;
  } else {
    clearNightStyleSheet ( this );
    if ( state.histogramWidget ) {
      clearNightStyleSheet ( state.histogramWidget );
    }
    config.nightMode = 0;
  }
  update();
}


void
MainWindow::enableHistogram ( void )
{
  if ( histogramOpt->isChecked()) {
    if ( !state.histogramWidget ) {
      state.histogramWidget = new HistogramWidget();
      // need to do this to be able to uncheck the menu item on closing
      state.histogramWidget->setAttribute ( Qt::WA_DeleteOnClose );
      if ( config.histogramOnTop ) {
        state.histogramWidget->setWindowFlags ( Qt::WindowStaysOnTopHint );
      }
      connect ( state.histogramWidget, SIGNAL( destroyed ( QObject* )), this,
          SLOT ( histogramClosed()));
    }
    state.histogramWidget->show();
    state.histogramOn = 1;
    config.showHistogram = 1;
    oldHistogramState = -1;
  } else {
    if ( state.histogramWidget ) {
      state.histogramWidget->hide();
    }
    state.histogramOn = 0;
    config.showHistogram = 0;
  }
}


void
MainWindow::histogramClosed ( void )
{
  state.histogramWidget = 0;
  state.histogramOn = 0;
  if ( !doingQuit ) {
    histogramOpt->setChecked ( 0 );
  }
  oldHistogramState = 0;
  // We don't want to change this if the histogram window is closing because
  // we exited
  if ( !doingQuit ) {
    config.showHistogram = 0;
  }
}


void
MainWindow::setNightStyleSheet ( QWidget* window )
{
  QString style = QString (
      "background: rgb( 92, 0, 0 ); "
      "border-color: rgb( 192, 0, 0 ); "
  );

  if ( state.needGroupBoxBorders ) {
    style += styleGroupBoxBorders;
  }

  window->setStyleSheet ( style );
}


void
MainWindow::clearNightStyleSheet ( QWidget* window )
{
  window->setStyleSheet ( state.needGroupBoxBorders ?
      styleGroupBoxBorders : "" );
}


void
MainWindow::enableReticle ( void )
{
  config.showReticle = reticle->isChecked() ? 1 : 0;
}


void
MainWindow::setFlipX ( int state )
{
  flipX->setChecked ( state );
}


void
MainWindow::setFlipY ( int state )
{
  flipY->setChecked ( state );
}


void
MainWindow::enableFlipX ( void )
{
  int flipState = flipX->isChecked() ? 1 : 0;

  config.flipX = flipState;
  if ( state.camera->isInitialised() &&
      state.camera->hasControl ( OA_CAM_CTRL_HFLIP )) {
    state.camera->setControl ( OA_CAM_CTRL_HFLIP, flipState );
    config.controlValues [ OA_CAM_CTRL_HFLIP ] = flipState;
    if ( state.settingsWidget ) {
      state.settingsWidget->enableFlipX ( flipState );
    }
    return;
  }
  state.previewWidget->enableFlipX ( flipState );
  if ( flipState ) {
    mosaicFlipWarning();
  }
}


void
MainWindow::enableFlipY ( void )
{
  int flipState = flipY->isChecked() ? 1 : 0;

  config.flipY = flipState;
  if ( state.camera->isInitialised() &&
      state.camera->hasControl ( OA_CAM_CTRL_VFLIP )) {
    state.camera->setControl ( OA_CAM_CTRL_VFLIP, flipState );
    config.controlValues [ OA_CAM_CTRL_VFLIP ] = flipState;
    if ( state.settingsWidget ) {
      state.settingsWidget->enableFlipY ( flipState );
    }
    return;
  }
  state.previewWidget->enableFlipY ( flipState );
  if ( flipState ) {
    mosaicFlipWarning();
  }
}


void
MainWindow::mosaicFlipWarning ( void )
{
  int format = state.camera->videoFramePixelFormat();

  if ( OA_ISBAYER ( format )) {
    QMessageBox::warning ( this, APPLICATION_NAME,
        tr ( "Flipping a raw camera image may require a different colour mask to be used for demosaicking " ));
  }
}


void
MainWindow::enableDemosaic ( void )
{
  int demosaicState = demosaicOpt->isChecked() ? 1 : 0;
  int format;

  config.demosaic = demosaicState;
  state.previewWidget->enableDemosaic ( demosaicState );
  if ( state.camera->isInitialised()) {
    format = state.camera->videoFramePixelFormat ( 0 );
    state.captureWidget->enableTIFFCapture (( !OA_ISBAYER( format ) ||
        ( config.demosaic && config.demosaicOutput )) ? 1 : 0 );
  }
}


void
MainWindow::aboutDialog ( void )
{
  QMessageBox::about ( this, tr ( "About " APPLICATION_NAME ),
      tr ( "<h2>" APPLICATION_NAME " " VERSION_STR "</h2>"
      "<p>Copyright &copy; " COPYRIGHT_YEARS " " AUTHOR_NAME "</p>"
      "<p>" APPLICATION_NAME " is an open source video capture application "
      "intended primarily for planetary imaging."
      "<p>Thanks are due to numerous forum members for testing and "
      "encouragement and to those camera manufacturers who have provided "
      "documentation, Linux SDKs and other help without which this "
      "application would have taken much longer to create.</p>"
      "<p>Kudos is also due to the FFmpeg project, the libusb project "
      "and libuvc, which I have hacked without mercy." ));
}


void
MainWindow::setProgress ( int progress )
{
  progressBar->setValue ( progress );
}


void
MainWindow::doGeneralSettings ( void )
{
  createSettingsWidget();
  state.settingsWidget->setActiveTab ( state.generalSettingsIndex );
  state.settingsWidget->show();
}


void
MainWindow::doCaptureSettings ( void )
{
  createSettingsWidget();
  state.settingsWidget->setActiveTab ( state.captureSettingsIndex );
  state.settingsWidget->show();
}


void
MainWindow::doCameraSettings ( void )
{
  createSettingsWidget();
  state.settingsWidget->setActiveTab ( state.cameraSettingsIndex );
  state.settingsWidget->show();
}


void
MainWindow::doProfileSettings ( void )
{
  createSettingsWidget();
  state.settingsWidget->setActiveTab ( state.profileSettingsIndex );
  state.settingsWidget->show();
}


void
MainWindow::doFilterSettings ( void )
{
  createSettingsWidget();
  state.settingsWidget->setActiveTab ( state.filterSettingsIndex );
  state.settingsWidget->show();
}


void
MainWindow::doAutorunSettings ( void )
{
  createSettingsWidget();
  state.settingsWidget->setActiveTab ( state.autorunSettingsIndex );
  state.settingsWidget->show();
}


void
MainWindow::doHistogramSettings ( void )
{
  createSettingsWidget();
  state.settingsWidget->setActiveTab ( state.histogramSettingsIndex );
  state.settingsWidget->show();
}


void
MainWindow::doDemosaicSettings ( void )
{
  createSettingsWidget();
  state.settingsWidget->setActiveTab ( state.demosaicSettingsIndex );
  state.settingsWidget->show();
}


void
MainWindow::createSettingsWidget ( void )
{
  if ( !state.settingsWidget ) {
    state.settingsWidget = new SettingsWidget();
    state.settingsWidget->setWindowFlags ( Qt::WindowStaysOnTopHint );
    state.settingsWidget->setAttribute ( Qt::WA_DeleteOnClose );
    state.settingsWidget->enableTab ( state.cameraSettingsIndex,
        state.camera->isInitialised() ? 1 : 0 );
    connect ( state.settingsWidget, SIGNAL( destroyed ( QObject* )), this,
        SLOT ( settingsClosed()));
  }
}


void
MainWindow::settingsClosed ( void )
{
  state.settingsWidget = 0;
}


void
MainWindow::doCameraMenu ( int replaceSingleItem )
{
  int numDevs;
  int i;

  // FIX ME -- move the device type stuff into the library

  devTypes[0] = "unknown";
  devTypes[1] = "V4L2";
  devTypes[2] = "PWC";
  devTypes[3] = "ZWASI";
  devTypes[4] = "QHY";
  devTypes[5] = "IIDC";
  devTypes[6] = "UVC";

  if ( rescan ) {
    rescan->setEnabled( 0 );
  }
  if ( connectedCameras && !replaceSingleItem ) {
    for ( i = 0; i < connectedCameras; i++ ) {
      cameraMenu->removeAction ( cameras[i] );
      delete cameras[i];
    }
    delete signalMapper;
  }

  bzero ( devs, sizeof ( devs ));
  numDevs = state.camera->listConnected ( devs );

  if ( !replaceSingleItem ) {
    if ( numDevs > 0 ) {
      signalMapper = new QSignalMapper ( this );
      for ( i = 0; i < numDevs && devs[i]; i++ ) {
        QString iface ( devTypes[ devs[i]->interface ]);
        QString name ( devs[i]->deviceName );
        menuEntry[i] = "(" + iface + ") " + name;
        cameras[i] = new QAction ( menuEntry[i], this );
        if ( cameraMenuCreated ) {
          cameraMenu->insertAction ( cameraMenuSeparator, cameras[i] );
        } else {
          cameraMenu->addAction ( cameras[i] );
        }
        signalMapper->setMapping ( cameras[i], i );
        connect ( cameras[i], SIGNAL( triggered()), signalMapper, SLOT( map()));
      }
      connect ( signalMapper, SIGNAL( mapped ( int )), this,
          SLOT( connectCamera ( int )));
    }

    if ( !cameraMenuCreated ) {
      cameraMenuSeparator = cameraMenu->addSeparator();
      rescan = new QAction ( tr ( "Rescan" ), this );
      rescan->setStatusTip ( tr ( "Scan for newly connected devices" ));
      connect ( rescan, SIGNAL( triggered()), this, SLOT( rescanCameras() ));
      disconnect = new QAction ( tr ( "Disconnect" ), this );
      connect ( disconnect, SIGNAL( triggered()), this,
          SLOT( disconnectCamera()));
      disconnect->setEnabled( 0 );
      cameraMenu->addAction ( rescan );
      cameraMenu->addAction ( disconnect );
    }
  } else {
    if ( numDevs == 1 ) { // this should hopefully be true
      QString iface ( devTypes[ devs[0]->interface ]);
      QString name ( devs[0]->deviceName );
      menuEntry[0] = "(" + iface + ") " + name;
      cameras[0]->setText ( menuEntry[0] );
    } else {
      for ( i = 0; i < connectedCameras; i++ ) {
        cameras[i]->setEnabled(0);
      }
    }
  }

  cameraMenuCreated = 1;
  connectedCameras = numDevs;
  if ( rescan ) {
    rescan->setEnabled( 1 );
  }
}


void
MainWindow::closeSettingsWindow ( void )
{
  if ( state.settingsWidget ) {
    state.settingsWidget->close();
    state.settingsWidget = 0;
  }
}


void
MainWindow::destroyLayout ( QLayout* layout )
{
  QLayoutItem* item;
  QLayout*     sublayout;
  QWidget*     widget;

  if ( !layout ) { return; }

  while (( item = layout->takeAt(0))) {
    if (( sublayout = item->layout())) {
      destroyLayout ( sublayout );
    }
    else if (( widget = item->widget())) {
      widget->hide();
      delete widget;
    } else {
      delete item;
    }
  }

  delete layout;
}


void
MainWindow::styleStatusBarTemp ( int state )
{
  tempLabel->setEnabled ( state );
  tempValue->setEnabled ( state );
}


void
MainWindow::styleStatusBarDroppedFrames ( int state )
{
  droppedLabel->setEnabled ( state );
  droppedValue->setEnabled ( state );
}
