/*****************************************************************************
 *
 * previewWidget.cc -- class for the preview window in the UI (and more)
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#include <QtGui>
#include <math.h>

extern "C" {
#include <openastro/oacam.h>
#include <openastro/oademosaic.h>
}

#include "configuration.h"
#include "previewWidget.h"
#include "outputHandler.h"
#include "histogramWidget.h"
#include "state.h"


// FIX ME -- Lots of this stuff needs refactoring or placing elsewhere
// as it's really not anything to do with the actual preview window
// any more

PreviewWidget::PreviewWidget ( QWidget* parent ) : QFrame ( parent )
{
  currentZoom = 100;
  int zoomFactor = state.zoomWidget->getZoomFactor();
  // setAttribute( Qt::WA_NoSystemBackground, true );
  connect ( state.camera, SIGNAL( frameGrabbed ( void*, int )), this,
    SLOT ( updatePreview ( void*, int )));
  image = 0;
  lastCapturedFramesUpdateTime = 0;
  capturedFramesDisplayInterval = 200;
  lastDisplayUpdateTime = 0;
  frameDisplayInterval = 1000/15; // display frames per second
  previewEnabled = 1;
  videoFramePixelFormat = OA_PIX_FMT_RGB24;
  framesInLastSecond = 0;
  secondForFrameCount = 0;
  secondForTemperature = 0;
  secondForDropped = 0;
  flipX = 0;
  flipY = 0;
  movingReticle = 0;
  rotatingReticle = 0;
  rotationAngle = 0;
  recalculateDimensions ( zoomFactor );
  previewImageBuffer[0] = writeImageBuffer[0] = 0;
  previewImageBuffer[1] = writeImageBuffer[1] = 0;
  expectedSize = config.imageSizeX * config.imageSizeY *
      OA_BYTES_PER_PIXEL( videoFramePixelFormat );
  demosaic = config.demosaic;

  for ( int i = 0; i < 256; i++ ) {
    greyscaleColourTable.push_back ( QColor ( i, i, i ).rgb());
  }
  // FIX ME -- this is a hack for a race condition that I can't quite
  // pin down yet.  Even this isn't really reliable
  screenUpdatesEnabled = 1;
}


PreviewWidget::~PreviewWidget()
{
  if ( previewImageBuffer[0] ) {
    free ( previewImageBuffer[0] );
    free ( previewImageBuffer[1] );
  }
  if ( writeImageBuffer[0] ) {
    free ( writeImageBuffer[0] );
    free ( writeImageBuffer[1] );
  }
  if ( image ) {
    delete image;
    image = 0;
  }
}


void
PreviewWidget::updatePreview ( void* imageData, int length )
{
  // don't do anything if the length is not as expected
  if ( length != expectedSize ) {
    // qWarning() << "size mismatch.  have:" << length << " expected: "
    //   << expectedSize;
    state.camera->releaseImageData();
    return;
  }

  // assign the temporary buffers for image transforms if they
  // don't already exist

  if ( !previewImageBuffer[0] ) {
    previewBufferLength = config.imageSizeX * config.imageSizeY * 3;
    if (!( previewImageBuffer[0] = malloc ( previewBufferLength ))) {
      qWarning() << "preview image buffer 0 malloc failed";
      return;
    }
    if (!( previewImageBuffer[1] = malloc ( previewBufferLength ))) {
      qWarning() << "preview image buffer 1 malloc failed";
      return;
    }
  }
  if ( !writeImageBuffer[0] ) {
    writeBufferLength = config.imageSizeX * config.imageSizeY * 3;
    if (!( writeImageBuffer[0] = malloc ( writeBufferLength ))) {
      qWarning() << "write image buffer 0 malloc failed";
      return;
    }
    if (!( writeImageBuffer[1] = malloc ( writeBufferLength ))) {
      qWarning() << "write image buffer 1 malloc failed";
      return;
    }
  }

  if ( image ) {
    delete image;
    image = 0;
  }

  // FIX ME -- this is not the right way to fix the problem of this
  // code executing after the camera has been disconnected.  The race
  // condition remains, but the window is much smaller.
  // Probably should copy the data to a buffer that doesn't get removed
  // by the camera class, but then we have to deal with resizing on
  // resolution changes etc.

  if ( state.camera->captureRunning()) {
    struct timeval	t;
    int			doDisplay = 0;
    int			doHistogram = 0;
    int			previewPixelFormat, writePixelFormat;

    // write straight from the data if possible
    void* previewBuffer = imageData;
    void* writeBuffer = imageData;
    previewPixelFormat = writePixelFormat = videoFramePixelFormat;

    // do a vertical/horizontal flip if required
    if ( flipX || flipY ) {
      // this is going to make a mess for data we intend to demosaic.
      // the user will have to deal with that
      ( void ) memcpy ( writeImageBuffer[0], writeBuffer, length );
      processFlip ( writeImageBuffer[0], length, writePixelFormat );
      // both preview and write will come from this buffer for the
      // time being.  This may change later on
      previewBuffer = writeImageBuffer[0];
      writeBuffer = writeImageBuffer[0];
    }

    int reducedGreyscaleBitDepth = 0;
    int currentPreviewBuffer = -1;
    if ( OA_PIX_FMT_GREY16BE == previewPixelFormat ||
        OA_PIX_FMT_GREY16LE == previewPixelFormat ||
        OA_PIX_FMT_GRBG16LE == previewPixelFormat ||
        OA_PIX_FMT_GRBG16BE == previewPixelFormat ) {
      // As we're not already using the previewImageBuffer for preview we
      // need to copy the data to it from whatever we were going to use
      currentPreviewBuffer = 0;
      ( void ) memcpy ( previewImageBuffer[ currentPreviewBuffer ],
          previewBuffer, length );
      convert16To8Bit ( previewImageBuffer[ currentPreviewBuffer ], length,
          previewPixelFormat );
      previewBuffer = previewImageBuffer [ currentPreviewBuffer ];
      if (  OA_PIX_FMT_GREY16BE == previewPixelFormat ||
          OA_PIX_FMT_GREY16LE == previewPixelFormat ) {
        reducedGreyscaleBitDepth = 1;
      }
    }

    ( void ) gettimeofday ( &t, 0 );
    unsigned long now = ( unsigned long ) t.tv_sec * 1000 +
        ( unsigned long ) t.tv_usec / 1000;

    if ( previewEnabled ) {
      if (( lastDisplayUpdateTime + frameDisplayInterval ) < now ) {
        lastDisplayUpdateTime = now;
        doDisplay = 1;

        if ( demosaic && config.demosaicPreview ) {
          if ( OA_ISBAYER ( previewPixelFormat )) {
            currentPreviewBuffer = ( -1 == currentPreviewBuffer ) ? 0 :
                !currentPreviewBuffer;
            // Use the demosaicking to copy the data to the previewImageBuffer
            ( void ) oademosaic ( previewBuffer,
                previewImageBuffer[ currentPreviewBuffer ], config.imageSizeX,
                config.imageSizeY, 8, config.cfaPattern,
                config.demosaicMethod );
            previewBuffer = previewImageBuffer [ currentPreviewBuffer ];
          }
        }

        QImage* newImage;
        QImage* swappedImage = 0;
        if ( OA_PIX_FMT_GREY8 == videoFramePixelFormat ||
             ( OA_ISBAYER ( previewPixelFormat ) && ( !demosaic ||
             !config.demosaicPreview )) || reducedGreyscaleBitDepth ) {
          newImage = new QImage (( const uint8_t* ) previewBuffer,
              config.imageSizeX, config.imageSizeY, // config.imageSizeX,
              QImage::Format_Indexed8 );
          newImage->setColorTable ( greyscaleColourTable );
          swappedImage = newImage;
        } else {
          newImage = new QImage (( const uint8_t* ) previewBuffer,
              config.imageSizeX, config.imageSizeY, // config.imageSizeX,
              QImage::Format_RGB888 );
          if ( OA_PIX_FMT_BGR24 == previewPixelFormat ) {
            swappedImage = new QImage ( newImage->rgbSwapped());
          } else {
            swappedImage = newImage;
          }
        }

        int zoomFactor = state.zoomWidget->getZoomFactor();
        if ( zoomFactor && zoomFactor != currentZoom ) {
          recalculateDimensions ( zoomFactor );
        }
        // FIX ME -- optimise this 100% check again
//      if ( currentZoom != 100 ) {
          QImage scaledImage = swappedImage->scaled ( currentZoomX,
            currentZoomY );
          image = new QImage ( scaledImage );
          if ( swappedImage != newImage ) {
            delete swappedImage;
          }
          delete newImage;
//      } else {
//        image = newImage;
//      }
      }
    }

    OutputHandler* output = state.captureWidget->getOutputHandler();
    if ( output ) {
      output->addFrame ( writeBuffer );
      if (( lastCapturedFramesUpdateTime + capturedFramesDisplayInterval ) <
          now ) {
        state.mainWindow->setCapturedFrames ( output->getFrameCount());
        lastCapturedFramesUpdateTime = now;
      }
      state.captureIndex++;
    }

    state.camera->releaseImageData();

    framesInLastSecond++;
    if ( t.tv_sec != secondForFrameCount ) {
      secondForFrameCount = t.tv_sec;
      state.mainWindow->setActualFrameRate ( framesInLastSecond );
      framesInLastSecond = 0;
      if ( state.histogramOn ) {
        state.histogramWidget->process ( writeBuffer, length,
            writePixelFormat );
        doHistogram = 1;
      }
    }

    if ( hasTemp && t.tv_sec != secondForTemperature && t.tv_sec % 5 == 0 ) {
      state.mainWindow->setTemperature ( state.camera->getTemperature());
      secondForTemperature = t.tv_sec;
    }
    if ( hasDroppedFrames && t.tv_sec != secondForDropped &&
        t.tv_sec % 2 == 0 ) {
      state.mainWindow->setDroppedFrames (
          state.camera->readControl ( OA_CAM_CTRL_DROPPED ));
      secondForTemperature = t.tv_sec;
    }

    if ( doDisplay ) {
      update();
    }

    // check histogram control here just in case it got changed
    // this ought to be done rather more safely
    if ( state.histogramOn && state.histogramWidget && doHistogram ) {
      state.histogramWidget->update();
    }

    if ( output && config.limitEnabled ) {
      int finished = 0;
      float percentage = 0;
      int frames = output->getFrameCount();
      switch ( config.limitType ) {
        case 0: // FIX ME -- nasty magic number
          // start and current times here are in ms, but the limit value is in
          // secs, so rather than ( current - start ) / time * 100 to get the
          // %age, we do ( current - start ) / time / 10
          percentage = ( now - state.captureWidget->recordingStartTime ) /
              config.secondsLimitValue / 10;
          if ( now > state.captureWidget->recordingEndTime ) {
            finished = 1;
          }
          break;
        case 1: // FIX ME -- nasty magic number
          percentage = frames / config.framesLimitValue * 100;
          if ( frames >= config.framesLimitValue ) {
            finished = 1;
          }
          break;
      }
      if ( finished ) {
        state.captureWidget->doStopRecording();
        // these two are really just tidying up the display
        state.mainWindow->setProgress ( 100 );
        state.mainWindow->setCapturedFrames ( frames );
        if ( state.autorunEnabled ) {
          // returns non-zero if more runs are left
          if ( state.captureWidget->singleAutorunFinished()) {
            state.autorunStartNext = now + 1000 * config.autorunDelay;
          }
        }
      } else {
        state.mainWindow->setProgress ( percentage );
      }
    }

    if ( state.autorunEnabled && state.autorunStartNext &&
        now > state.autorunStartNext ) {
      state.autorunStartNext = 0;
      state.captureWidget->startNewAutorun();
    }
  }
}


void
PreviewWidget::setBuffer ( QImage* buffer )
{
  image = buffer;
}


void
PreviewWidget::configure ( void )
{
  // setGeometry ( 0, 0, config.imageSizeX, config.imageSizeY );
}


void
PreviewWidget::updatePreviewSize ( void )
{
  int zoomFactor = state.zoomWidget->getZoomFactor();
  recalculateDimensions ( zoomFactor );
  expectedSize = config.imageSizeX * config.imageSizeY *
      OA_BYTES_PER_PIXEL( videoFramePixelFormat );
  int newBufferLength = config.imageSizeX * config.imageSizeY * 3;
  if ( newBufferLength > previewBufferLength ) {
    if (!( previewImageBuffer[0] = realloc ( previewImageBuffer[0],
        newBufferLength ))) {
      qWarning() << "preview image buffer realloc failed";
    }
    if (!( previewImageBuffer[1] = realloc ( previewImageBuffer[1],
        newBufferLength ))) {
      qWarning() << "preview image buffer realloc failed";
    }
    previewBufferLength = newBufferLength;
  }
  diagonalLength = sqrt ( config.imageSizeX * config.imageSizeX +
      config.imageSizeY * config.imageSizeY );
}


void
PreviewWidget::recalculateDimensions ( int zoomFactor )
{
  currentZoom = zoomFactor;
  currentZoomX = config.imageSizeX * zoomFactor / 100;
  currentZoomY = config.imageSizeY * zoomFactor / 100;
  if ( savedXSize && savedYSize ) {
    reticleCentreX = ( reticleCentreX * currentZoomX ) / savedXSize;
    reticleCentreY = ( reticleCentreY * currentZoomY ) / savedYSize;
  } else {
    reticleCentreX = currentZoomX / 2;
    reticleCentreY = currentZoomY / 2;
  }
  setGeometry ( 0, 0, currentZoomX, currentZoomY );
  savedXSize = currentZoomX;
  savedYSize = currentZoomY;
  if ( 0 == rotationAngle ) {
    rotationTransform.reset();
  } else {
    rotationTransform = QTransform().translate ( reticleCentreX,
        reticleCentreY ).rotate ( rotationAngle ).translate ( -reticleCentreX,
        -reticleCentreY );
  }
}


void
PreviewWidget::paintEvent ( QPaintEvent* event )
{
  Q_UNUSED( event );

  // FIX ME -- there's a race condition here when switching cameras that
  // needs sorting.

  if ( image ) {
    QPainter painter ( this );

    if ( screenUpdatesEnabled ) {
      painter.drawImage ( 0, 0, *image );

      painter.setTransform ( rotationTransform );
      painter.setRenderHint ( QPainter::Antialiasing, true );
      painter.setPen ( QPen ( Qt::red, 2, Qt::SolidLine, Qt::FlatCap ));

      if ( config.showReticle ) {
        switch ( config.reticleStyle ) {

          case RETICLE_CIRCLE:
            painter.drawEllipse ( reticleCentreX - 50, reticleCentreY - 50,
              100, 100 );
            painter.drawEllipse ( reticleCentreX - 100, reticleCentreY - 100,
              200, 200 );
            painter.drawEllipse ( reticleCentreX - 200, reticleCentreY - 200,
              400, 400 );
            painter.drawLine ( reticleCentreX, reticleCentreY - 5,
              reticleCentreX, reticleCentreY + 5 );
            painter.drawLine ( reticleCentreX - 5, reticleCentreY,
              reticleCentreX + 5, reticleCentreY );
            break;

          case RETICLE_CROSS:
            // drawing lines at least the diagonal length long will mean
            // they always run to the edge of the displayed image.  The
            // widget will crop the lines as appropriate
            painter.drawLine ( reticleCentreX, reticleCentreY - diagonalLength,
                reticleCentreX, reticleCentreY - 10 );
            painter.drawLine ( reticleCentreX, reticleCentreY + diagonalLength,
                reticleCentreX, reticleCentreY + 10 );
            painter.drawLine ( reticleCentreX - diagonalLength, reticleCentreY,
                reticleCentreX - 10, reticleCentreY );
            painter.drawLine ( reticleCentreX + diagonalLength, reticleCentreY,
                reticleCentreX + 10, reticleCentreY );
            break;

          case RETICLE_TRAMLINES:
            // see comments for cross
            painter.drawLine ( reticleCentreX - 10,
                reticleCentreY - diagonalLength, reticleCentreX - 10,
                reticleCentreY + diagonalLength );
            painter.drawLine ( reticleCentreX + 10,
                reticleCentreY - diagonalLength, reticleCentreX + 10,
                reticleCentreY + diagonalLength );
            painter.drawLine ( reticleCentreX - diagonalLength,
                reticleCentreY - 10, reticleCentreX + diagonalLength,
                reticleCentreY - 10 );
            painter.drawLine ( reticleCentreX - diagonalLength,
                reticleCentreY + 10, reticleCentreX + diagonalLength,
                reticleCentreY + 10 );
            break;
        }
      }
    }
  }
}


void
PreviewWidget::mousePressEvent ( QMouseEvent* event )
{
  if ( event->button() == Qt::LeftButton ) {
    lastPointerX = event->x();
    lastPointerY = event->y();
    if (( abs ( lastPointerX - reticleCentreX ) < 20 ) &&
        ( abs ( lastPointerY - reticleCentreY ) < 20 )) {
      movingReticle = 1;
    }
  }
}


void
PreviewWidget::mouseMoveEvent ( QMouseEvent* event )
{
  int x = event->x();
  int y = event->y();
  if ( movingReticle ) {
    reticleCentreX += ( x - lastPointerX );
    reticleCentreY += ( y - lastPointerY );
  }
  lastPointerX = x;
  lastPointerY = y;
}

void
PreviewWidget::mouseReleaseEvent ( QMouseEvent* event )
{
  if ( event->button() == Qt::LeftButton ) {
    movingReticle = rotatingReticle = 0;
  }
}


void
PreviewWidget::wheelEvent ( QWheelEvent* event )
{
  int change = event->delta();
  int direction = ( change > 0 ) - ( change < 0 );
  if ( !direction ) {
    event->ignore();
  } else {
    int x = event->x();
    qreal scale = fabs ( x - reticleCentreX ) / 50;
    rotationAngle += direction * scale;
    if ( 0 == rotationAngle ) {
      rotationTransform.reset();
    } else {
      rotationTransform = QTransform().translate ( reticleCentreX,
          reticleCentreY ).rotate ( rotationAngle ).translate ( -reticleCentreX,
          -reticleCentreY );
    }
    event->accept();
  }
}


void
PreviewWidget::recentreReticle ( void )
{
  reticleCentreX = currentZoomX / 2;
  reticleCentreY = currentZoomY / 2;
}


void
PreviewWidget::derotateReticle ( void )
{
  rotationTransform.reset();
  rotationAngle = 0;
}


void
PreviewWidget::setCapturedFramesDisplayInterval ( int millisecs )
{
  capturedFramesDisplayInterval = millisecs;
}


void
PreviewWidget::setEnabled ( int state )
{
  previewEnabled = state;
}


void
PreviewWidget::setVideoFramePixelFormat ( int format )
{
  videoFramePixelFormat = format;
  expectedSize = config.imageSizeX * config.imageSizeY *
      OA_BYTES_PER_PIXEL( videoFramePixelFormat );
}


void
PreviewWidget::enableTempDisplay ( int state )
{
  hasTemp = state;
}


void
PreviewWidget::enableDroppedDisplay ( int state )
{
  hasDroppedFrames = state;
}


void
PreviewWidget::enableFlipX ( int state )
{
  flipX = state;
}


void
PreviewWidget::enableFlipY ( int state )
{
  flipY = state;
}


void
PreviewWidget::enableDemosaic ( int state )
{
  demosaic = state;
}


void
PreviewWidget::enableScreenUpdates ( int state )
{
  screenUpdatesEnabled = state;
}


void
PreviewWidget::setDisplayFPS ( int fps )
{
  frameDisplayInterval = 1000 / fps;
}


void
PreviewWidget::processFlip ( void* imageData, int length, int format )
{
  uint8_t* data = ( uint8_t* ) imageData;
  int assumedFormat = format;

  // fake up a format for mosaic frames here as properly flipping a
  // mosaicked frame would be quite hairy

  if ( OA_ISBAYER8 ( format )) {
    assumedFormat = OA_PIX_FMT_GREY8;
  } else {
    if ( OA_ISBAYER16 ( format )) {
      assumedFormat = OA_PIX_FMT_GREY16BE;
    }
  }

  switch ( assumedFormat ) {
    case OA_PIX_FMT_GREY8:
      processFlip8Bit ( data, length );
      break;
    case OA_PIX_FMT_GREY16BE:
    case OA_PIX_FMT_GREY16LE:
      processFlip16Bit ( data, length );
      break;
    case OA_PIX_FMT_RGB24:
    case OA_PIX_FMT_BGR24:
      processFlip24BitColour ( data, length );
      break;
    default:
      qWarning() << __FUNCTION__ << " unable to flip format " << format;
      break;
  }
}


void
PreviewWidget::processFlip8Bit ( uint8_t* imageData, int length )
{
  if ( flipX && flipY ) {
    uint8_t* p1 = imageData;
    uint8_t* p2 = imageData + length - 1;
    uint8_t s;
    while ( p1 < p2 ) {
      s = *p1;
      *p1++ = *p2;
      *p2-- = s;
    }
  } else {
    if ( flipX ) {
      uint8_t* p1;
      uint8_t* p2;
      uint8_t s;
      for ( int y = 0; y < config.imageSizeY; y++ ) {
        p1 = imageData + y * config.imageSizeX;
        p2 = p1 + config.imageSizeX - 1;
        while ( p1 < p2 ) {
          s = *p1;
          *p1++ = *p2;
          *p2-- = s;
        }
      }
    }
    if ( flipY ) {
      uint8_t* p1;
      uint8_t* p2;
      uint8_t s;
      p1 = imageData;
      for ( int y = config.imageSizeY - 1; y > config.imageSizeY / 2; y-- ) {
        p2 = imageData + y * config.imageSizeX;
        for ( int x = 0; x < config.imageSizeX; x++ ) {
          s = *p1;
          *p1++ = *p2;
          *p2++ = s;
        }
      }
    }
  }
}


void
PreviewWidget::processFlip16Bit ( uint8_t* imageData, int length )
{
  if ( flipX && flipY ) {
    uint8_t* p1 = imageData;
    uint8_t* p2 = imageData + length - 2;
    uint8_t s;
    while ( p1 < p2 ) {
      s = *p1;
      *p1++ = *p2;
      *p2++ = s;
      s = *p1;
      *p1++ = *p2;
      *p2 = s;
      p2 -= 2;
    }
  } else {
    if ( flipX ) {
      uint8_t* p1;
      uint8_t* p2;
      uint8_t s;
      for ( int y = 0; y < config.imageSizeY; y++ ) {
        p1 = imageData + y * config.imageSizeX;
        p2 = p1 + ( config.imageSizeX - 1 ) * 2;
        while ( p1 < p2 ) {
          s = *p1;
          *p1++ = *p2;
          *p2++ = s;
          s = *p1;
          *p1++ = *p2;
          *p2 = s;
          p2 -= 2;
        }
      }
    }
    if ( flipY ) {
      uint8_t* p1;
      uint8_t* p2;
      uint8_t s;
      p1 = imageData;
      for ( int y = config.imageSizeY - 1; y > config.imageSizeY / 2; y-- ) {
        p2 = imageData + y * config.imageSizeX * 2;
        for ( int x = 0; x < config.imageSizeX * 2; x++ ) {
          s = *p1;
          *p1++ = *p2;
          *p2++ = s;
        }
      }
    }
  }
}


void
PreviewWidget::processFlip24BitColour ( uint8_t* imageData, int length )
{
  if ( flipX && flipY ) {
    uint8_t* p1 = imageData;
    uint8_t* p2 = imageData + length - 3;
    uint8_t s;
    while ( p1 < p2 ) {
      s = *p1;
      *p1++ = *p2;
      *p2++ = s;
      s = *p1;
      *p1++ = *p2;
      *p2++ = s;
      s = *p1;
      *p1++ = *p2;
      *p2 = s;
      p2 -= 5;
    }
  } else {
    if ( flipX ) {
      uint8_t* p1;
      uint8_t* p2;
      uint8_t s;
      for ( int y = 0; y < config.imageSizeY; y++ ) {
        p1 = imageData + y * config.imageSizeX * 3;
        p2 = p1 + ( config.imageSizeX - 1 ) * 3;
        while ( p1 < p2 ) {
          s = *p1;
          *p1++ = *p2;
          *p2++ = s;
          s = *p1;
          *p1++ = *p2;
          *p2++ = s;
          s = *p1;
          *p1++ = *p2;
          *p2 = s;
          p2 -= 5;
        }
      }
    }
    if ( flipY ) {
      uint8_t* p1;
      uint8_t* p2;
      uint8_t s;
      p1 = imageData;
      for ( int y = config.imageSizeY - 1; y > config.imageSizeY / 2; y-- ) {
        p2 = imageData + y * config.imageSizeX * 3;
        for ( int x = 0; x < config.imageSizeX * 3; x++ ) {
          s = *p1;
          *p1++ = *p2;
          *p2++ = s;
        }
      }
    }
  }
}


void
PreviewWidget::convert16To8Bit ( void* imageData, int length, int format )
{
  uint8_t* t = ( uint8_t* ) imageData;
  uint8_t* s = ( uint8_t* ) imageData;

  if ( OA_PIX_FMT_GREY16LE == format || OA_PIX_FMT_GRBG16LE == format ) {
    s++;
  }
  for ( int i = 0; i < length; i += 2, s += 2 ) {
    *t++ = *s;
  }
}
